/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_cache_mem.h"
#include <assert.h>

#include <wfs_misc.h>

bool operator==( const wfs_cache_mem_key &a, const wfs_cache_mem_key &b ) {
  return (a.fid == b.fid) && (a.version == b.version) &&
    (a.block_no == b.block_no);
}

bool operator!=( const wfs_cache_mem_key &a, const wfs_cache_mem_key &b ) {
  return (a.fid != b.fid) || (a.version != b.version) ||
    (a.block_no != b.block_no);
}

bool operator<( const wfs_cache_mem_key &a, const wfs_cache_mem_key &b ) {

  bool rt;       ;
  if( a.fid != b.fid ) {
    rt = (a.fid < b.fid);
  } else if( a.version != b.version ) {
    rt = (a.version < b.version);
  } else {
    rt = (a.block_no < b.block_no);
  }

  return rt;

}

wfs_cache_mem::wfs_cache_mem( std::string cache_dir, unsigned int cache_size ) :
  wfs_cache_blockfile(cache_dir), _cache_size(cache_size), _lru(NULL),
  _mru(NULL), _curr_size(0) {

  assert( pthread_mutex_init( &_frans_cachelock, NULL ) == 0 );

}

wfs_cache_mem::~wfs_cache_mem() {
  pthread_mutex_destroy( &_frans_cachelock );
  // TODO: delete everything in the cache?
}

// Assume _frans_cachelock is already held
void wfs_cache_mem::update_ru( cache_mem_entry *entry ) {

  // new LRU?
  if( _lru == NULL ) {
    assert( entry->prev == NULL );
    assert( entry->next == NULL );
    _lru = entry;
  } else if( _lru->key == entry->key && _lru->next != NULL ) {
    _lru = _lru->next;
  }

  // make this entry the new MRU
  if( _mru != NULL && _mru->key != entry->key ) {
    if( entry->prev != NULL ) {
      entry->prev->next = entry->next;
    }
    if( entry->next != NULL ) {
      entry->next->prev = entry->prev;
    }
    entry->next = NULL;
    entry->prev = _mru;
    _mru->next = entry;
  }
  _mru = entry;

}

bool wfs_cache_mem::read_block( wfs_file_handle fh, wfs_version version,
				unsigned int block_no ) {

  wfs_cache_mem_key key( fh.fid, version, block_no );
  pthread_mutex_lock( &_frans_cachelock );
  cache_mem_entry *entry = _cache[key];
  pthread_mutex_unlock( &_frans_cachelock );
  if( entry != NULL ) {
    wfs_log(WFS_DBG_1, "Read block mem true for %016"PRIx64":%d:%d\n",
	      fh.fid, version.version_num, block_no );
    return true;
  } else {
    return wfs_cache_blockfile::read_block( fh, version, block_no );
  }
}

bool wfs_cache_mem::read_block( wfs_file_handle fh, wfs_version version,
				unsigned int block_no,
				std::string &buf ) {

  wfs_cache_mem_key key( fh.fid, version, block_no );
  pthread_mutex_lock( &_frans_cachelock );
  cache_mem_entry *entry = _cache[key];
  if( entry != NULL ) {
    buf = entry->buf;
    update_ru( entry );
    assert( _mru->key.fid == fh.fid );
    wfs_log(WFS_DBG_1, "Cache-reading block %d for %016"PRIx64", size %u, currsize %d\n",
	      block_no, fh.fid, (unsigned)buf.length(), _curr_size );
    pthread_mutex_unlock( &_frans_cachelock );
    apply_buf_writes( fh, version, block_no, buf );
    return true;
  } else {
    pthread_mutex_unlock( &_frans_cachelock );
    return wfs_cache_blockfile::read_block( fh, version, block_no, buf );
  }

}


void wfs_cache_mem::store_block( wfs_file_handle fh, wfs_version version,
				 unsigned int block_no,
				 std::string buf ) {

  wfs_cache_mem_key key( fh.fid, version, block_no );
  pthread_mutex_lock( &_frans_cachelock );
  cache_mem_entry *entry = _cache[key];
  if( entry != NULL ) {
    // update the buffer
    int size_diff = buf.length() - entry->buf.length();
    _curr_size += size_diff*sizeof(char);
    entry->buf = buf;
  } else {
    entry = new cache_mem_entry();
    entry->buf = buf;
    entry->key = key;
    _cache[key] = entry;
    _curr_size += sizeof(*entry) + buf.length()*sizeof(char);
  }
  wfs_log(WFS_DBG_1, "Caching block %d for %016"PRIx64", size %u, currsize %d\n",
	    block_no, fh.fid, (unsigned)buf.length(), _curr_size );
  update_ru( entry );
  //writeout_md( fh, version );

  while( _curr_size > _cache_size ) {
    // need to eject someone to disk
    cache_mem_entry *lru = _lru;
    _lru = lru->next;
    if( _lru != NULL ) {
      _lru->prev = NULL;
    } else {
      // if we delete the last item, then there certainly can't be an MRU
      _mru = NULL;
    }
    _curr_size -= sizeof(*lru) + lru->buf.length()*sizeof(char);
    wfs_log(WFS_DBG_1, "Ejecting block %d for %016"PRIx64", size %u, currsize %d\n",
	      lru->key.block_no, lru->key.fid, (unsigned)lru->buf.length(), _curr_size );
    _cache[lru->key] = NULL;
    wfs_file_handle fhlru;
    fhlru.fid = lru->key.fid;
    // TODO: separate thread for disk writes?  Unlock cache?
    wfs_cache_blockfile::store_block( fhlru, lru->key.version,
				      lru->key.block_no, lru->buf );
    delete lru;
  }

  pthread_mutex_unlock( &_frans_cachelock );

}

