/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "paxos.h"
#include <signal.h>
#include <sys/time.h>

// Use the Paxos algorithm to agree on a list of members.  As nodes
// join and leave, each node uses Paxos to move from one view number
// to the next view number, and all nodes go through all view
// numbers. If node 2 joins after the first node, it starts out in
// view 1 with the first node (as specified in the arguments of the
// paxos_rsm constructor) as the only member. It will propose to
// replicas in view_1 (i.e. node 1) a new view_2 containing node 1 and
// 2. When Paxos succeeds, view_2 is formed, i.e.view_2={1,2}. When
// node 3 joins, it attempts to propose view_2={1,3}, but will realize
// that view_2={1,2} already exists. It will then attempt to propose
// to replicas in view 2 a new view_3={1,2,3}. And so on.  When a node
// discovers that some node of the current view is not responding, it
// kicks off Paxos to propose a new view without the node that isn't
// responding.
//
// Typically nodes re-join the system after they recover
// from failure, and start from the latest view they were part of. To
// be able to do so correctly, Paxos logs view changes and other
// Paxos-related state.
//
// To be able to bring other nodes up to date to the latest formed
// view, each a node will have a complete history of all view numbers
// and their values that it knows about. At any time a node can reboot
// and when it re-joins, it may be many views behind; by remembering
// all views, the other nodes can bring this re-joined node up to
// date.

static void *
heartbeatthread(void *x)
{
  paxos *r = (paxos *) x;
  r->heartbeater();
  return 0;
}

static void *
managerthread(void *x)
{
  paxos *r = (paxos *) x;
  r->manager();
  return 0;
}

bool
operator> (const prop_t &a, const prop_t &b)
{
  return (a.n > b.n || (a.n == b.n && a.m > b.m));
}

bool
operator>= (const prop_t &a, const prop_t &b)
{
  return (a.n > b.n || (a.n == b.n && a.m >= b.m));
}

static bool break1 = false;
static bool break2 = false;

/**
 * Call this from your code between Phases 1 and 2 of Paxos
 */
static void
breakpoint1()
{
  if (break1) {
    wfs_log(WFS_DBG_1,"Dying at breakpoint 1!\n");
    exit(1);
  }
}

/**
 * Call this from your code between Phases 2 and 3 of Paxos
 */
static void
breakpoint2()
{
  if (break2) {
  	wfs_log(WFS_DBG_1,"Dying at breakpoint 2!\n");
    exit(1);
  }
}

void handle_usersig(int param)
{
  if (param == SIGUSR1) {
  	wfs_log(WFS_DBG_1,"Got SIGUSR1\n");
    break1 = true;
  } else if (param == SIGUSR2) {
    wfs_log(WFS_DBG_1,"Got SIGUSR2\n");
    break2 = true;
  }

}

paxos::paxos(std::string _first, std::string _me, std::string log_dir,
	     wfs_client_manager *clients, int _delay_max,
	     paxos_view_change *_vc)
  : _clients(clients), c_vid(0), stable(true), time_since_unstable(0),
    me(_me), vc(_vc), delay(true), delay_max(_delay_max)
{
  pthread_t th;

  signal( SIGUSR1, handle_usersig );
  signal( SIGUSR2, handle_usersig );

  assert (pthread_mutex_init(&pxs_mutex, NULL) == 0);

  l = new paxos_log (this, me, log_dir);
  sockaddr_in mysock;
  make_sockaddr(me.c_str(), &mysock);
  pxs = new rpcs(mysock.sin_port);

  pxs->reg(paxos_protocol::preparereq, this, &paxos::preparereq);
  pxs->reg(paxos_protocol::acceptreq, this, &paxos::acceptreq);
  pxs->reg(paxos_protocol::decidereq, this, &paxos::decidereq);
  pxs->reg(paxos_protocol::heartbeat, this, &paxos::heartbeat);

  assert(pthread_mutex_lock(&pxs_mutex)==0);
  //if the vid=0, i.e. there is no view information from the log,
  //designate the first view consisting only the "first" argument
  std::vector<std::string> initial_v;
  vid_h = c_vid;
  my_n.n = 0;
  my_n.m = me;
  n_h.n = 0;
  n_h.m = me;
  if (vid_h == 0) {
    if (me == _first) {
      initial_v.push_back(me);
      switchview(1, initial_v);
    } else {
      initial_v.push_back(_first);
      c_mems = initial_v;
      stable = false;
      delay = false;
    }
  }

  assert(pthread_cond_init(&manager_cond, NULL) == 0);
  assert(pthread_create(&th, NULL, &managerthread, (void *) this) == 0);

  if (!stable)
    pthread_cond_signal(&manager_cond);
  assert(pthread_mutex_unlock(&pxs_mutex)==0);

  assert(pthread_cond_init(&heartbeat_cond, NULL) == 0);
  assert (pthread_create(&th, NULL, &heartbeatthread, (void *) this) == 0);
}

void
paxos::get_members(std::vector<std::string> &mems)
{
  pthread_mutex_lock(&pxs_mutex);
  mems = c_mems;  //copy c_mems to mems
  pthread_mutex_unlock(&pxs_mutex);
}

rpcc *
paxos::get_cl(std::string m)
{
  return get_cl1(m);
}

rpcc *
paxos::get_cl1(std::string m)
{
  return _clients->get_client(make_node(m.c_str()));
}

void
paxos::return_cl(std::string m)
{
  _clients->done_with_client(make_node(m.c_str()));
}

std::string
paxos::print_members(const std::vector<std::string> &mems)
{
  std::string s;
  s.clear();
  for (unsigned i = 0; i < mems.size(); i++) {
    s += mems[i];
    if (i < (mems.size()-1))
      s += ",";
  }
  return s;
}

bool
paxos::isamember(std::string m, const std::vector<std::string> &mems)
{
  for (unsigned i = 0; i < mems.size(); i++) {
    if (mems[i] == m) return 1;
  }
  return 0;
}


//check if the servers in l2 contains a majority of servers in l1
bool
paxos::majority(const std::vector<std::string> &l1, const std::vector<std::string> &l2)
{
  unsigned n = 0;

  for (unsigned i = 0; i < l1.size(); i++) {
    if (isamember(l1[i], l2))
      n++;
  }
  return n >= (l1.size() >> 1) + 1;
}

void
paxos::switchview(unsigned vid, std::vector<std::string> ms)
{
  //assume pxs_mutex is held

  wfs_log(WFS_DBG_3,"switchview: switch to vid=%d v= %s\n", vid,
   		print_members(ms).c_str());

  views[vid] = ms;
  l->logview(vid, ms);
  c_vid = vid;
  c_mems = ms;
  if (vid > vid_h) {

    wfs_log(WFS_DBG_3,"highestacceptevid = %d\n", vid);

    vid_h = vid;
    my_n.n = 1;
    my_n.m = me;
    n_h.n = 0;
    n_h.m = me;
    n_a.n = 0;
    n_a.m = me;
    v_a.clear();
    stable = true;
    time_since_unstable = 0;

    //temporarily drop lock to prevent possible deadlock
    //because vc->view_change() needs to grab rsm_mutex
    //and rsm_mutex spans over pxs_mutex sometimes (see rsm.cc, start_paxos)
    assert(pthread_mutex_unlock(&pxs_mutex)==0);
    if (vc) vc->view_change();
    assert(pthread_mutex_lock(&pxs_mutex)==0);
  }
}

void
paxos::startvc()
{
  my_n.n = n_h.n + 1 > my_n.n + 1 ? n_h.n + 1 : my_n.n + 1;
}

paxos_protocol::status
paxos::heartbeat(std::string m, unsigned vid, int &r)
{
  r = (int)c_vid;

  wfs_log(WFS_DBG_4,"heartbeat from %s vid %d c_vid %d\n", m.c_str(), vid, c_vid);

  if (vid >= c_vid) {
    return paxos_protocol::OK;
  } else {
    return paxos_protocol::ERR;
  }
}

void
paxos::doheartbeat(std::string m)
{
  //lock pxs_mutex is not held
  int ret;
  int r;
  unsigned vid = c_vid;

  wfs_log(WFS_DBG_4,"doheartbeater to %s (%d)\n", m.c_str(), vid);

  rpcc *which_cl = get_cl(m);
  if( which_cl == NULL ) {
    ret = rpc_const::timeout_failure;
  } else {
    ret = which_cl->call(paxos_protocol::heartbeat, me, vid, r,
			 rpcc::to(10000));
    return_cl(m);
  }
  if (ret != paxos_protocol::OK) {

  	wfs_log(WFS_DBG_4,"doheartbeat: problem with %s (%d) my vid %d his vid %d start paxos\n", m.c_str(), ret, vid, r);

    start_paxos();
  }

  wfs_log(WFS_DBG_3,"doheartbeat done %d\n",ret);

}

void
paxos::heartbeater()
{
  struct timeval now;
  struct timespec next_timeout;
  std::string m;

  assert(pthread_mutex_lock(&pxs_mutex)==0);
  while (1) {
    gettimeofday(&now, NULL);
    next_timeout.tv_sec = now.tv_sec + 3;
    next_timeout.tv_nsec = 0;
    pthread_cond_timedwait(&heartbeat_cond, &pxs_mutex, &next_timeout);
    if (!stable) {
      // if it's been more than a minute since we got a preparereq, give up
      // on it and start our own
      if( time_since_unstable != 0 &&
	  (time(NULL) - time_since_unstable) > 60 ) {
	// can't just call start_paxos, since that requires stable to be true
	pthread_cond_signal(&manager_cond);
      }
      continue;
    }

    std::vector<std::string> mems = c_mems;
    assert(pthread_mutex_unlock(&pxs_mutex)==0);

    wfs_log(WFS_DBG_4,"heartbeater: current membership %s\n", print_members(c_mems).c_str());


    if (!isamember(me,c_mems)) {
      //this node has not joined the current membership
      start_paxos();
    }else{
      //find the node with the smallest id
      m = me;
      for (unsigned i = 0; i < mems.size(); i++) {
        if (m > mems[i])
          m = mems[i];
      }

      if (m == me) {
        //if i am the one with smallest id, ping the rest of the nodes
        for (unsigned i = 0; i < mems.size(); i++) {
          if (mems[i]!=me)
            doheartbeat(mems[i]);
        }
      } else {
        //the rest of the nodes ping the one with smallest id
        doheartbeat(m);
      }
    }
    assert(pthread_mutex_lock(&pxs_mutex)==0);
  }
}

void
paxos::start_paxos()
{
  pthread_mutex_lock(&pxs_mutex);
  // initiate Paxos by signaling manager thread
	if (stable) {
		stable = false;
		pthread_cond_signal(&manager_cond);
	}
  pthread_mutex_unlock(&pxs_mutex);
}

void
paxos::manager()
{

  // Loop continuously, initiating a new iteration of Paxos whenever
  // signalled and this node is stable (i.e., when current view is stable).

  std::vector<std::string> accepts;
  std::vector<std::string> mems;
  std::vector<std::string> mems1;
  std::vector<std::string> v;
  int vid;

  assert(pthread_mutex_lock(&pxs_mutex)==0);
  while(1) {

    wfs_log(WFS_DBG_3,"paxos::manager: highest %d mems = %s\n", vid_h,
        print_members(c_mems).c_str());


    if (stable) {
      wfs_log(WFS_DBG_3,"paxos::manager: done\n");

      pthread_cond_wait(&manager_cond, &pxs_mutex);
    }

    wfs_log(WFS_DBG_3,"paxos::manager: run paxos\n");

    startvc();
    accepts.clear();
    mems.clear();
    v.clear();
    mems = c_mems;
    if (!isamember(me, mems)) {

      wfs_log(WFS_DBG_3,"paxos::manager: add myself\n");

      mems.push_back(me);
    }
    vid = vid_h + 1;
    if (prepare(vid, accepts, mems, v)) {

      if (majority(views[vid_h], accepts)) {

        wfs_log(WFS_DBG_3,"paxos::manager: received a majority of prepare responses\n");

        if (v.size() == 0) {
          v = accepts;
          if (!isamember(me, v)) {

            wfs_log(WFS_DBG_3,"paxos::manager: add myself\n");

            v.push_back(me);
          }
        }

        breakpoint1();

        mems1 = accepts;
        accepts.clear();
        std::vector<std::string> rejects;
        accept(vid, accepts, rejects, mems1, v);

        if (majority(views[vid_h], accepts)) {

          wfs_log(WFS_DBG_3,"paxos::manager: received a majority of accept responses\n");

          breakpoint2();

          // must combine rejects and accepts, cause they all get decides
          for( unsigned i = 0; i < rejects.size(); i++ ) {
            accepts.push_back(rejects[i]);
          }

          decide(vid, accepts, v);
        } else {

          wfs_log(WFS_DBG_2,"paxos::manager: no majority of accept responses\n");

          dodelay();
        }
      } else {

        wfs_log(WFS_DBG_2,"paxos::manager: no majority of prepare responses\n");

        dodelay();
      }
    } else {

      wfs_log(WFS_DBG_2,"paxos::manager: we are behind stable? %d\n", stable);

      dodelay();
    }
    assert(pthread_mutex_unlock(&pxs_mutex)==0);
  }
}

/**
 * Wait a random amount of time between 0 and 10 seconds
 */
void
paxos::dodelay()
{
  struct timeval now;
  struct timespec next_timeout;
  int delay = random() % delay_max;
  gettimeofday(&now, NULL);
  next_timeout.tv_sec = now.tv_sec + delay;
  next_timeout.tv_nsec = 0;

  wfs_log(WFS_DBG_3,"dodelay: %ds\n", delay);

  pthread_cond_timedwait(&manager_cond, &pxs_mutex, &next_timeout);
}

bool
paxos::prepare(unsigned vid, std::vector<std::string> &accepts,
         std::vector<std::string> mems,
         std::vector<std::string> &v)
{

  wfs_log(WFS_DBG_3,"prepare: send prepare for vid %d prop %d v=%s\n",
      vid, my_n.n, print_members(mems).c_str());

  for (unsigned i = 0; i < mems.size(); i++) {
    paxos_protocol::preparearg a;
    paxos_protocol::prepareres r;
    a.vid = vid;
    a.n = my_n;
    rpcc *which_cl = get_cl1(mems[i]);
    assert(pthread_mutex_unlock(&pxs_mutex)==0);

    wfs_log(WFS_DBG_3,"prepare: about to send prepare vid %d to %s \n",
        	vid, mems[i].c_str());

    int ret;
    if( which_cl == NULL ) {
      ret = rpc_const::timeout_failure;
    } else {
      ret = which_cl->call(paxos_protocol::preparereq, me, a, r,
			   rpcc::to(10000));
      return_cl(mems[i]);
    }

    wfs_log(WFS_DBG_3,"prepare: send prepare vid %d to %s reply %d\n", vid,
     	mems[i].c_str(), ret);

    assert(pthread_mutex_lock(&pxs_mutex)==0);
    if (ret != paxos_protocol::OK) {

    	wfs_log(WFS_DBG_2,"prepare: mem %s didn't respond (%d)\n", mems[i].c_str(), ret);

    } else {
      if (r.oldview) {

        wfs_log(WFS_DBG_3,"prepare: member %s: view %d has been accepted\n",
            	mems[i].c_str(), a.vid);

        switchview(a.vid, r.mems);
        stable = false;
        delay = false;
        return false;
      } else if (!r.accept) {

      	wfs_log(WFS_DBG_3,"prepare: member %s rejects\n", mems[i].c_str());

        return false;
      } else {

      	wfs_log(WFS_DBG_3,"prepare: member %s accepts\n", mems[i].c_str());

        accepts.push_back(mems[i]);
        if (r.n_a.n > 0) {

          wfs_log(WFS_DBG_3,"prepare: an accepted value for prop %d(%s) w. v = %s\n",
              		r.n_a.n, r.n_a.m.c_str(), print_members(r.v_a).c_str());

          v = r.v_a;
        }
      }
    }
  }
  return true;
}


void
paxos::accept(unsigned vid, std::vector<std::string> &accepts,
    std::vector<std::string> &rejects,
    std::vector<std::string> mems,
    std::vector<std::string> v)
{

  wfs_log(WFS_DBG_3,"accept runs: vid %d prop %d(%s) with v = %s\n", vid,
      my_n.n, my_n.m.c_str(), print_members(v).c_str());

  for (unsigned i = 0; i < mems.size(); i++) {
    paxos_protocol::acceptarg a;
    int r;
    a.vid = vid;
    a.n = my_n;
    a.v = v;
    rpcc *which_cl = get_cl1(mems[i]);
    assert(pthread_mutex_unlock(&pxs_mutex)==0);
    int ret;
    if( which_cl == NULL ) {
      ret = rpc_const::timeout_failure;
    } else {
      ret = which_cl->call(paxos_protocol::acceptreq,
          me, a, r, rpcc::to(10000));
      return_cl(mems[i]);
    }
    assert(pthread_mutex_lock(&pxs_mutex)==0);
    if (ret != paxos_protocol::OK) {

      wfs_log(WFS_DBG_2,"accept: member %s didn't respond\n", mems[i].c_str());

      continue;
    } else {
      if (!r) {

        wfs_log(WFS_DBG_3,"accept: member %s rejects view\n", mems[i].c_str());

        rejects.push_back(mems[i]);
      } else {

        wfs_log(WFS_DBG_3,"accept: member %s accepts view\n", mems[i].c_str());

        accepts.push_back(mems[i]);
      }
    }
  }
}

void
paxos::decide(unsigned vid, std::vector<std::string> accepts,
    std::vector<std::string> v)
{

  wfs_log(WFS_DBG_3,"decide: vid_h %d vid %d\n", vid_h, vid);

  for (unsigned i = 0; i < accepts.size(); i++) {
    paxos_protocol::decidearg a;
    int r;
    a.vid = vid;
    a.v = v;
    rpcc *which_cl = get_cl1(accepts[i]);
    assert(pthread_mutex_unlock(&pxs_mutex)==0);

    wfs_log(WFS_DBG_3,"decide: about to send to %s\n", accepts[i].c_str());

    int ret;
    if( which_cl == NULL ) {
      ret = rpc_const::timeout_failure;
    } else {
      ret = which_cl->call(paxos_protocol::decidereq, me, a, r,
          rpcc::to(10000));
      return_cl(accepts[i]);
    }

    wfs_log(WFS_DBG_3,"decide: finished send to %s\n", accepts[i].c_str());

    assert(pthread_mutex_lock(&pxs_mutex)==0);
    if (ret != paxos_protocol::OK) {

      wfs_log(WFS_DBG_2,"decide: member %s didn't respond %d\n", accepts[i].c_str(), ret);

      continue;
    }
  }

  wfs_log(WFS_DBG_3,"decide: vid_h %d vid %d finished\n", vid_h, vid);

}

paxos_protocol::status
paxos::preparereq(std::string src, paxos_protocol::preparearg a,
    paxos_protocol::prepareres &r)
{

  // handle a preparereq message in Phase 1 of Paxos
  bool run_paxos = false;
  assert(pthread_mutex_lock(&pxs_mutex)==0);


  wfs_log(WFS_DBG_3,"preparereq: %s for vid %d n %d(%s) hvid %d n_h %d(%s) n_a %d(%s)\n",
      src.c_str(), a.vid, a.n.n, a.n.m.c_str(),
   	  vid_h, n_h.n, n_h.m.c_str(), n_a.n, n_a.m.c_str());


  if (a.vid <= vid_h) {
    r.oldview = true;
    r.mems = views[a.vid];

  	wfs_log(WFS_DBG_3,"preparereq: prepare for an already accepted vid v = %s\n",
        print_members(r.mems).c_str());

  } else if (a.vid > vid_h + 1) {

  	wfs_log(WFS_DBG_3,"preparereq: i am behind!\n");

    r.n_a = n_h;
    r.accept = false;
    r.oldview = false;
    run_paxos = true;
  } else {    // the next view
    r.oldview = false;
    if (a.n > n_h) {

      wfs_log(WFS_DBG_3,"preparereq: a higher proposal\n");

      n_h = a.n;
      l->loghigh(n_h);
      stable = false;
      time_since_unstable = time(NULL);
      r.accept = true;
      r.n_a = n_a;
      r.v_a = v_a;
    } else {

      wfs_log(WFS_DBG_3,"preparereq: a lower proposal---reject\n");

      r.n_a = n_h; //jy: tells the failed proposer my n_h value to try again
      r.accept = false;
    }
  }
  assert(pthread_mutex_unlock(&pxs_mutex)==0);
  if (run_paxos)
    start_paxos(); //run my own paxos to catch up
  return paxos_protocol::OK;

}

paxos_protocol::status
paxos::acceptreq(std::string src, paxos_protocol::acceptarg a, int &r)
{

  // handle an acceptreq message in Phase 2 of Paxos

  assert(pthread_mutex_lock(&pxs_mutex)==0);

  wfs_log(WFS_DBG_3,"acceptreq: prop (%d,%s) for view %d w. v = %s\n",
      a.n.n, a.n.m.c_str(), a.vid, print_members(a.v).c_str());

  if (a.vid == vid_h + 1) {
    if (a.n >= n_h) {

      wfs_log(WFS_DBG_3,"acceptreq: accept proposal  n_h = %d(%s)\n",
          n_h.n, n_h.m.c_str());

      n_a = a.n;
      v_a = a.v;
      l->logprop(n_a, v_a);
      r = 1;
    } else {

      wfs_log(WFS_DBG_3,"acceptreq: reject proposal n_h = %d(%s)\n",
          n_h.n, n_h.m.c_str());

      r = 0;
    }
  } else
    r = 0;
  assert(pthread_mutex_unlock(&pxs_mutex)==0);
  return paxos_protocol::OK;
}

paxos_protocol::status
paxos::decidereq(std::string src, paxos_protocol::decidearg a, int &r)
{

  // handle an decide message in Phase 3 of Paxos
  assert(pthread_mutex_lock(&pxs_mutex)==0);

  wfs_log(WFS_DBG_3,"decidereq for accepted vid %d (my vid %d) v=%s a=%s\n", a.vid,
	   vid_h, print_members(v_a).c_str(), print_members(a.v).c_str());
  if (a.vid == vid_h + 1) {
    switchview(a.vid, a.v);
  } else if (a.vid <= vid_h) {
    // we are ahead ignore.
  } else {
    // we are behind
    assert(0);
  }
  assert(pthread_mutex_unlock(&pxs_mutex)==0);
  return paxos_protocol::OK;
}

