/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "rsm_client.h"
#include <vector>
#include <arpa/inet.h>

// XXX need lock to protect dsts
// XXX members should timeout?

rsm_client::rsm_client(std::string dst, wfs_client_manager *clients) :
  _clients(clients)
{
  std::vector<std::string> mems;

  master.id = dst;
  master.node = make_node(dst.c_str());
  assert(pthread_mutex_init(&_master_lock, NULL) == 0);
  assert (init_members(true));
}

rsm_client::~rsm_client() {
  assert(pthread_mutex_destroy(&_master_lock) == 0);
}

void
rsm_client::master_failure()
{
  assert(pthread_mutex_lock(&_master_lock) == 0);
  if (known_mems.back() == master.id) {
    known_mems.pop_back();
    known_mems.push_front(master.id);
  }
  assert(pthread_mutex_unlock(&_master_lock) == 0);
  init_members(false);
}

rsm_protocol::status
rsm_client::invoke(int proc, std::string req, std::string &rep)
{
  int ret;
  while (1) {

    wfs_log(WFS_DBG_3,"rsm_client::invoke proc %x master %s\n", proc, master.id.c_str());

	wfs_node node;
	rpcc *client = get_rpcc(node);
	if( client == NULL ) {
	  ret = rpc_const::timeout_failure;
	} else {
	  ret = client->call(rsm_client_protocol::invoke, proc, req,
			     rep, rpcc::to(10000),"host");
	  return_rpcc(node);
	}

    wfs_log(WFS_DBG_3,"rsm_client::invoke proc %x master %s done %d\n", proc,
     	  master.id.c_str(), ret);

    if (ret == rsm_client_protocol::OK) {
      break;
    }
    // you have to handle the cases when the RSM replies BUSY, NOTMASTER or simply fails.
    if (ret == rsm_client_protocol::BUSY) {

      wfs_log(WFS_DBG_2,"rsm is busy %s\n", master.id.c_str());

      sleep(3);
      continue;
    }
    if (ret == rsm_client_protocol::NOTMASTER) {

      wfs_log(WFS_DBG_2,"master %s isn't the master--let's get a complete list of mems\n",
          master.id.c_str());

      if (init_members(true))
        continue;
    }

    wfs_log(WFS_DBG_2,"master %s failed ret %d\n", master.id.c_str(), ret);

    master_failure();

    wfs_log(WFS_DBG_2,"rsm_client::invoke: retry new master %s\n", master.id.c_str());

  }
  return ret;
}

bool
rsm_client::init_members(bool send_member_rpc)
{
  if (send_member_rpc) {
    std::vector<std::string> v;
    wfs_node node;
    rpcc *client = get_rpcc(node);
    int ret;
    if( client == NULL ) {
      ret = rpc_const::timeout_failure;
    } else {
      ret = client->call(rsm_client_protocol::members, 0, v,
			 rpcc::to(10000),"host");
      return_rpcc(node);
    }
    if (ret != rsm_protocol::OK || v.size() == 0){

      wfs_log(WFS_DBG_1, "init_members failed %d %u\n", ret, (unsigned)v.size() );
      return false;
    }

    for (unsigned i =0; i < v.size(); i++) {
      std::list<std::string>::iterator j;
      for (j = known_mems.begin(); j != known_mems.end(); j++) {
        if ((*j) == v[i])
          break;
      }
      if (j == known_mems.end()) {
        known_mems.push_back(v[i]);
      } else if (i == v.size()-1) {
        //the last member in v is the master
        known_mems.erase(j);
        known_mems.push_back(v[i]);
      }
    }
  }

  if (known_mems.size() < 1) {

  	wfs_log(WFS_DBG_1,"rsm_client::init_members does not know any members!\n");
    assert(0);
  }

  std::string new_master = known_mems.back();
  known_mems.pop_back();
  known_mems.push_front(new_master);

  if (new_master != master.id) {
    assert(pthread_mutex_lock(&_master_lock) == 0);
    master.id = new_master;
    master.node = make_node(new_master.c_str());
    assert(pthread_mutex_unlock(&_master_lock) == 0);
  }
  return true;
}

