/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef rsm_client_h
#define rsm_client_h

#include "rpc.h"
#include "rsm_protocol.h"
#include "wfs_misc.h"
#include <wfs_client_manager.h>
#include <string>
#include <vector>

//
// rsm client interface.
//
// The client stubs package up an rpc, and then call the invoke procedure
// on the replicated state machine passing the RPC as an argument.  This way
// the replicated state machine isn't service specific; any server can use it.
//

class rsm_client {

  struct master_t {
    wfs_node node;
    std::string id;
  };

 protected:
  master_t master;
  std::list<std::string> known_mems;
  wfs_client_manager *_clients;
  pthread_mutex_t _master_lock;

  void master_failure();
  bool init_members(bool send_mem_rpc);
 public:
  
  rpcc * get_rpcc( wfs_node &node ) { 
    assert(pthread_mutex_lock(&_master_lock) == 0);
    node.id = master.node.id;
    node.addr = master.node.addr;
    node.site_id = master.node.site_id;
    assert(pthread_mutex_unlock(&_master_lock) == 0);
    return _clients->get_client(node);
  }
  void return_rpcc( wfs_node node ) {_clients->done_with_client(node);}
  rsm_client(std::string dst, wfs_client_manager *clients);
  ~rsm_client();
  rsm_protocol::status invoke(int proc, std::string req, std::string &rep);

  template<class R, class A1>
    int call(unsigned int proc, const A1 & a1, R &r);

  template<class R, class A1, class A2>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, R &r);

  template<class R, class A1, class A2, class A3>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3,
	     R &r);

  template<class R, class A1, class A2, class A3, class A4>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3,
	     const A4 & a4, R &r);

  template<class R, class A1, class A2, class A3, class A4, class A5>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3,
	     const A4 & a4, const A5 & a5, R &r);
};

template<class R, class A1> int
  rsm_client::call(unsigned int proc, const A1 & a1, R & r)
{
  marshall m;
  std::string rep;
  std::string res;
  m << a1;
  int intret = invoke(proc, m.str(), rep);
  unmarshall u(rep);
  u >> intret;
  u >> res;
  unmarshall u1(res);
  u1 >> r;
  return intret;
}

template<class R, class A1, class A2> int
  rsm_client::call(unsigned int proc, const A1 & a1, const A2 & a2, R & r)
{
  marshall m;
  std::string rep;
  std::string res;
  m << a1;
  m << a2;
  int intret = invoke(proc, m.str(), rep);
  unmarshall u(rep);
  u >> intret;
  u >> res;
  unmarshall u1(res);
  u1 >> r;
  return intret;
}

template<class R, class A1, class A2, class A3> int
  rsm_client::call(unsigned int proc, const A1 & a1,
		const A2 & a2, const A3 & a3, R & r)
{
  marshall m;
  std::string rep;
  std::string res;
  m << a1;
  m << a2;
  m << a3;
  int intret = invoke(proc, m.str(), rep);
  unmarshall u(rep);
  u >> intret;
  u >> res;
  unmarshall u1(res);
  u1 >> r;
  return intret;
}

template<class R, class A1, class A2, class A3, class A4> int
  rsm_client::call(unsigned int proc, const A1 & a1,
		   const A2 & a2, const A3 & a3, const A4 & a4, R & r)
{
  marshall m;
  std::string rep;
  std::string res;
  m << a1;
  m << a2;
  m << a3;
  m << a4;
  int intret = invoke(proc, m.str(), rep);
  unmarshall u(rep);
  u >> intret;
  u >> res;
  unmarshall u1(res);
  u1 >> r;
  return intret;
}

template<class R, class A1, class A2, class A3, class A4, class A5> int
  rsm_client::call(unsigned int proc, const A1 & a1,
		   const A2 & a2, const A3 & a3, const A4 & a4, const A5 & a5,
		   R & r)
{
  marshall m;
  std::string rep;
  std::string res;
  m << a1;
  m << a2;
  m << a3;
  m << a4;
  m << a5;
  int intret = invoke(proc, m.str(), rep);
  unmarshall u(rep);
  u >> intret;
  u >> res;
  unmarshall u1(res);
  u1 >> r;
  return intret;
}

#endif
