/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// Fault tolerant Configuration service. Uses a RSM and paxos - Main file

#include <arpa/inet.h>
#include <getopt.h>
#include <assert.h>
#include <netdb.h>
#include <rpc.h>
#include <vivaldi.h>
#include <wfs_misc.h>
#include <assert.h>
#include <signal.h>

#include "rpc.h"
#include "paxos.h"
#include "rsm.h"
#include "wfs_rsm_cserver.h"


static void usage() {
  fprintf( stderr, "Usage for wfs_rsm_cserver:\n\n"
  	   "\t--master <master:port>\n"
	   "\t--this-host <me:port>\n"
	   "\t--log-dir <directory>\n"
	   "\t--debug (may be given multiple times)\n"
	   "\t--timeout <timeout_secs>\n"
	   "\t--pxs-delay-max <delay_secs>\n"
       "\t--dbg-dir <debug_log_directory>\n"
#ifdef WFS_WITH_LIBSSH
	   "\t[--secure]\n"
	   "\t[--keydir <path>] (defaults to ~/.wfs/)\n"
#endif
	   );
  exit(-1);

}

void clean_exit( int signal ) {
  printf( "clean exit!\n" );
  exit(0);
}

int main( int argc, char *argv[] ) {

  std::string master;
  std::string me;
  std::string log_dir = ".";
  std::string dbg_dir = ".";
  int debug_level = 0;
  bool secure = false;
  const char *key_dir = default_keydir;

  int c;
  optind = 0;  // reset getopt
  int timeout = 60; // in seconds
  int pxs_delay_max = 10; // in seconds

  while(1) {
  	static struct option long_options[] =
  	{
  	  {"master", required_argument, 0, 'm'},
      {"this-host", required_argument, 0, 'l'},
      {"log-dir", required_argument, 0, 'o'},
      {"timeout", required_argument, 0, 't'},
      {"paxos-delay-max", required_argument, 0, 'p'},
      {"debug", no_argument, 0, 'd'},
      {"dbg-dir", required_argument, 0, 'g'},
#ifdef WFS_WITH_LIBSSH
      {"secure", no_argument, 0, 'S'},
      {"keydir", required_argument, 0, 'K'},
#endif
      {"help", no_argument, 0, 'h'},
      {0, 0, 0, 0}
    };

    int option_index = 0;
    c = getopt_long( argc, argv, "m:l:o:dt:g:p:Sh?",
		     long_options, &option_index );

    if( c == -1 ) {
      break;
    }

    switch(c) {

    case 'm':
      master = optarg;
      break;

    case 'l':
      me = optarg;
      break;

    case 'o':
      log_dir = optarg;
      break;

    case 't':
      timeout = atoi(optarg);
      assert( timeout > 0 );
      break;

    case 'p':
      pxs_delay_max = atoi(optarg);
      assert( pxs_delay_max > 0 );
      break;

    case 'd':
      debug_level++;
      break;

    case 'g':
      dbg_dir = optarg;
      break;
#ifdef WFS_WITH_LIBSSH
    case 'S':
      secure = true;
      break;
    case 'K':
      key_dir = optarg;
      break;
#endif

    case '?':
    case 'h':
      /* getopt_long already printed an error message. */
      usage();
      break;

    default:
      usage();
    }
  }
  TidWFSLogAugmentor * taug = new TidWFSLogAugmentor();
  WFSLogInit  __loginit("cserver."+me, dbg_dir, 100, 5, taug);

  if( debug_level > 0 ) {
    wfs_set_debug( debug_level );
    wfs_log( WFS_DBG_1, "DEBUG LEVEL: %d\n", debug_level );
  }

  if (secure) {
    rpc_enable_security(wfs_public_key_identifies_user,
			wfs_get_private_key,
			"host");
    wfs_set_key_dir(key_dir);
  }

  struct timeval tv;
  gettimeofday(&tv, NULL);
  sockaddr_in mysock;
  make_sockaddr(me.c_str(), &mysock);
  srandom( tv.tv_usec^mysock.sin_addr.s_addr^mysock.sin_port );

  wfs_thread_id_init();

  if( me.empty() || master.empty() ) {
    fprintf( stderr, "Required parameter is missing.\n" );
    usage();
  }

  signal( SIGINT, &clean_exit );

  fflush(stdout);
  setlinebuf(stdout);

  //server.setlossy(true);
  //TODO: add vivaldy wrapper to RSM
  //viv = new vivaldi();
  //server.set_vivaldi(viv);
  vivaldi viv;
  wfs_client_manager clients(&viv);

  rsm rsm(master, me, log_dir, &clients, pxs_delay_max);

  //wfs_rsm_cserver cserver( timeout, &rsm);
  wfs_rsm_cserver cserver(&rsm, secure);

  // New Handlers
  rsm.set_state_transfer((rsm_state_transfer *)&cserver);
  rsm.reg(wcc_protocol::acquire, &cserver, &wfs_rsm_cserver::acquire,&wfs_rsm_cserver::can_acquire);
  rsm.reg(wcc_protocol::release, &cserver, &wfs_rsm_cserver::release,&wfs_rsm_cserver::can_release);
  rsm.reg(wcc_protocol::renew, &cserver,&wfs_rsm_cserver::renew, &wfs_rsm_cserver::can_renew);
  // Read only handlers.
  rsm.reg_ro(wcc_protocol::fetch_slices, &cserver, &wfs_rsm_cserver::fetch_slices);

  // master stuff
  rsm.reg(rsm_reserved::became_master, &cserver,
	  &wfs_rsm_cserver::became_master);
  rsm.reg(rsm_reserved::lost_master, &cserver,
	  &wfs_rsm_cserver::lost_master);

  // auth stuff
  rsm.reg(wcc_protocol::update_auth_host_keys,
	  &cserver, &wfs_rsm_cserver::update_auth_host_keys,
	  &wfs_rsm_cserver::can_update_auth_host_keys);
  rsm.reg(wcc_protocol::update_auth_user_keys,
	  &cserver, &wfs_rsm_cserver::update_auth_user_keys,
	  &wfs_rsm_cserver::can_update_auth_user_keys);

  rsm.recover_from_log();

  while(1)
    sleep(1000);
}
