/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_rsm_cclient.h"
#include "wfs_misc.h"
#include "host_marshall.h"
#include "method_thread.h"
#include <iostream>

// Need to actually define it somewhere
const double wfs_rsm_cclient::RENEWAL_MARGIN;

wfs_rsm_cclient::wfs_rsm_cclient(std::string xdst, host me,
				 wfs_client_manager *clients)
: this_host(me) , _shutting_down(false)
{
  pthread_mutex_init(&renew_slices_mutex, NULL);
  assert(pthread_cond_init(&schedule_change, 0) == 0);
  rsmc = new rsm_client(xdst, clients);
  if((th_renewer = method_thread(this, false, &wfs_rsm_cclient::slice_renewer)) == 0){
    printf("rpcc::rpcc pthread_create");
    exit(1);
  }
}
wfs_rsm_cclient::~wfs_rsm_cclient()
{
  _shutting_down = true;
  pthread_cond_signal(&schedule_change);
  lockedmap::iterator iter;
  //assert(pthread_cancel(th_renewer) == 0);
  assert(pthread_join(th_renewer, NULL) == 0);
  release_set.clear();
  assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
  for(iter = held_locks.begin();iter!=held_locks.end();iter++){
    delete iter->second;
  }
  assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
  pthread_mutex_destroy(&renew_slices_mutex);
  assert(pthread_cond_destroy(&schedule_change) == 0);
}

int
wfs_rsm_cclient::acquire_slice(slice_assignment* psa,const unsigned int period, bool renew)
{
  //struct timeval start,end;
  wcc_protocol::ret ret;
  int r;
  const std::string state = psa->to_canonical_string();

  std::ostringstream ost;
  ost << psa->get_primary();
  wfs_slice_id sliceid = psa->slice.val();
  // TODO: Check whether the primary is this host. O/W return an error.

  wfs_log(WFS_DBG_3,"wfs_rsm_cclient::acquire host:%s slice:%016"PRIx64" period:%u \n\tslice state: %s\n",
        ost.str().c_str(),sliceid,period,state.c_str());

#ifdef PLAB_TEST
  // TODO:tests only
  gettimeofday(&start, NULL);
#endif
  r = rsmc->call(wcc_protocol::acquire,psa->get_primary(),psa->slice,state, period, ret);

  wfs_log(WFS_DBG_3,"wfs_rsm_cclient::acquire for slice:%016"PRIx64" return code:%d.\n",
        sliceid,r);

  switch(r){
  case wcc_protocol::OK:

    wfs_log(WFS_DBG_3,"wfs_rsm_cclient::acquire slice:%016"PRIx64" acquired successfully for %d seconds.\n", sliceid,ret.tryagain);

#ifdef PLAB_TEST
    // TODO: tests only
    gettimeofday(&end, NULL);
    ifdef printf("**** ACQUIRE SUCCEEDED. Latency:%d ms\n",(end.tv_usec - start.tv_usec)/1000);
#endif
    break;
  case wcc_protocol::RPCERR:
  case wcc_protocol::IOERR:

    wfs_log(WFS_DBG_2,"wfs_rsm_cclient::acquire slice:%016"PRIx64" not acquired. returned code is:%d\n",sliceid,r);

#ifdef PLAB_TEST
    printf("**** ACQUIRE FAILED. Latency:%d ms\n",(end.tv_usec - start.tv_usec)/1000);
#endif
    break;
  case wcc_protocol::RETRY:
    // it could be that we got a retry, but an early call by rsm succeeded,
    // so we really do own it
    if( ret.cur_state == state ) {
      r = wcc_protocol::OK;
      wfs_log(WFS_DBG_2,"wfs_rsm_cclient::acquire slice:%016"PRIx64" acquired successfully (via retry) for %d seconds.\n", sliceid,ret.tryagain);
    } else {
      wfs_log(WFS_DBG_2,"wfs_rsm_cclient::acquire slice:%016"PRIx64" is already locked for the next %d seconds. Slice State:%s\n",sliceid,ret.tryagain,ret.cur_state.c_str());
    }
    break;
  default:

    wfs_log(WFS_DBG_1,"wfs_rsm_cclient::acquire slice:%016"PRIx64" unknown return code :%d\n",sliceid,r);

    r = wcc_protocol::IOERR;
    break;
  }
  if(r == wcc_protocol::OK){
    assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
    hold_lock(sliceid,psa->get_primary(),state,period,
	      (unsigned)((period > 1/RENEWAL_MARGIN)?(period - RENEWAL_MARGIN*period):(period)),
	      renew);
    assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
  }
  return r;
}

/*
 * Atomic operations: A lock is either at the held_locks or at the release_set but not in both at the same time.
 * a lock_date->is_renewable is set iff  it is in the priority queue.
 * Assume renew_slices is held
 */
void
wfs_rsm_cclient::hold_lock( wfs_slice_id sliceid, host primary, 
			    std::string state, unsigned int period, 
			    unsigned int expire_from_now,
			    bool renew ) {

  lockedmap::iterator iter;
  lock_data * plock = NULL;
  // Put the lock in the held_set
  wfs_log(WFS_DBG_3, "hold_lock: locking %016"PRIx64"\n", sliceid );
  if((iter = held_locks.find(sliceid)) == held_locks.end() ){
    plock = new lock_data(sliceid,primary,state,period,time(NULL));
    held_locks[sliceid] = plock;
  } else {
    plock = iter->second;
    plock->last_renewed = time(NULL);
  }
  if(renew && !plock->renewable){
    // Record this slice and mark it for renewal
    unsigned int now = time(NULL);
    unsigned int next =  now + expire_from_now;
    
    plock->renewable = true;
    renewal_ent ent(plock,next);
    
    wfs_log(WFS_DBG_3, "Setting next slice renewal to %d\n", next );
    
    pq.push(ent);
    pthread_cond_signal(&schedule_change);
  }
  // Make sure it is not in the release set.
  release_set.erase(sliceid);
}

int
wfs_rsm_cclient::release(slice_assignment * psa)
{
  wcc_protocol::ret ret;
  int r;
  lockedmap::iterator iter;
  const std::string state = psa->to_canonical_string();

  // Release at the server
  r = rsmc->call(wcc_protocol::release,psa->get_primary(),psa->slice,state, ret);

  if(r==wcc_protocol::OK){
    assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
    // Remove from held lock
    if((iter = held_locks.find(psa->slice.val())) != held_locks.end() ){
      // If not renewable, safe to delete - else, Will be deleted by the renewal thread
      if(iter->second->renewable){
        // If in PQ, make sure that this slice will not be renewed.
        release_set.insert(psa->slice);
      } else {
        delete iter->second;
      }
      iter->second = NULL;
      held_locks.erase(iter);
    }
    assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
    return 0;
  }
  return r;
}
//Note: If a slice changes its state, then the server must have incremented its viewstamp, So we are
// assured that this slice will be returned in sl and checked against the current state on this client.
int wfs_rsm_cclient::fetch_slices(const viewstamp clntvs ,std::vector<slice_lock> &sl, viewstamp& srvvs)
{
  wcc_protocol::fetch_ret ret;
  int r;
  wfs_node node;
  lockedmap::iterator iter;
  lock_data * ld;
  rpcc *client = rsmc->get_rpcc(node);

  assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
  if( client == NULL ) {
    r = rpc_const::timeout_failure;
  } else {
    r = client->call(wcc_protocol::fetch_slices,clntvs,ret,rpcc::to(5000),
		     "host");
    rsmc->return_rpcc(node);
  }
  if(r == wcc_protocol::OK) {

    wfs_log(WFS_DBG_3,"wfs_rsm_cclient::fetch_slices: Got the following Slices:\n");

    for( uint i = 0; i < ret.slices.size(); i++ ) {

      wfs_log(WFS_DBG_3,"\t%s\n",ret.slices[i].toString().c_str());

      // For Each slice returned, make sure that if we have it in held_locks, it has the
      // same state as on the server, so that renewals succeed. Don't care if the locks expired.
      if((iter = held_locks.find(ret.slices[i].name)) != held_locks.end() ){
        ld = iter->second;
        if(ld->lock_state != ret.slices[i].state){
          wfs_log(WFS_DBG_3,"wfs_rsm_cclient::fetch_slices: Updating slice:%016"PRIx64" from stale state:%s \nt"
                "to new state:%s\n",ld->slice.val(),ld->lock_state.c_str(),ret.slices[i].state.c_str());

          ld->lock_state = ret.slices[i].state;
        }
      } else {
	
	// make sure that we know that we hold locks that the server
	// wants us to hold
	if( ret.slices[i].primary.addr != 0 &&
	    ret.slices[i].primary == this_host && 
	    ret.slices[i].avail == slice_lock::OWNED ) {
	  hold_lock(ret.slices[i].name, this_host, ret.slices[i].state,
		    ret.slices[i].lease_period, 
		    1 /** renew immediately */, 
		    true);
	}

      }

    }
    sl = ret.slices;
    srvvs = ret.srvvs;

    // auth stuff -- hosts
    if( ret.auth_host_keys != "" ) {
      auth_update_hosts_file( ret.auth_host_keys );
    }

    // auth stuff -- users
    if( ret.auth_user_keys.size() > 0 ) {
      auth_update_users_dir( ret.auth_user_keys );
    }

  } else {
    wfs_log(WFS_DBG_1,"wfs_rsm_cclient::get_all_slices: Call to fetch_slices failed. Error Code:%d\n",r);
  }
  assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
  return r;
}

int
wfs_rsm_cclient::renew_slice(slice_assignment * pNewSa)
{
  int ret;
  lockedmap::const_iterator iter;
  std::string newstate;
  lock_data * pld;
  int r;
  assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
  if((iter = held_locks.find(pNewSa->slice.val())) == held_locks.end() ){

    wfs_log(WFS_DBG_2, "Not holding lock for slice %d, can't renew\n",
	      pNewSa->slice.val() );

    ret = wcc_protocol::NOENT;
  } else {
    pld = iter->second;
    newstate = pNewSa->to_canonical_string();
    r = rsmc->call(wcc_protocol::renew,pld->primary,pld->slice,pld->lock_state,
		   newstate,pld->period,ret);
    if(r == wcc_protocol::OK){
      ret = r;
      pld->last_renewed = time(NULL);
      pld->lock_state = newstate;
      // We don't touch the priority queue so this lock might be renewed ahead of time for the next time. OK.

    } else {

      wfs_log(WFS_DBG_1, "Bad RPC lock for slice %d, can't renew (%d)\n",
		  pNewSa->slice.val(), r );

      ret = wcc_protocol::RPCERR;
    }
  }
  assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
  return ret;
}

void
wfs_rsm_cclient::slice_renewer()
{
  //struct timeval start,end;
  wcc_protocol::ret ret,ret2;
  lockedmap::iterator iter;
  int r,r2;
  struct timespec ts;
  ts.tv_nsec = 0;
  ts.tv_sec = time(NULL) + 10000;	// Start by sleeping a long time until first condition

  // don't renew slices for a non-server
  if(this_host.addr == 0) {
    return;
  }

  assert(pthread_mutex_lock(&renew_slices_mutex) == 0);

  while(!_shutting_down) {
    pthread_cond_timedwait(&schedule_change,&renew_slices_mutex,&ts);
    // We are under lock
    wfs_log(WFS_DBG_3,"wfs_rsm_cclient::slice_renewer: Woke up\n");

    if(pq.empty() || _shutting_down){
      continue; // spurious wake up
    }
    std::vector<reacquire_data> reacquire;
    unsigned int now = time(NULL);
    // See if there are acquire calls to be executed
    while(!pq.empty() && (pq.top().next_renewal <= now)){
      renewal_ent ent = pq.top() ;
      pq.pop();
      // Was this lock released? If so, don't renew.
      if(release_set.find(ent.ldata->slice.val()) != release_set.end()){
        wfs_log(WFS_DBG_3,"wfs_rsm_cclient::slice_renewer: removed slice:%016"PRIx64"\n",ent.ldata->slice.val());
        // Make sure it is not in held_locks. (Should not happen).
        lockedmap::iterator iter;
        if((iter = held_locks.find(ent.ldata->slice.val())) != held_locks.end() ){
          held_locks.erase(iter);
        }
        delete ent.ldata;
        continue;
      }
      // call renew
      wfs_slice_id sliceid = ent.ldata->slice.val();
#ifdef PLAB_TEST
      // TODO:tests only
      gettimeofday(&start, NULL);
#endif
      r = rsmc->call(wcc_protocol::renew,ent.ldata->primary,ent.ldata->slice,ent.ldata->lock_state,
		     ent.ldata->lock_state,ent.ldata->period,ret);
#ifdef PLAB_TEST
      // TODO: tests only
      gettimeofday(&end, NULL);
#endif

      wfs_log(WFS_DBG_3,"wfs_rsm_cclient::slice_renewer: for slice:%016"PRIx64" return code:%d.\n",sliceid,r);

      switch(r){
      case wcc_protocol::OK:

        wfs_log(WFS_DBG_3,"wfs_rsm_cclient::slice_renewer: renewed slice:%016"PRIx64" lock for %d seconds successfully.\n", sliceid,ent.ldata->period);

#ifdef PLAB_TEST
        printf("**** RENEW SUCCEEDED. Latency:%d ms\n",(end.tv_usec - start.tv_usec)/1000);
#endif
        // Should already have the lock in held locks
        assert((iter = held_locks.find(sliceid)) != held_locks.end() );
        iter->second->last_renewed = time(NULL);

        break;
      case wcc_protocol::RPCERR:
      case wcc_protocol::IOERR:
        wfs_log(WFS_DBG_1,"wfs_rsm_cclient::slice_renewer: slice:%016"PRIx64" not acquired. returned code is:%d\n",sliceid,r);

#ifdef PLAB_TEST
        printf("**** RENEW Failed. Latency:%d ms\n",(end.tv_usec - start.tv_usec)/1000);
#endif
        break;
      case wcc_protocol::NOENT:
        // If the server does not know about this slice, try and acquire it first.

        wfs_log(WFS_DBG_1,"wfs_rsm_cclient::slice_renewer: slice:%016"PRIx64" is unknown to the server. "
              "Perhaps we are recovering from a failure.\nWill try and acquire this slice.\n",sliceid);

        r2 = rsmc->call(wcc_protocol::acquire,ent.ldata->primary,ent.ldata->slice,
			ent.ldata->lock_state,ent.ldata->period, ret2);
        if(r2 == wcc_protocol::OK){
          // Should already have the lock in held locks
          assert((iter = held_locks.find(sliceid)) != held_locks.end() );
          iter->second->last_renewed = time(NULL);
        }
        break;
      case wcc_protocol::NOOWN:

	// if the lock is free, try to acquire!
	if( slice_assignment(ret.cur_state).replicas.size() == 0 ) {

	  wfs_log(WFS_DBG_1,"wfs_rsm_cclient::slice_renewer: The server thinks that this host does not own slice:%016"PRIx64". "
		   "It's free though, so trying to acquire.\n",sliceid);

	  reacquire_data rd;
	  rd.state = ent.ldata->lock_state;
	  rd.period = ent.ldata->period;
	  reacquire.push_back(rd);
	} else {
	  // If we don't own a slice anymore, give up.
	  wfs_log(WFS_DBG_1,"wfs_rsm_cclient::slice_renewer: The server thinks that this host does not own slice:%016"PRIx64"."
		   "Perhaps we are recovering from a failure.\nWill not try to renew this slice anymore.\n",sliceid);

	  if((iter = held_locks.find(sliceid)) != held_locks.end()){
	    held_locks.erase(iter);
	  }
	  release_set.insert(sliceid);
	}
	break;
      case wcc_protocol::STALE:

        wfs_log(WFS_DBG_1,"wfs_rsm_cclient::slice_renewer: slice:%016"PRIx64"'s state is stale. Perhaps we are recovering from a failure.\n",sliceid);

        break;
      default:

        wfs_log(WFS_DBG_1,"wfs_rsm_cclient::slice_renewer: slice:%016"PRIx64" unknown return code :%d\n",sliceid,r);

        break;
      }
      now = time(NULL); // Call might have taken time
      // Compute next call time
      ent.next_renewal = now + (unsigned)((ent.ldata->period > 1/RENEWAL_MARGIN)?(ent.ldata->period - RENEWAL_MARGIN*ent.ldata->period):(ent.ldata->period));
      // Back to the queue;
      pq.push(ent);
    }

    // now reacquire
    if(reacquire.size() > 0) {
      assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
      for(unsigned i = 0; i < reacquire.size(); i++) {
	slice_assignment oldSa(reacquire[i].state);
	// make sure we're still the primary in our own minds . . .
	if(oldSa.replicas.size() == 0) {
	  oldSa.replicas.push_back(this_host);
	} else if(oldSa.get_primary() != this_host) {
	  oldSa.replicas.push_front(this_host);
	  if(oldSa.replicas.size() > oldSa.policy.rep_num) {
	    oldSa.replicas.pop_back();
	  }
	}
	acquire_slice(&oldSa, reacquire[i].period, true);
      }
      assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
    }

    if(pq.empty()) {
      // Go to sleep for a long time
      ts.tv_sec = time(NULL) + 10000;
    } else {
      ts.tv_sec = pq.top().next_renewal;
    }

    wfs_log(WFS_DBG_3, "New slice renewal time %d\n", (int)ts.tv_sec );

    usleep(10000);

  }
  assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);

}

bool wfs_rsm_cclient::is_slice_locked(wfs_slice_id sid) const
{
  bool ret;
  lockedmap::const_iterator iter;
  assert(pthread_mutex_lock(&renew_slices_mutex) == 0);
  if((iter = held_locks.find(sid)) == held_locks.end() ){
    ret = false;
  } else {
    ret = (time(NULL) - iter->second->last_renewed) < iter->second->period;
  }
  assert(pthread_mutex_unlock(&renew_slices_mutex) == 0);
  return ret;
}




