/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef WFS_RSM_CCLIENT_H_
#define WFS_RSM_CCLIENT_H_

#include <string>
#include <list>
#include <assert.h>
#include <queue>
#include <vector>
#include <set>
#include <wfs_client_manager.h>
#include "wcc_protocol.h"
#include "rpc.h"
#include "rsm_client.h"
#include "replication_policy.h"
#include "wfs_rsm_cserver.h"

/*
 * wfs_rsm_cclient is a simple client interface to the rsm configuration service.
 * The higher logic is handled by users such as node_manager_rsm.
 */
class wfs_rsm_cclient {
  private:

  struct reacquire_data {
    std::string state;
    int period;
  };

    /*
     * Need to keep track of which locks are locked and when were they granted.
     * When an acquire succeeds, an entry is created and inserted. On a release an entry is deleted.
     * On a successful renewal an entry is updated.
     * Use renew_slices_mutex to protect this operation
     */
    struct lock_data {
      wfs_slice slice;
      host primary;
      std::string lock_state;
      unsigned int period;
      unsigned int last_renewed;
      bool renewable;

      lock_data(wfs_slice_id _sl,const host prim,const std::string state, unsigned int per ,unsigned int lr = 0)
      : slice(_sl), primary(prim), lock_state(state),period(per),last_renewed(lr), renewable(false){}
    };

    struct renewal_ent {
      renewal_ent(lock_data * ld, unsigned int next = 0)
      : ldata(ld), next_renewal(next) {}
      ~renewal_ent() {};
      lock_data * ldata;
      unsigned int next_renewal;    // Time from Epoch.
      // If next_renewal is 0 (not yet scheduled), always return the other entry as the smaller (i.e. any future time happens before "unshceduled").
      bool operator<(const renewal_ent& ent) const { return (next_renewal)?(next_renewal > ent.next_renewal):(ent.next_renewal); }
    };

    rsm_client *rsmc;
    host this_host;
    static const double RENEWAL_MARGIN = .5;
    typedef std::priority_queue<renewal_ent,std::vector<renewal_ent> > slicepq;
    slicepq pq;
    mutable pthread_mutex_t renew_slices_mutex;
    bool _shutting_down;

    /*
     * STL's priority queue does not support remove or iteration. To prevent locks that have been released from being periodically
     * renewed, we keep a set of released slices and whenever a slice is scheduled to be renewed it is checked against that set.
     * If it is in the set we just pop that entry from the priority queue and do nothing else.
     * All set operation must be done under cur_slices_mutex locked.
     */
    std::set<wfs_slice_id> release_set;
    pthread_t th_renewer;
    pthread_cond_t schedule_change;

    typedef std::map<wfs_slice_id,lock_data *> lockedmap;
    lockedmap held_locks;


    void slice_renewer();
    void hold_lock( wfs_slice_id sliceid, host primary, 
		    std::string state, unsigned int period, 
		    unsigned int expire_from_now, bool renew );
  public:

    wfs_rsm_cclient(std::string xdst, host me, 
		    wfs_client_manager *clients);
    ~wfs_rsm_cclient();
    /*
     * acquire_slice
     * 	acquires a slice and renews it before expiration if requested.
     */
    int acquire_slice(slice_assignment * psa,const unsigned period, bool renew);
    /*
     * Releases a slice and stops it from being renewed.
     */
    int release(slice_assignment * psa);

    /*
     * The clients current viewstamp is passed in current and the server's current viewstamp is returned there as well.
     */
    int fetch_slices(const viewstamp clntvs ,std::vector<slice_lock> &sl, viewstamp& srvvs);

    /*
     * Is the slice locked. Is the slice in the locked set and is (now-last_renewal) < period
     */
    bool is_slice_locked(wfs_slice_id sid) const;

    int renew_slice(slice_assignment * pNewSa);

};


#endif // WFS_RSM_CCLIENT_H_
