/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// Fault tolerant Configuration service. Uses a RSM and paxos.
#include "paxos.h"
#include "rsm.h"
#include <arpa/inet.h>
#include <getopt.h>
#include <assert.h>
#include <netdb.h>
#include <rpc.h>
#include <vivaldi.h>
#include <wfs_misc.h>
#include <assert.h>
#include <sys/stat.h>
#include <method_thread.h>

#include "host_marshall.h"
#include "wfs_rsm_cserver.h"

//static vivaldi* viv;

using namespace std;

wfs_rsm_cserver::wfs_rsm_cserver(class rsm *_rsm, bool secure) :
  rsm (_rsm), _secure(secure), _auth_hosts_modified_viewstamp(0,0),
  _auth_hosts_modified_time(0), _auth_users_modified_viewstamp(0,0),
  _auth_users_modified_time(0) {

  assert( pthread_mutex_init( &_slices_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &_hosts_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &_auth_lock, NULL ) == 0 );

  // start the background auth monitoring thread
  if( _secure ) {
    method_thread( this, true, &wfs_rsm_cserver::monitor_auth_files );
  }

}

wfs_rsm_cserver::~wfs_rsm_cserver() {
  pthread_mutex_destroy( &_slices_lock );
  pthread_mutex_destroy( &_hosts_lock );
}

int wfs_rsm_cserver::ping( host h, host &dummy ) {
  // Not used
  return wcc_protocol::RPCERR;
}

int
wfs_rsm_cserver::acquire(const host requester,const wfs_slice slice,
    const std::string newstate, const unsigned period, wcc_protocol::ret& ret)
{
  slice_lock * slcp;
  ostringstream oshost;
  oshost << requester;

  pthread_mutex_lock(&_slices_lock);
  if (locks.find(slice) == locks.end()) {
    slcp = new slice_lock(slice);
    locks[slice] = slcp;
  } else {
    slcp = locks[slice];
  }
  slcp->avail = slice_lock::OWNED;
  slcp->lease_period = period;
  slcp->state = newstate;
  // TODO, see if we know requester, and if so get a pointer from that set.
  slcp->primary = host(requester);
  // Only the master sets the granting time.
  bool master = rsm->amimaster_mx();

  if(master){
    unsigned int now = time(NULL);
    slcp->granted_on = now;
    slcp->last_modified = rsm->get_latest_viewstamp();

    wfs_log(WFS_DBG_3,"acquire(on master): granting lock to %s. Slice:%016"PRIx64". Period:%d Now:%d\n"
          ,oshost.str().c_str(),slice.val(),period,now);

  } else {

    wfs_log(WFS_DBG_3,"acquire(on slave): granting lock to %s. Slice:%016"PRIx64". Period:%d \n"
          ,oshost.str().c_str(),slice.val(),period);

    // if any of our locks had a granted_on, that means we used to be master
    // but no longer are.  in that case we have to clear all the granted_ons
    // so that if we become master again in the future, the locks don't all
    // automatically expire
    if( slcp->granted_on != 0 ) {
      clear_all_locks();
    }
  }
  pthread_mutex_unlock(&_slices_lock);
  // Leave Old State empty and set period to the one that was requested.
  ret.tryagain = period;
  ret.cur_state = "";
  return wcc_protocol::OK;
}

int
wfs_rsm_cserver::became_master(std::string old_master, int& dummy) {
  pthread_mutex_lock(&_slices_lock);
  renew_all_locks();
  pthread_mutex_unlock(&_slices_lock);
}

int
wfs_rsm_cserver::lost_master(std::string new_master, int& dummy) {
  pthread_mutex_lock(&_slices_lock);
  clear_all_locks();
  pthread_mutex_unlock(&_slices_lock);
}

void
wfs_rsm_cserver::clear_all_locks() {

  wfs_log(WFS_DBG_2,"Lost being a master!  Clearing all locks.\n" );

  for (LOCKMAP::iterator iter_lock = locks.begin(); iter_lock != locks.end();
       iter_lock++) {
    slice_lock *slock = iter_lock->second;
    slock->granted_on = 0;
  }

}

/**
 * Called whenever the node does an operation that only a master should be
 * doing, and finds that it has a granted lock with no granted_on field
 * set -- this means that the node just became the master, and must regrant
 * a full lease period to all known locks, since the renew operation is not
 * replicated
 */
void
wfs_rsm_cserver::renew_all_locks() {


  wfs_log(WFS_DBG_2,"Became master!  Renewing all locks.\n" );


  unsigned int now = time(NULL);

  for (LOCKMAP::iterator iter_lock = locks.begin(); iter_lock != locks.end();
       iter_lock++) {
    slice_lock *slock = iter_lock->second;
    if(slock->avail == slice_lock::OWNED && slock->granted_on == 0){
      slock->granted_on = now;
    }
  }

}

rsm::dec_ret
wfs_rsm_cserver::can_acquire(const host requester,const wfs_slice slice,const std::string newstate,
    const unsigned period, wcc_protocol::ret& ret, int& wccret)
{

  slice_lock * slcp;
  int tte;
  ostringstream oshost;
  oshost << requester;

  pthread_mutex_lock(&_slices_lock);

  if (locks.find(slice) == locks.end()) {
    // This means this is a new slice and can be acquired.

    wfs_log(WFS_DBG_3,"can_acquire: %s: requested to lock new slice:%016"PRIx64". Approve!\n"
          ,oshost.str().c_str(),slice.val());

    pthread_mutex_unlock(&_slices_lock);
    return rsm::YES_REP;
  }
  // Existing lock
  slcp = locks[slice];
  /*
   * This code is only executed only by the master so if for an existing locked lock granted_on is 0, then
   * this is the first time the master sees this lock as a master and must renew it for the entire period
   * to avoid clock synchronization problems. We do this by setting granted_on to now.
   * TODO: Might as well run on all other OWNED locks and set their grant time for now.
   */
  if(slcp->avail == slice_lock::OWNED && slcp->granted_on == 0){
    renew_all_locks();
  }
  // Has not expired?
  if((tte = slcp->ttexpire()) > 0) {
    assert(tte < MAX_LEASE_PERIOD);
    ret.tryagain = tte;
    ret.cur_state = slcp->state;
    wccret = wcc_protocol::RETRY;

    wfs_log(WFS_DBG_2,"can_acquire: %s: requested to lock slice:%016"PRIx64" with %d remaining seconds on lease. Reject!\n"
          ,oshost.str().c_str(),slice.val(),tte);

    pthread_mutex_unlock(&_slices_lock);
    return rsm::NO;
  }
  // Expired Existing locks. Can be acquired
  wfs_log(WFS_DBG_3, "can_acquire: %s: requested to lock unlocked slice:%016"PRIx64". Approve!\n"
        ,oshost.str().c_str(),slice.val());

  pthread_mutex_unlock(&_slices_lock);
  return rsm::YES_REP;
}

// Call when lock mutex is locked
rsm::dec_ret
wfs_rsm_cserver::ownership_and_state_test(const host requester,const wfs_slice & slice,
    const std::string& oldstate ,const std::string& newstate,const unsigned period,
    wcc_protocol::ret& ret, int& wccret, std::string debug_call, bool renewal_check)
{
  slice_lock * slcp;

  if (locks.find(slice) == locks.end()) {

    wfs_log(WFS_DBG_2,"%s: %s: requested to release/renew non-existent slice:%016"PRIx64". Reject!\n"
          ,debug_call.c_str(),requester.str().c_str(),slice.val());

    wccret = wcc_protocol::NOENT;
    return rsm::NO;
  }
  // Existing locks
  slcp = locks[slice];
  // Ownership test
  if(slcp->primary.addr == 0 || slcp->primary != requester ||
     slcp->ttexpire() <= 0){

    wfs_log(WFS_DBG_2,"%s: %s: requested to release/renew slice:%016"PRIx64" which is does not own. Reject!\n"
          ,debug_call.c_str(),requester.str().c_str(),slice.val());

    wccret = wcc_protocol::NOOWN;
    if(slcp->ttexpire() <= 0){
      slcp->state = "";
      slcp->avail = slice_lock::FREE;
    }
    // The client can use the state to find out who owns the lock
    ret.cur_state = slcp->state;
    return rsm::NO;
  }
  // Freshness test
  if(slcp->state != oldstate){

    wfs_log(WFS_DBG_3,"%s: %s: requested to release/renew slice:%016"PRIx64" but state is stale. Reject!\n"
          ,debug_call.c_str(),requester.str().c_str(),slice.val());

    wccret = wcc_protocol::STALE;
    ret.cur_state = slcp->state;
    return rsm::NO;

  }

  wfs_log(WFS_DBG_3,"%s: %s's request to release/renew slice:%016"PRIx64" is valid. Approve!\n"
        ,debug_call.c_str(),requester.str().c_str(),slice.val());

  if(renewal_check && (oldstate == newstate) && (period == slcp->lease_period)){

    wfs_log(WFS_DBG_3,"%s: No need to replicate this operation! (Requester:%s , Slice:%016"PRIx64")\n"
          ,debug_call.c_str(),requester.str().c_str(),slice.val());

    return rsm::YES_NOREP;
  } else {
    return rsm::YES_REP;
  }
}

/*
 * NOTE:This procedure is only replicated if the lock state or the renewal period change.
 * This is possible since slaves don't keep track of actual grant time.
 */
int
wfs_rsm_cserver::renew(const host requester,const wfs_slice slice,const std::string oldstate,const std::string newstate,
    const unsigned period, wcc_protocol::ret& ret)
{
  slice_lock * slcp;
  ostringstream oshost;
  oshost << requester;
  bool replicate;

  pthread_mutex_lock(&_slices_lock);

  assert(locks.find(slice) != locks.end());

  wfs_log(WFS_DBG_3,"renew: renew slice %016"PRIx64": owned by:%s for %d more seconds.\n"
        ,slice.val(),oshost.str().c_str(),period);

  slcp = locks[slice];

  replicate = ((oldstate != newstate) || (period != slcp->lease_period));

  slcp->avail = slice_lock::OWNED;
  slcp->lease_period = period;
  slcp->state = newstate;

  // No need to change primary. Only an owner can renew its own lock.

  // Only the master sets the granting time.
  bool master = rsm->amimaster_mx();
  if(master){
    unsigned int now = time(NULL);
    slcp->granted_on = now;
    if(replicate){
      slcp->last_modified = rsm->get_latest_viewstamp();
    }
    wfs_log(WFS_DBG_3,"renew(on master): renew slice %016"PRIx64": owned by:%s for %d more seconds. Now:%d\n"
          ,slice.val(),oshost.str().c_str(),period,now);

  }
  pthread_mutex_unlock(&_slices_lock);
  return wcc_protocol::OK;
}


rsm::dec_ret
wfs_rsm_cserver::can_renew(const host requester,const wfs_slice slice,const std::string oldstate,const std::string newstate,
    const unsigned period, wcc_protocol::ret& ret, int& wccret)
{
  rsm::dec_ret answer;
  pthread_mutex_lock(&_slices_lock);
  answer = ownership_and_state_test(requester,slice,oldstate,newstate,period,ret,wccret,"can_renew",true);
  pthread_mutex_unlock(&_slices_lock);
  return answer;
}

int
wfs_rsm_cserver::release(const host requester,const wfs_slice slice,const std::string oldstate, wcc_protocol::ret& ret)
{
  slice_lock * slcp;
  ostringstream oshost;
  oshost << requester;

  pthread_mutex_lock(&_slices_lock);
  assert(locks.find(slice) != locks.end());

  wfs_log(WFS_DBG_3,"release: releasing slice %016"PRIx64": previously owned by:%s\n"
        ,slice.val(),oshost.str().c_str());

  slcp = locks[slice];
  slcp->avail = slice_lock::FREE;
  slcp->lease_period = 0;
  slcp->state = "";
  slcp->granted_on = 0;
  // TODO: Just decrement reference count and delete from active primaries on 0
  slcp->primary = host();
  // Only the master should update the operation viewstamp.
  if(rsm->amimaster_mx()){
    slcp->last_modified = rsm->get_latest_viewstamp();
  }

  pthread_mutex_unlock(&_slices_lock);
  return wcc_protocol::OK;
}

rsm::dec_ret
wfs_rsm_cserver::can_release(const host requester,const wfs_slice slice,const std::string oldstate,
    wcc_protocol::ret& ret, int& wccret)
{
  rsm::dec_ret answer;
  int dummy = 20;
  pthread_mutex_lock(&_slices_lock);
  answer = ownership_and_state_test(requester,slice,oldstate,oldstate,dummy,ret,wccret,"can_release");
  pthread_mutex_unlock(&_slices_lock);
  return answer;
}

int wfs_rsm_cserver::fetch_slices(const viewstamp clntvs ,wcc_protocol::fetch_ret& ret)
{

  LOCKMAP::iterator iter_lock;
  slice_lock *slock;
  viewstamp useclntvs;
  int j = 0;

  pthread_mutex_lock(&_slices_lock);

  viewstamp latest = rsm->get_latest_viewstamp();
  if(clntvs > latest){
    // Should not happen. Give client all updates.
    useclntvs.vid = 0; useclntvs.seqno = 0;

    wfs_log(WFS_DBG_1,"wfs_rsm_cserver::fetch_slices: Client reported viewstamp in the future. Using 0 viewstamp.\n");

  } else {
    useclntvs = clntvs;
  }

  viewstamp max_seen = useclntvs;

  for (iter_lock = locks.begin(); iter_lock != locks.end(); iter_lock++) {
    bool expired = false;
    slock = iter_lock->second;

    // if we get fetch_slices, we must be the master, so make sure the
    // lock state makes sense
    if(slock->avail == slice_lock::OWNED && slock->granted_on == 0){
      renew_all_locks();
    }

    wfs_log(WFS_DBG_4,"fetch_slices: looking at lock %016"PRIx64"\n", slock->name );

    // If this lock has a viewstamp of zero, we might just became a master - stamp the lock with the latest viewstamp.
    if(slock->last_modified.vid == 0 && slock->last_modified.seqno == 0){
      slock->last_modified = latest;

      wfs_log(WFS_DBG_3,"wfs_rsm_cserver::fetch_slices: Updating viewstamp for slice %016"PRIx64"\n"
              ,slock->name);

    }
    // We have to send expired lock regardless of viewstamp because expiration operation is not stamped.
    if((slock->granted_on != 0) && (slock->ttexpire() <= 0)) {
      expired = true;
    }
    // Since the both no-operation-yet and the first operation get (1,0), need to send all when client sends (1,0)
    if(expired || slock->last_modified > useclntvs || (useclntvs.vid == 1 && useclntvs.seqno == 0)){
      j++;
      slice_lock sl(*slock);
      /*
       * Check to see if this lock has expired, If so, tell client that it is FREE.
       * Note that since this operation is not replicated, the state of expired locks remains LOCKED
       * In essence the client may see a different state than what the server has.
       * The only scenario where the client will see an inconsistent state is where a new host
       * becomes the master and renews all expired locks for the lease period, in that period
       * a client may think these locks are free while in fact they will be locked. This is not
       * a problem.
       */
      // Has  expired?
      if(expired) {
        sl.avail = slice_lock::FREE;
      }
      if(slock->last_modified > max_seen) {
	max_seen = slock->last_modified;
      }
      ret.slices.push_back(sl);
    }
  }
  pthread_mutex_unlock(&_slices_lock);

  // does the client have the latest keys?
  if( _secure ) {

    if( _auth_hosts_modified_viewstamp > useclntvs || 
	(useclntvs.vid == 0 && useclntvs.seqno == 0) ) {
      ret.auth_host_keys = auth_hosts_file_string();
      if(_auth_hosts_modified_viewstamp > max_seen) {
	max_seen = _auth_hosts_modified_viewstamp;
      }
    } else {
      ret.auth_host_keys = "";
    }

    if( _auth_users_modified_viewstamp > useclntvs || 
	(useclntvs.vid == 0 && useclntvs.seqno == 0) ) {
      ret.auth_user_keys = auth_user_keys_map();
      if(_auth_users_modified_viewstamp > max_seen) {
	max_seen = _auth_users_modified_viewstamp;
      }
    }

    ret.srvvs = max_seen;
    assert(max_seen <= latest);


  } else {
    ret.auth_host_keys = "";
  }

  wfs_log(WFS_DBG_3,"wfs_rsm_cserver::fetch_slices: Client's viewstamp (%d,%d). Sent %d Entries and returned server's last-op viewstamp (%d,%d).\n"
        ,clntvs.vid,clntvs.seqno,j,latest.vid,latest.seqno);


  return wcc_protocol::OK;
}

std::string
wfs_rsm_cserver::marshal_state()
{
  ostringstream ost;
  string r;

  pthread_mutex_lock(&_slices_lock);

  wfs_log(WFS_DBG_3,"wfs_rsm_cserver::marshal_state %zu locks\n", locks.size());

  // hosts file
  std::string hosts_file = auth_hosts_file_string();
  ost << hosts_file.size();
  ost << hosts_file;

  // users dir
  std::map< std::string, std::string > users_map = auth_user_keys_map();
  marshall m;
  m << users_map;
  std::string mstr = m.str();
  ost << mstr.size();
  ost << mstr;

  ost << locks.size();
  ost << " ";

  LOCKMAP::const_iterator iter_lock;
  for (iter_lock = locks.begin(); iter_lock != locks.end(); iter_lock++) {
    slice_lock *slock = locks[iter_lock->first];
    assert(slock);

    wfs_log(WFS_DBG_3,"wfs_rsm_cserver::marshal_state slice %s\n",slock->toString().c_str());

    // Write lock
    //slock->operator <<(ost);
    ost << *slock;
  }
  // TODO: when/if added, write hosts state.
  pthread_mutex_unlock(&_slices_lock);
  r = ost.str();
  wfs_log(WFS_DBG_3,"wfs_rsm_cserver::marshal_state done. state length = %zu\n", r.length());
  return r;
}

void
wfs_rsm_cserver::unmarshal_state(std::string state)
{
  int nlock;
  slice_lock *slock;
  istringstream ist(state);

  int hosts_file_size;
  ist >> hosts_file_size;
  char *c = new char[hosts_file_size+1];
  ist.get( c, hosts_file_size+1, '\0' );
  auth_update_hosts_file( std::string(c) );
  _auth_hosts_modified_viewstamp = rsm->get_latest_viewstamp();
  _auth_hosts_modified_time = time(NULL);
  delete [] c;

  int ustr_size;
  ist >> ustr_size;
  char *c2 = new char[ustr_size+1];
  ist.get( c2, ustr_size+1, '\'' );
  unmarshall u( c2, ustr_size );
  std::map< std::string, std::string > users_map;
  umap( u, users_map );
  auth_update_users_dir( users_map );
  u.take_buf(); // no double delete
  _auth_users_modified_viewstamp = rsm->get_latest_viewstamp();
  _auth_users_modified_time = time(NULL);
  delete [] c2;

  ist >> nlock;

  wfs_log(WFS_DBG_3,"wfs_rsm_cserver::unmarshal_state %zu bytes. containing %d locks\n", state.length(), nlock);

  pthread_mutex_lock(&_slices_lock);
  for (int j = 0; j < nlock; j++) {
    slock = new slice_lock();
    ist >> *slock;
    locks[slock->name] = slock;

    wfs_log(WFS_DBG_3,"wfs_rsm_cserver::unmarshal_state slice %s\n",slock->toString().c_str());

  }
  pthread_mutex_unlock(&_slices_lock);

  wfs_log(WFS_DBG_3,"wfs_rsm_cserver::unmarshal_state done\n");

  return;
}

void
wfs_rsm_cserver::monitor_auth_files() {

  while(1) {

    sleep(10);

    if( !rsm->amimaster() ) {
      continue;
    }

    assert( pthread_mutex_lock( &_auth_lock ) == 0 );
    _auth_code = random();

    const char *authorized_hosts_file = wfs_get_authorized_hosts_file();
    struct stat sb;
    if( stat( authorized_hosts_file, &sb ) == 0 ) {
      if( sb.st_mtime > _auth_hosts_modified_time ) {
	std::string hosts_file = auth_hosts_file_string();
	marshall m;
	m << hosts_file;
	m << _auth_code;
	std::string rs;
	wfs_log( WFS_DBG_3, "Found an update to the host keys (%d %d)\n", 
		 sb.st_mtime,  _auth_hosts_modified_time );
	rsm_client_protocol::status ret = 
	  rsm->client_invoke( wcc_protocol::update_auth_host_keys, m.str(),
			      rs );
      }
    }


    const char *authorized_keys_dir = wfs_get_authorized_keys_dir();
    if( stat( authorized_keys_dir, &sb ) == 0 ) {
      // TODO: should probably check individual times for all files in the dir
      if( sb.st_mtime > _auth_users_modified_time ) {
	std::map< std::string, std::string > users_map = auth_user_keys_map();
	marshall m;
	m << users_map;
	m << _auth_code;
	std::string rs;
	wfs_log( WFS_DBG_3, "Found an update to the user keys\n" );
	rsm_client_protocol::status ret = 
	  rsm->client_invoke( wcc_protocol::update_auth_user_keys, m.str(),
			      rs );
      }
    }

    assert( pthread_mutex_unlock( &_auth_lock ) == 0 );

  }

}

rsm::dec_ret 
wfs_rsm_cserver::can_update_auth_host_keys(const std::string hosts_file, 
					   int auth_code, 
					   int &dummy, int& wccret)
{
  if( auth_code == _auth_code ) {
    return rsm::YES_REP;
  } else {
    return rsm::NO;
  }
}

int 
wfs_rsm_cserver::update_auth_host_keys(const std::string hosts_file, 
				       int auth_code, int &dummy) {
  if( !rsm->amimaster_mx() ) {
    auth_update_hosts_file( hosts_file );
  }
  _auth_hosts_modified_viewstamp = rsm->get_latest_viewstamp();
  _auth_hosts_modified_time = time(NULL);
  return wcc_protocol::OK;
}


rsm::dec_ret
wfs_rsm_cserver::can_update_auth_user_keys(const std::map< std::string, std::string > users_map, 
					   int auth_code, 
					   int &dummy, int& wccret)
{
  if( auth_code == _auth_code ) {
    return rsm::YES_REP;
  } else {
    return rsm::NO;
  }
}

int
wfs_rsm_cserver::update_auth_user_keys(const std::map< std::string, std::string > users_map, 
				       int auth_code, int &dummy) {
  if( !rsm->amimaster_mx() ) {
    auth_update_users_dir( users_map );
  }
  _auth_users_modified_viewstamp = rsm->get_latest_viewstamp();
  _auth_users_modified_time = time(NULL);
  return wcc_protocol::OK;
}
