/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef WFS_RSM_CSERVER_H_
#define WFS_RSM_CSERVER_H_

#include <string>
#include <map>
#include <rpc.h>
#include <unistd.h>
#include "rsm_state_transfer.h"
#include "rsm.h"
#include "host.h"
#include "wcc_protocol.h"
#include "wfs_slice.h"
#include "updown.h"
#include "slice_lock.h"


/*
 * The configuration server in this implementation is unaware of the internals of a slice policy.
 * it is aware of its string form which is slice_state := string_representation(policy,replica_set)
 * The RSM state consists of:
 * 	the set of mappings from fid slices (locks) -> <primary,slice_state,expiration>
 * where expiration is only maintained by the current master according to its count from the epoch.
 *
 * The implicit state is
 * 	the set of live primary nodes. Which is a set of nodes which hold locks.
 */
class wfs_rsm_cserver : public rsm_state_transfer {
 public:

 private:

  static const int MAX_LEASE_PERIOD = 60*10; // 10 Minutes. Too large?

  typedef std::map<wfs_slice_id, slice_lock*> LOCKMAP;
  LOCKMAP locks;
  pthread_mutex_t _slices_lock;

  std::vector<host1> _hosts;	// All current active nodes (nodes that are known to be primaries for at least one slice)
  pthread_mutex_t _hosts_lock;
  class rsm *rsm;

  rsm::dec_ret ownership_and_state_test(const host requester,const wfs_slice& slice,const std::string& oldstate,
      const std::string& newstate,const unsigned period,wcc_protocol::ret& ret,
      int& wccret, std::string debug_call, bool renewal_check = false);
  void clear_all_locks();
  void renew_all_locks();

  //friend std::ostream & operator<<(std::ostream &os,const wfs_rsm_cserver::slice_lock &l);
  //friend std::istream & operator>>(std::istream &is, wfs_rsm_cserver::slice_lock &l);

  // auth stuff
  bool _secure;
  pthread_mutex_t _auth_lock;
  int _auth_code;
  viewstamp _auth_hosts_modified_viewstamp;
  time_t _auth_hosts_modified_time;
  viewstamp _auth_users_modified_viewstamp;
  time_t _auth_users_modified_time;
  std::string hosts_file_string();
  std::map<std::string, std::string> user_keys_map();

 public:
  wfs_rsm_cserver(class rsm *rsm = 0, bool secure = false);
  ~wfs_rsm_cserver();

  // Read-only Handlers
  int ping( host h, host & );
  int fetch( int dummy, std::vector<host> &hosts );

  int became_master(std::string old_master, int& dummy);
  int lost_master(std::string new_master, int& dummy);

  /*
   * acquire:
   * parameters:
   * 	requester: node_id of the requester (primary).
   * 	slice: name of the lock (12 msbs of the 64 bit fid)
   * 	newstate: the requested new state for this lock.
   * 	period: the requested lease period for this lock.
   * 	ret: placeholder to hint the client when to try again for an unexpired lock.
   *
   * Note: I chose to use the host struct as the unique requester id. An alternative choice would be to use
   * the wfs_node_id. The reasons for this choice: host contains the site_id in it, it already has marshalling
   * support and overloaded boolean comparators which give the same effect as using wfs_node_id which is
   * just a hash of IP:port . The price is a slightly larger payload (Y.S.)
   *
   */
  int acquire(const host requester,const wfs_slice slice,const std::string newstate,
			  const unsigned int period, wcc_protocol::ret& ret);

  /*
   * Decision function - Checks two things:
   * 	That the lock is not held under an unexpired lease.
   * 	That the requester's reported slice_state matches the slice_state on the server.
   * 	Since we treat the string as an opaque state - this is just a simple string comparison.
   * This implies that if a client want's to make changed to locks that it holds, it needs to release
   * them first and then re-acquire them with the new slice_state or just use renew.
   */
  rsm::dec_ret can_acquire(const host requester,const wfs_slice slice,const std::string newstate,
			const unsigned period, wcc_protocol::ret& ret, int& wccret);

  /*
   * renew:
   * Used by nodes that hold a lock and its up-to-date state and would like to renew it before it expired
   * parameters:
   * 	requester: node_id of the requester (primary).
   * 	slice: name of the lock (12 msbs of the 64 bit fid)
   * 	oldstate: the state associated with the lock, as the client knows it.
   * 	newstate: the requested new state for this lock (can be the same as old state)
   * 	period: the requested lease period for this lock.
   * 	ret: placeholder to tell the client the current state if it is not equal to oldstate.
   *
   */
  int renew(const host requester,const wfs_slice slice,const std::string oldstate,const std::string newstate,
  			  const unsigned period, wcc_protocol::ret& ret);

  /*
   * Decision function can_renew - Checks two things:
   * 	That the lock is held by that client
   * 	That the requester's reported slice_state matches the slice_state on the server.
   * 	Since we treat the string as an opaque state - this is just a simple string comparison.
   */
  rsm::dec_ret can_renew(const host requester,const wfs_slice slice,const std::string oldstate,const std::string newstate,
  			const unsigned period, wcc_protocol::ret& ret, int& wccret);

  /*
   * gracefully release a lock that that client previously acquired.
   * We must require that the client has the same lock state as the current state on the rsm.
   */
  int release(const host requester,const wfs_slice slice,const std::string oldstate, wcc_protocol::ret& ret);

  /*
   * release decision function - See if the client holds this lock,and if client's state is ok.
   *  if so we can replicate this action.
   */
  rsm::dec_ret can_release(const host requester,const wfs_slice slice,const std::string oldstate, wcc_protocol::ret& ret, int& wccret);

  int fetch_slices(const viewstamp clntvs, wcc_protocol::fetch_ret& ret);


  // auth stuff
  rsm::dec_ret can_update_auth_host_keys(const std::string hosts_file, 
					 int auth_code, int &dummy, 
					 int& wccret);
  int update_auth_host_keys(const std::string hosts_file, 
			    int auth_code, int &dummy);
  rsm::dec_ret can_update_auth_user_keys(const std::map< std::string, std::string > users_map, 
					 int auth_code, int &dummy, 
					 int& wccret);
  int update_auth_user_keys(const std::map< std::string, std::string > users_map, 
			    int auth_code, int &dummy);

  void monitor_auth_files();

  // State transfer parameters
  std::string marshal_state();
  void unmarshal_state(std::string state);

};


#endif /*WFS_RSM_CSERVER_H_*/
