/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PRTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// RPC stubs for overfuse to talk to overserve.

#include "wfs_cache_client.h"

#include <host_marshall.h>
#include <method_thread.h>
#include <wfs_misc.h>
#include "wfs_server.h"

#include <math.h>
#include <time.h>

wfs_cache_client::wfs_cache_client( sockaddr_in addr, wfs_node me,
				    slice_manager *slices, wfs_cache *cache,
				    rpcs *rpc_server, vivaldi *viv,
				    wfs_client_manager *clients,
				    bool read_ahead, int prefetch_threads )
  : wfs_client( me, slices, viv, clients ),
    _cache(cache), _host( addr.sin_addr.s_addr, addr.sin_port ),
    _fileprefetch( _cache, this, me, slices, viv, clients ),
    _dirprefetch( _cache, this, &_leases, me, slices, viv, clients ) {

  _bv_me = _me;
  _bv_me.id++;

  rpc_server->set_vivaldi(_viv);
  // register lease callback RPCs
  rpc_server->reg( wfs_protocol::invalidate, this,
                   &wfs_cache_client::invalidate_lease );

  // start the background lease renewal thread
  method_thread( this, true, &wfs_cache_client::lease_renewer );

  _read_ahead = read_ahead;

  //start the prefetching threads
  _fileprefetch.start(prefetch_threads);
  _dirprefetch.start(prefetch_threads);
}

void wfs_cache_client::lease_renewer() {

  while(1) {

    wfs_node node;
    std::vector<wfs_file_handle> fhs;

    wfs_log(WFS_DBG_1, "About to wait for new lease batch\n" );

    _leases.wait_for_next_batch( node, fhs );

    wfs_log(WFS_DBG_1, "New lease renewal batch from node %016"PRIx64", %u fhs\n", node.id,
	      (unsigned)fhs.size() );

    if( fhs.size() > 0 ) {
      int dummy;
      const static rpcc::TO to =
	rpcc::to(wfs_lease_manager::LEASE_BATCH_MARGIN*1000);
      wfs_protocol::status s;

      // get the right node id (cause metadata node ids are incremented)
      wfs_node node_right =
	make_node( inet_ntoa( node.addr.sin_addr ),
		   htons( node.addr.sin_port ) );

      rpcc *client = _clients->get_client( node_right, to );
      if( client == NULL ) {
	s = rpc_const::timeout_failure;
      } else {
	s = client->call( wfs_protocol::renew_lease_batch,
			  fhs, _host, dummy, to, "host" );
	_clients->done_with_client(node_right);
      }
      if( s == wfs_protocol::OK ) {
	_leases.renew_leases( node, fhs );
      } else {
        wfs_log(WFS_DBG_1, "Couldn't renew leases on node %016"PRIx64": %d\n", node.id, s );
      }
    }

  }

}

wfs_protocol::status
wfs_cache_client::getfile( wfs_file_handle &fh,
			   const std::string &user,
			   wfs_protocol::fileinfo &fin,
                           host *client /*ignored*/ ) {

  // let's set any permanent cues we have cached for this guy, in case
  // we can't reach the file's server
  wfs_protocol::fileinfo tmp_fin;
  if( _cache->read_any_md( fh, tmp_fin ) ) {
    fh.set_perm_cues( tmp_fin.perm_cues );
  }

  wfs_node file_node = get_primary( file_handle_to_file_id(fh) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  //claim this lookup from the prefetching thread
  _dirprefetch.lookup(fh);

  // Manage file meta-data storage using leases

  if( (((fh.get_cue( WFS_CUE_CONSISTENCY ).val == WFS_CUE_EVENTUAL) &&
	_leases.has_lease( fh, _bv_me )) ||
       _leases.has_lease( fh, file_node )) &&
      _cache->read_md( fh, fin ) ) {

    wfs_log(WFS_DBG_1, "Read file md for %016"PRIx64" out of cache; size=%lld\n", fh.fid,
	      fin.size);

    if( fh.version.version_num >= 0 && fh.version != fin.version) {
      return wfs_protocol::BADVERSION;
    }

    _cache->store_md( fh, fin );

    // version files don't contain the regular contents, so the size should
    // be however big the ASCII equivalent of an int is.
    if( fh.is_version ) {
      fin.size = WFS_VERSION_FILE_SIZE;
    }

    // Check for permanent cues, since they might not yet be set in the
    // file handle
    fh.set_perm_cues( fin.perm_cues );
    return wfs_protocol::OK;

  } else {

    _leases.reg_pending_lease(fh);

    // TODO: In the future, with a best version cue, we may want to
    // send lookups to multiple nodes in parallel, in case the responsible
    // node fails and we can find a newer version than our cached version
    // somewhere else within the time limit.

    wfs_protocol::status status = wfs_client::getfile( fh, user, fin, &_host );

    bool lease_ok = _leases.unreg_pending_lease(fh);

    // asking for the best version means that if we can't find the original,
    // it's ok to use the one in our cache
    if( status == -1 &&
        fh.get_cue( WFS_CUE_CONSISTENCY ).val == WFS_CUE_EVENTUAL &&
        _cache->read_any_md( fh, fin ) ) {


      wfs_log(WFS_DBG_1, "Read file md for %016"PRIx64" out of cache under bestversion\n", fh.fid );

      fh.set_perm_cues( fin.perm_cues );
      if( !fh.is_version ) {
        _cache->store_md( fh, fin );
      }
      // HACK: Give yourself a very short lease when looking at best
      // version data, so subsequent lookups will be fast.  It's possible
      // this will cause problems, since it won't get invalidated
      _leases.add_lease( fh, _bv_me, 2 );
      return wfs_protocol::OK;

    } else {

      if( status == wfs_protocol::OK_FLUSH ) {
        _cache->clear_md( fh, fin.version );
      }

      if( lease_ok && !fh.is_version &&
          (status == wfs_protocol::OK_FLUSH || status == wfs_protocol::OK) ) {
        _leases.add_lease( fh, file_node, fhlease(fh) );
	wfs_protocol::fileinfo xfin;
	if( _cache->read_md( fh, xfin ) && xfin.version == fin.version ) {
	  // it hasn't changed on the server, so what we have (which could
	  // include a bunch of changes) is still good
	  // TODO: what if it has changed on the server?  Our lease would
	  // be invalidated, but what would happen to buffered writes?
	} else {
	  _cache->store_md( fh, fin );
	}
      }
      return status;

    }

  }

}

wfs_protocol::status wfs_cache_client::getdir( wfs_file_handle &fh,
					       const std::string &user,
                                               wfs_protocol::dirinfo &fin,
                                               host *client /*ignored*/ ) {

  // we use fileinfo objects for directory infos, since they're a superset.
  // if this ever changes, we're screwed

  wfs_protocol::fileinfo file_fin;
  wfs_node file_node = get_primary( fh.fid );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_log(WFS_DBG_1, "Getdir %016"PRIx64" from node %016"PRIx64"\n",
	    fh.fid, file_node.id );

  // hack to tell directory listings apart from meta-data
  file_node.id++;

  if( (((fh.get_cue( WFS_CUE_CONSISTENCY ).val == WFS_CUE_EVENTUAL) &&
	_leases.has_lease( fh, _bv_me )) ||
       _leases.has_lease( fh, file_node )) &&
      _cache->read_md( fh, file_fin ) ) {

    fin.perm_cues = file_fin.perm_cues;
    fin.permissions = file_fin.permissions;
    fin.size = file_fin.size;
    fin.atime = file_fin.atime;
    fin.ctime = file_fin.ctime;
    fin.mtime = file_fin.mtime;
    wfs_log(WFS_DBG_1, "Read md for dir %016"PRIx64" out of cache\n", fh.fid );

  } else {

    _leases.reg_pending_lease(fh);
    wfs_log(WFS_DBG_1, "About to getdir %016"PRIx64" from node %016"PRIx64"\n",
	      fh.fid, file_node.id );
    wfs_protocol::status s = wfs_client::getdir( fh, user, fin, &_host );
    wfs_log(WFS_DBG_1, "Finished getdir %016"PRIx64" from node %016"PRIx64" -- %d\n",
	      fh.fid, file_node.id, s );
    bool lease_ok = _leases.unreg_pending_lease(fh);

    if( s == -1 &&
        fh.get_cue( WFS_CUE_CONSISTENCY ).val == WFS_CUE_EVENTUAL &&
        _cache->read_any_md( fh, file_fin ) ) {

      _leases.add_lease( fh, _bv_me, 2 );
      fin.perm_cues = file_fin.perm_cues;
      fin.permissions = file_fin.permissions;
      fin.size = file_fin.size;
      fin.atime = file_fin.atime;
      fin.ctime = file_fin.ctime;
      fin.mtime = file_fin.mtime;

      wfs_log(WFS_DBG_1, "Read md for dir %016"PRIx64" out of cache under bestversion\n", fh.fid );

    } else {

      if( s != wfs_protocol::OK ) {
	return s;
      }

      if( s == wfs_protocol::OK_FLUSH ) {
	_cache->clear_md( fh, wfs_version() /*version not used*/ );
      }

      file_fin.perm_cues = fin.perm_cues;
      file_fin.permissions = fin.permissions;
      file_fin.size = fin.size;
      file_fin.atime = fin.atime;
      file_fin.ctime = fin.ctime;
      file_fin.mtime = fin.mtime;
      file_fin.version = wfs_version();
      if( lease_ok &&
	  (s == wfs_protocol::OK_FLUSH || s == wfs_protocol::OK) ) {
	_leases.add_lease( fh, file_node, fhlease(fh) );
	_cache->store_md( fh, file_fin );
      }

    }

  }

  fh.set_perm_cues( fin.perm_cues );
  return wfs_protocol::OK;

}

wfs_protocol::status wfs_cache_client::lookup( wfs_file_handle dir_handle,
                                               std::string name,
					       const std::string &user,
                                               wfs_file_handle &xfh,
                                               host *client /*ignored*/ ) {

  wfs_file_handle dummy;
  // let wfs_client handle cue stuff
  if( dir_handle.is_nodes || dir_handle.is_sites || dir_handle.site != "" ||
      search_and_set_cues( name, dummy ) ) {
    return wfs_client::lookup( dir_handle, name, user, xfh );
  }

  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  // only read it from the cache if we have a valid lease
  // TODO: If the directory is eventual but we've hit the maxtime limit with
  // the other nodes, maybe we can use what's in the cache . . .
  if( _leases.has_lease( dir_handle, dir_node ) ) {
    wfs_protocol::fileinfo di;
    if (!_cache->read_md(dir_handle, di)) {
      return wfs_protocol::IOERR;
    } else if (!has_permission(di.permissions, user, WFS_PERMISSION_READ)) {
      return wfs_protocol::PERM;
    }
    xfh = dir_handle;
    if( _cache->dir_lookup( dir_handle, name, xfh ) ) {
      wfs_log(WFS_DBG_1, "Read name %s from dir %016"PRIx64" from cache under a lease\n",
                name.c_str(), dir_handle.fid );
      xfh.inc_depth();
      if( xfh.fid == 0 ) {
        return wfs_protocol::NOENT;
      } else {
	return wfs_protocol::OK;
      }
    }
  }

#ifdef MAIL_HACK

  /*******
   * HACK FOR OSDI!!!!!
   *
   * Preload the root directory cache
   *
   ******/
  if( dir_handle.fid == 0x00000001 ) {

    wfs_storage storage( "/tmp/dir12345", (_local_server == NULL) );
    wfs_file_id fid;
    if( storage.lookup( dir_handle, name, fid ) ) {
      if( fid != 0 ) {
	xfh = dir_handle;
	xfh.fid = fid;
	_leases.add_lease( dir_handle, dir_node, WFS_LEASE_SEC_DEFAULT );
	_cache->add_dir_entry( dir_handle, name, xfh );
	wfs_log(WFS_DBG_1, "Direct-caching name %s,%016"PRIx64" from dir %016"PRIx64" from cache under a lease\n",
		  name.c_str(), fid, dir_handle.fid );
	return wfs_protocol::OK;
      }
    }
  }
  /*****
   * END HACK
   ****/
#endif

  _leases.reg_pending_lease(dir_handle);

  wfs_protocol::lookupinfo li;
  wfs_protocol::status s = wfs_client::lookup( dir_handle, name, user,
					       li, &_host );
  xfh = li.fh;

  bool lease_ok = _leases.unreg_pending_lease(dir_handle);

  // asking for the best version means that if we can't find the original,
  // it's ok to use the one in our cache
  if( s == -1 &&
      dir_handle.get_cue( WFS_CUE_CONSISTENCY ).val == WFS_CUE_EVENTUAL &&
      _cache->dir_lookup( dir_handle, name, xfh ) ) {

    wfs_log(WFS_DBG_1, "Read name %s from dir %016"PRIx64" from cache under bestversion\n",
	      name.c_str(), dir_handle.fid );

    s = wfs_protocol::OK;
    _leases.add_lease( dir_handle, _bv_me, 2 );

  } else {

    // flush the cache either if the server says to, or if our lease expired
    // (or we never had one to begin with)
    if( s == wfs_protocol::OK_FLUSH ||
	!_leases.has_lease( dir_handle, dir_node ) ) {
      _cache->flush_dir( dir_handle );
    }

    if( lease_ok &&
	(s == wfs_protocol::OK_FLUSH || s == wfs_protocol::OK ||
	 s == wfs_protocol::NOENT) ) {
      // a new lease is implicit with every successful reply from the remote
      // node
      _leases.add_lease( dir_handle, dir_node, fhlease(dir_handle) );
      _cache->add_dir_entry( dir_handle, name, xfh );

      // if the node was the primary, cache any extra info we might
      // have gotten.
      if( li.extra_info.size() > 0 &&
	  dir_node.id == get_primary( file_handle_to_file_id(xfh) ).id ) {
	if( isdir(xfh) ) {
	  dir_node.id++;
	}
	_leases.add_lease( xfh, dir_node, fhlease(xfh) );
	wfs_protocol::fileinfo xfin;
	if( _cache->read_md( xfh, xfin ) &&
	    xfin.version == li.extra_info[0].version ) {
	  // it hasn't changed on the server, so what we have (which could
	  // include a bunch of changes) is still good
	  // TODO: what if it has changed on the server?  Our lease would
	  // be invalidated, but what would happen to buffered writes?
	} else {
	  _cache->store_md( xfh, li.extra_info[0] );
	}
      }

      // if somehow we didn't have the meta-data for the parent yet
      // (apparently some systems like PlanetLab don't issue a lookup
      // for the root parent before doing a lookup in the parent),
      // we need to get the parent' meta-data into the cache
      wfs_protocol::fileinfo file_fin;
      if( !_cache->read_md( dir_handle, file_fin ) ) {
	wfs_protocol::dirinfo dir_fin;
	getdir( dir_handle, user, dir_fin, client );
      }

    }

  }

  return s;

}

wfs_protocol::status wfs_cache_client::readdir( wfs_file_handle fh,
                                                unsigned long long ind,
                                                unsigned int n,
						const std::string &user,
                                                std::vector<wfs_storage::dirent> &e,
                                                host *client /*ignored*/) {
  if( fh.is_nodes || fh.is_sites || fh.site != "" ) {
    wfs_log(WFS_DBG_2, "returning a special directory\n");
    return wfs_client::readdir( fh, ind, n, user, e );
  }

  wfs_node dir_node = get_primary( file_handle_to_file_id( fh ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  if( _leases.has_lease( fh, dir_node ) ) {
    wfs_protocol::fileinfo di;
    if (!_cache->read_md(fh, di)) // If we have the lease, we should have metadata?
      return wfs_protocol::IOERR;
    else if (!has_permission(di.permissions, user, WFS_PERMISSION_READ)) {
      return wfs_protocol::PERM;
    }
    if( _cache->readdir( fh, ind, n, e ) ) {
      wfs_log(WFS_DBG_1, "Readdir %016"PRIx64" from cache under a lease\n", fh.fid );

      if (fh.get_cue( WFS_CUE_PREFETCH ).val == WFS_CUE_WHOLEFILE) {
        wfs_log(WFS_DBG_1, "DIR PREFETCHING: ON\n");
        _dirprefetch.directory( fh, e, user );
      }
      return wfs_protocol::OK;
    }
  }

  _leases.reg_pending_lease(fh);

  wfs_protocol::status s = wfs_client::readdir( fh, ind, n, user, e, &_host );

  bool lease_ok = _leases.unreg_pending_lease(fh);

  // flush the cache either if the server says to, or if our lease expired
  // (or we never had one to begin with)
  if( s == wfs_protocol::OK_FLUSH ||
      !_leases.has_lease( fh, dir_node ) ) {
    _cache->flush_dir( fh );
  }

  if( lease_ok &&
      (s == wfs_protocol::OK_FLUSH || s == wfs_protocol::OK) ) {
    // a new lease is implicit with every successful reply from the remote
    // node
    _leases.add_lease( fh, dir_node, fhlease(fh) );
    // assume we've cached the whole directory if we've read less than
    // the requested number of entries
    bool complete = (e.size() < n);
    _cache->add_readdir_entry( fh, ind, e, complete );

  }
  if (fh.get_cue( WFS_CUE_PREFETCH ).val == WFS_CUE_WHOLEFILE) {
    wfs_log(WFS_DBG_1, "DIR PREFETCHING: ON\n");
    _dirprefetch.directory( fh, e, user );
  }
  return s;

}

wfs_protocol::status
wfs_cache_client::read_uncached_data( wfs_file_handle fh,
                                      wfs_version version,
                                      unsigned long long off,
                                      unsigned int n,
				      const std::string &user,
                                      std::string &buf ) {

  return wfs_client::readfile( fh, version, off, n, user, buf );

}
wfs_protocol::status
wfs_cache_client::readfile( wfs_file_handle fh,
                            wfs_version version, unsigned long long off,
                            unsigned int n,
			    const std::string &user,
                            std::string &buf ) {
  // check the cache for any cached blocks
  // TODO: Verify that we have the latest version number before returning
  // cached blocks?
  unsigned int num_blocks;
  unsigned int first_block = calc_block_no( off, n, num_blocks );
  int block_off = offset_within_block( first_block, off );
  assert( block_off >= 0 && (unsigned)block_off < WFS_BLOCK_SIZE );
  unsigned int size = n;
  buf = "";

  //find out the size of the file
  wfs_protocol::fileinfo info;
  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  bool r = _cache->read_md(fh, info);
  if( !r || !_leases.has_lease( fh, file_node ) ) {
    wfs_protocol::fileinfo info2;
    bool r2 = (getfile( fh, user, info2 ) == wfs_protocol::OK);
    // if we still have the newest working copy of the file,
    // keep our old metadata -- it may have important
    // size info from unbuffered writes
    if( r && info2.version <= info.version ) {
      _cache->store_md( fh, info );
    } else // Make sure we're using the latest permissions
      info = info2;
    r = r2;
  }

  // Check permissions here in case we're reading all blocks out of the cache
  if (!has_permission(info.permissions, user, WFS_PERMISSION_READ))
      return wfs_protocol::PERM;

  //check that we are not reading off of the end of the file
  if( r && (off >= info.size) ) {
    wfs_log(WFS_DBG_1, "offset is past the end of the file: %llu\n",off);
    return wfs_protocol::OK; //? what should I return??
  }

  if( r && ((off+n) > info.size) ) {
    wfs_log(WFS_DBG_1, "reading off the end of the file: offset %llu size %d\n",
	     off,n);
    first_block = calc_block_no(off, info.size-off, num_blocks);
    size = info.size-off;
  }

  wfs_protocol::status status;
  int threads = 0;
  wfs_filefetch fetch( _cache, this, _bv_me, _slices, _viv, _clients);
  for( unsigned int i = first_block; i < (first_block+num_blocks); i++ ) {
    if( fh.is_version ||
	!_cache->read_block( fh, version, i ) ||
	!_fileprefetch.lookup( fh, i ) ) {
      //add the block to be fetched
      fetch.add(fh, i, version, user);
      threads++;
    }
  }
  if (threads > 5) {
    threads = 5;
  } 
  pthread_t t[threads];
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  fetch.start(threads,t,&attr);
  pthread_attr_destroy(&attr);
    
  //start up prefetching
  if (fh.get_cue( WFS_CUE_PREFETCH ).val == WFS_CUE_WHOLEFILE) {
    wfs_log( WFS_DBG_1, "FILE PREFETCHING: ON\n");
    _fileprefetch.wholefile( fh, version, user, first_block, num_blocks );
  }
  else if (_read_ahead){
    wfs_log(WFS_DBG_1, "FILE READ-AHEAD: ON\n");
    _fileprefetch.readahead( fh, version, user, first_block, num_blocks );
  }
  
  //join the threads
  for( int i = 0; i < threads; i++) {
    int r = pthread_join(t[i], NULL);
    assert(r == 0);
  }

  std::string block_buf;
  for( unsigned int i = first_block; i < (first_block+num_blocks); i++ ) {
    _fileprefetch.wait(fh, i);
    status = fetch.get(fh, i);

    if( status != wfs_protocol::OK ) {
      if( status == wfs_protocol::BADVERSION ) {
	_cache->clear_md( fh, version );
      }
      return status;
    } else if( !_cache->read_block( fh, version, i, block_buf) ) {
      wfs_log(WFS_DBG_1, "couldn't find block in cache\n");
      return wfs_protocol::IOERR;
    }

    unsigned int curr_size;
    if( block_off + size < block_buf.length() ) {
      curr_size = size;
    } else {
      curr_size = block_buf.length() - block_off;
    }

    buf.append( block_buf.substr( block_off, curr_size ) );
    size -= curr_size;
    block_off = 0;

  }
  return wfs_protocol::OK;

}

wfs_protocol::status wfs_cache_client::writefile( wfs_file_handle fh,
                                                  wfs_version version,
                                                  unsigned long long off,
                                                  const std::string &buf,
						  const std::string &user,
                                                  wfs_version &new_version,
                                                  host *client /*ignored*/ ) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  // also important to update the metadata
  // ~xavid: get the metadata first to check permissions, so we don't
  // let unauthorized users pollute the cache
  wfs_protocol::fileinfo finfo;

  // we need to know what the real size should be, so if the md's not
  // already cached, fetch it
  bool r = _cache->read_md(fh, finfo);
  if( !r || !_leases.has_lease( fh, file_node ) ) {
    wfs_protocol::fileinfo info2;
    wfs_protocol::status r2 = getfile( fh, user, info2 );
    if (r2 == wfs_protocol::OK){
      // if we still have the newest working copy of the file,
      // keep our old metadata -- it may have important
      // size info from unbuffered writes
      if( r && info2.version <= finfo.version ) {
	_cache->store_md( fh, finfo );
      } else {
	finfo = info2;
      }
    } else {
      return r2;
    }
  }

  if (!has_permission(finfo.permissions, user, WFS_PERMISSION_WRITE))
    return wfs_protocol::PERM;

  // if this is the first version, better store a blank block in case
  // the user reads it without closing it first
  // TODO: This is sort of a hack, and probably won't work for initial
  // writes past the first block . . .
  if( version.version_num == 1 && _cache->num_buf_writes(fh) == 0 ) {
    unsigned int num_blocks;
    unsigned int first_block = calc_block_no(off, buf.size(), num_blocks);
    for( unsigned i = 0; i < num_blocks; i++ ) {
      _cache->store_block( fh, version, first_block+i, "" );
    }
  }

  new_version.version_num = version.version_num;
  new_version.node_id = file_node.id;

  // cheat and just write it to the cache
  if( !_cache->buf_write( fh, version, off, false, buf ) ) {
    // there's not enough room to cache this guy, so flush everything first
    wfs_protocol::status s = flush( fh, user, new_version, client );
    if( s != wfs_protocol::OK ) {
      return s;
    }
    assert( _cache->buf_write( fh, version, off, false, buf ) );
  }

  // Update the metadata
  finfo.version = new_version;
  finfo.mtime = time(NULL);
  finfo.ctime = finfo.mtime;

  unsigned long long new_size = off + buf.length();
  if( finfo.size < new_size ) {
    finfo.size = new_size;
    wfs_log(WFS_DBG_1, "New size of file %016"PRIx64" is %lld (%lld)\n", fh.fid, finfo.size,
	      new_size );
  }
  _cache->store_md( fh, finfo );

  return wfs_protocol::OK;

}

wfs_protocol::status wfs_cache_client::flush( wfs_file_handle fh,
					      const std::string &user,
                                              wfs_version &new_version,
                                              host *client /*ignored*/ ) {


  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  std::vector<wfs_buf_write> writes;
  _cache->get_buf_writes( fh, writes );

  if( writes.size() == 0 ) {
    return wfs_protocol::OK;
  }

  // tentatively update cached metadata (see comment in
  // wfs_fuse_client.cc::flush)
  // xavid: this is weird if the writes fail due to, say, permissions
  wfs_protocol::fileinfo finfo;
  if( _cache->read_md( fh, finfo ) ) {
    finfo.version.version_num++;
    _cache->store_md( fh, finfo );
  }

  // TODO: could be smarter here about not sending writes that get
  // cancelled out by other, later writes
  std::vector<wfs_protocol::write> batch;
#define MAX_BATCH_SZ 5*1024*1024
  int batch_sz = 0;
  wfs_protocol::status ret = wfs_protocol::OK;

  for( unsigned i = 0; i < writes.size(); i++ ) {

    wfs_protocol::write w;

    w.off = writes[i].offset;
    w.data = writes[i].buf;
    w.truncate = writes[i].truncate;

    batch.push_back(w);

    batch_sz += w.data.size();
    if( batch_sz > MAX_BATCH_SZ || (i+1) == writes.size() ) {
      ret = wfs_client::writefile_batch( fh, batch, user,
					 new_version, &_host );
      if( ret != wfs_protocol::OK ) {
        return ret;
      }
      batch_sz = 0;
      batch.clear();
    }
  }

  // if I hold a lease, update my cached md for the file
  if( ret == wfs_protocol::OK && _leases.has_lease( fh, file_node ) ) {
    if( _cache->read_md( fh, finfo ) ) {
      finfo.version = new_version;
      _cache->store_md( fh, finfo );
    }
  }


  // TODO: we should be able to be smart here, and pre-cache the blocks
  // for this new version since we know about them already.  However,
  // there's a chance someone else's writes slipped in between our
  // write calls, so I'm not yet prepared to call any of the blocks we wrote
  // the definitive ones for the last version.

  return ret;

}

wfs_protocol::status wfs_cache_client::truncate( wfs_file_handle fh,
                                                 wfs_version version,
                                                 unsigned long long off,
						 const std::string &user,
                                                 wfs_version &new_version,
                                                 host *client /*ignored*/ ) {
  // Get metadata first to check permissions and prevent cache poisoning
  wfs_protocol::fileinfo finfo;

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  // we need to know what the real size should be, so if the md's not
  // already cached, fetch it
  bool r = _cache->read_md(fh, finfo);
  if( !r || !_leases.has_lease( fh, file_node ) ) {
    wfs_protocol::fileinfo info2;
    bool r2 = (getfile( fh, user, info2 ) == wfs_protocol::OK);
    // +xavid
    assert(r2);
    // if we still have the newest working copy of the file,
    // keep our old metadata -- it may have important
    // size info from unbuffered writes
    if( r && info2.version <= finfo.version ) {
      _cache->store_md( fh, finfo );
    } else {
      finfo = info2;
    }
  }

  if (!has_permission(finfo.permissions, user, WFS_PERMISSION_WRITE))
      return wfs_protocol::PERM;

  // cheat and just write it to the cache
  std::string empty = "";
  _cache->buf_write( fh, version, off, true, empty );

  // also important to update the metadata
  unsigned long long new_size = off;
  finfo.size = new_size;

  new_version.version_num = version.version_num;
  new_version.node_id = file_node.id;

  finfo.version = new_version;
  finfo.mtime = time(NULL);
  finfo.ctime = finfo.mtime;

  _cache->store_md( fh, finfo );

  return wfs_protocol::OK;

}

void wfs_cache_client::new_dir_entry( wfs_file_handle dir_handle,
				      wfs_node dir_node,
				      std::string name,
				      wfs_file_handle new_fh ) {
  // The permissions check is handled by caller
  if( _leases.has_lease( dir_handle, dir_node ) ) {
    wfs_protocol::fileinfo finfo;
    if( _cache->read_md( dir_handle, finfo ) ) {
      finfo.mtime = time(NULL);
      finfo.ctime = finfo.mtime;
      finfo.size = finfo.size + (name.size()+1+8 /*name+space+fid*/);
      _cache->store_md( dir_handle, finfo );
    }
    _cache->add_dir_entry( dir_handle, name, new_fh );
  }

}

wfs_protocol::status wfs_cache_client::create( wfs_file_handle dir_handle,
                                               std::string name,
					       const std::map<std::string,
					       wfs_permission> &permissions,
                                               wfs_file_type type,
					       const std::string &user,
                                               wfs_file_handle &new_fh,
                                               host *client /*ignored*/ ) {

  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_protocol::status s = wfs_client::create( dir_handle, name,
					       permissions, type,
					       user,
                                               new_fh, &_host );
  wfs_log(WFS_DBG_1, "create status for %s in dir %016"PRIx64" is %d\n", name.c_str(),
            dir_handle.fid, s );

  // if I hold a lease, update my cache for this directory
  if( s == wfs_protocol::OK ) {
    new_dir_entry( dir_handle, dir_node, name, new_fh );

    /**
    // if this is a new directory, create empty
    // cached entry for it
    if( is_dir ) {
      std::vector<wfs_storage::dirent> e;
      _cache->add_readdir_entry( new_fh, 0, e, true );
    }
    */

  }

  return s;

}

wfs_protocol::status wfs_cache_client::link( wfs_file_handle fh,
					     wfs_file_handle dir_handle,
					     std::string name,
					     bool create_file,
					     bool link_file,
					     wfs_file_type type,
					     const std::map<std::string,
					     wfs_permission> &permissions,
					     const std::string &user,
					     wfs_file_handle &new_fh,
					     wfs_version &new_file_version,
					     host *client ) {

  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_protocol::status s = wfs_client::link( fh, dir_handle, name,
					     create_file, link_file,
					     type, permissions,
					     user, new_fh, new_file_version,
					     &_host );

  wfs_log(WFS_DBG_1, "link status for %s in dir %016"PRIx64" is %d\n", name.c_str(),
	    dir_handle.fid, s );

  // if I hold a lease, update my cache for this directory and the file version
  if( s == wfs_protocol::OK ) {
    new_dir_entry( dir_handle, dir_node, name, fh );

    // if we have the md cached for the file, update the version
    wfs_protocol::fileinfo finfo;
    if( _cache->read_md(fh, finfo) ) {
      finfo.version = new_file_version;
      _cache->store_md(fh, finfo);
    }

  }

  return s;

}

wfs_protocol::status wfs_cache_client::unlink( wfs_file_handle dir_handle,
                                               std::string name, bool is_dir,
					       bool dec_link,
					       const std::string &user,
					       wfs_protocol::file_and_version &fv,
                                               host *client /*ignored*/ ) {

  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_protocol::status s = wfs_client::unlink( dir_handle, name, is_dir,
					       dec_link, user, fv,
                                               &_host );

  // if I hold a lease, update my cache for this directory
  if( s == wfs_protocol::OK && _leases.has_lease( dir_handle, dir_node ) ) {
    wfs_file_handle x;
    x.fid = 0;
    _cache->add_dir_entry( dir_handle, name, x );
    wfs_protocol::fileinfo finfo;
    if( _cache->read_md( dir_handle, finfo ) ) {
      finfo.mtime = time(NULL);
      finfo.ctime = finfo.mtime;
      finfo.size = finfo.size - (name.size()+1+8 /*name+space+fid*/);
      _cache->store_md( dir_handle, finfo );
    }
  }
  wfs_protocol::fileinfo finfo;
  if( dec_link && s == wfs_protocol::OK && _cache->read_md(fv.fh, finfo) ) {
    finfo.version = fv.ver;
    _cache->store_md(fv.fh, finfo);
  }

  return s;

}

wfs_protocol::status wfs_cache_client::rename( wfs_file_handle old_parent,
                                               std::string old_name,
                                               wfs_file_handle new_parent,
                                               std::string new_name,
					       const std::string &user,
                                               host *client /*ignored*/ ) {

  wfs_node dir_node1 = get_primary( file_handle_to_file_id( old_parent ) );
  if( dir_node1.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_node dir_node2 = get_primary( file_handle_to_file_id( new_parent ) );
  if( dir_node2.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  // necessary either way (TODO: pass fh back from rename?)
  wfs_file_handle old_fh;
  wfs_protocol::status s2 = lookup( old_parent, old_name, user, old_fh );

  wfs_protocol::status s = wfs_client::rename( old_parent, old_name,
                                               new_parent, new_name,
					       user,
                                               &_host );

  if( dir_node1.id == dir_node2.id ) {
    // if I hold a lease, update my cache for this directory
    if( s == wfs_protocol::OK && _leases.has_lease( old_parent, dir_node1 ) ) {
      wfs_file_handle x;
      x.fid = 0;
      _cache->add_dir_entry( old_parent, old_name, x );
      wfs_protocol::fileinfo finfo;
      if( _cache->read_md( old_parent, finfo ) ) {
        finfo.mtime = time(NULL);
        finfo.ctime = finfo.mtime;
        finfo.size = finfo.size - (old_name.size()+1+8 /*name+space+fid*/);
        _cache->store_md( old_parent, finfo );
      }
    }
  } else {
    // the unlink is handled by unlink()
  }

  if( s == wfs_protocol::OK && s2 == wfs_protocol::OK &&
      _leases.has_lease( new_parent, dir_node2 ) ) {
    _cache->add_dir_entry( new_parent, new_name, old_fh );
    wfs_protocol::fileinfo finfo;
    if( _cache->read_md( new_parent, finfo ) ) {
      finfo.mtime = time(NULL);
      finfo.ctime = finfo.mtime;
      finfo.size = finfo.size + (new_name.size()+1+8 /*name+space+fid*/);
      _cache->store_md( new_parent, finfo );
    }
  }

  return s;

}

wfs_protocol::status wfs_cache_client::chmod( wfs_file_handle fh,
					      const std::map<std::string,
					      wfs_permission> &permissions,
					      const std::string &user,
                                              wfs_version &new_version,
                                              host *client /*ignored*/) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );

  wfs_protocol::status s = wfs_client::chmod( fh, permissions, user,
					      new_version,
                                              &_host );

  // if I hold a lease, update my cache for this directory
  if( s == wfs_protocol::OK && _leases.has_lease( fh, file_node ) ) {
    wfs_protocol::fileinfo finfo;
    if( _cache->read_md( fh, finfo ) ) {
      // Only the mentioned permissions should change
      std::map<std::string,wfs_permission>::const_iterator i;
      for (i = permissions.begin(); i != permissions.end(); i++)
	finfo.permissions[i->first] = i->second;
      _cache->store_md( fh, finfo );
    }
  }

  return s;


}

wfs_protocol::status wfs_cache_client::invalidate_lease( wfs_file_handle fh,
                                                         int &dummy ) {

  wfs_log(WFS_DBG_1, "Invalidating lease for %016"PRIx64"\n", fh.fid );

  // invalidate all leases for everyone
  _leases.invalidate_lease( fh );

  if( isdir(fh) ) {
    _cache->flush_dir( fh );
  } else {
    // we only use file meta-data either:
    //   a) after checking whether we have the lease, or
    //   b) when doing a write/truncate operation, in which case we assume
    //      no write-sharing and just keep using our old meta-data no matter
    //      what.
  }

  return wfs_protocol::OK;

}

