/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_fetch.h"
#include "wfs_cache_client.h"
#include "wfs_options.h"
#include <host_marshall.h>
#include <math.h>
#include <wfs_misc.h>

static void *
fetchthread(void *x)
{
  wfs_fetch *p = (wfs_fetch *) x;
  p->fetcher();
  return 0;
}

wfs_fetch::wfs_fetch(wfs_cache *cache, wfs_cache_client *cc, wfs_node me,
		     slice_manager *slices, vivaldi *viv,
		     wfs_client_manager *clients)
  : _c(me, slices, viv, clients) {
  //wfs cache for checking if the thing to be fetched is already fetched
  _cache = cache;
  //cache client
  _cc = cc;

  //lock and condition variable for queue of things to fetch
  assert( pthread_mutex_init(&_q_lock, NULL) == 0 );
  assert( pthread_cond_init(&_q_wait, NULL) == 0 );

  //lock for hash table that looks up a fetch request using the tag
  assert( pthread_mutex_init(&_p_lock, NULL) == 0 );
}

wfs_fetch::~wfs_fetch() {
  pthread_mutex_destroy(&_q_lock);
  pthread_cond_destroy(&_q_wait);
  pthread_mutex_destroy(&_p_lock);

}

//actually start some threads
//must do this after the subclass is constructed
void
wfs_fetch::start(int num_threads, pthread_t threads[], const pthread_attr_t *attr) {
  //start the threads
  for (int thread_num = 0; thread_num < num_threads; thread_num++)
    assert( pthread_create(&(threads[thread_num]), attr,
			   &fetchthread, (void *) this) == 0 );
}

wfs_filefetch::wfs_filefetch(wfs_cache *cache, wfs_cache_client *cc, wfs_node me,
		     slice_manager *slices, vivaldi *viv,
		     wfs_client_manager *clients)
  : wfs_fetch(cache, cc, me, slices, viv, clients) { }

//fetches all of the requests in the queue and exits
//do not start threads until the queue is full!
void
wfs_filefetch::fetcher() {
  //tag of the current request we are working on
  req_tag tag;
  //information for the current request
  file_req *req;
  std::string buf;
  wfs_protocol::status status;

  pthread_mutex_lock(&_q_lock);
  while(!_queue.empty()) {
    //grab the first thing off of the queue
    tag = _queue.front();
    _queue.pop();
    pthread_mutex_unlock(&_q_lock);

    //get the information for the block request
    pthread_mutex_lock(&_p_lock);
    if( _pending.count(tag) > 0 ) {
      //found the request
      req = (file_req*)_pending[tag];
      pthread_mutex_unlock(&_p_lock);

      status = wfs_protocol::OK;
      if (!_cache->read_block(req->fh, req->version, req->bno)) {
        wfs_log(WFS_DBG_1, "About to read uncached block %d for file %016"PRIx64",%d, size=%d\n",
		 req->bno, req->fh.fid, req->version.version_num,
		 WFS_BLOCK_SIZE);

	status = _cc->read_uncached_data(req->fh, req->version,
					 (req->bno)*WFS_BLOCK_SIZE,
					 WFS_BLOCK_SIZE,
					 req->user, buf);

	if (status == wfs_protocol::OK) {
	  wfs_log(WFS_DBG_1, "Storing block %d to cache for file %016"PRIx64",%d, size=%d\n",
		   req->bno, req->fh.fid, req->version.version_num,
		   WFS_BLOCK_SIZE);
	  _cache->store_block(req->fh, req->version, req->bno, buf);
	  _cache->apply_buf_writes(req->fh, req->version, req->bno, buf);
	}
      }

      req->ret = status;
    } else {
      pthread_mutex_unlock(&_p_lock);
    }
    pthread_mutex_lock(&_q_lock);
  }
  pthread_mutex_unlock(&_q_lock);
  pthread_exit(NULL);
}

//add block request to fetched
//add requests before starting the fetching threads!
bool
wfs_filefetch::add(wfs_file_handle fh, unsigned int bno, wfs_version version,
		   std::string user) {

  req_tag tag = req_tag(fh.fid,bno);
  file_req *req = new file_req(fh, bno, version, user);
  bool r = false;
  pthread_mutex_lock(&_p_lock);
  if (_pending.count(tag) == 0) {
    _pending[tag] = req;
    r = true;
  }
  pthread_mutex_unlock(&_p_lock);

  if (r) {
    pthread_mutex_lock(&_q_lock);
    _queue.push(tag);
    pthread_mutex_unlock(&_q_lock);
  }
  return r;
}

wfs_protocol::status
wfs_filefetch::get(wfs_file_handle fh, unsigned int bno) {
  req_tag tag = req_tag(fh.fid, bno);
  wfs_protocol::status ret = wfs_protocol::OK;
  pthread_mutex_lock(&_p_lock);
  if (_pending.count(tag) > 0) {
    file_req *req = (file_req*) _pending[tag];
    ret = req->ret;
    _pending.erase(tag);
    wfs_log(WFS_DBG_2,"wfs_filefetch: deleting block %d\n",bno);
    delete req;
  }
  pthread_mutex_unlock(&_p_lock);
  return ret;
}
