/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_prefetch.h"
#include "wfs_cache_client.h"
#include "wfs_options.h"
#include <host_marshall.h>
#include <math.h>
#include <wfs_misc.h>

wfs_prefetch::wfs_prefetch(wfs_cache *cache, wfs_cache_client *cc, wfs_node me,
			   slice_manager *slices, vivaldi *viv,
			   wfs_client_manager *clients)
  : wfs_fetch(cache, cc, me, slices, viv, clients) {

  _req_window = 0;
  _req_window_size = 1;
#if RATE_CONTROLLED
  assert( pthread_mutex_init(&_win_lock, NULL) == 0 );
  for (int i = 0; i < 100; i++)
    _avg_latencies[i] = 0;
  _total_reqs = 0;
#endif

}

wfs_prefetch::~wfs_prefetch() {
#if RATE_CONTROLLED
  pthread_mutex_destroy(&_win_lock);
#endif
}

void
wfs_prefetch::start(int num_threads) {
  threadpool_size = num_threads;
#if RATE_CONTROLLED
  _req_window_size = 1;
#else
  _req_window_size = num_threads;
#endif
  pthread_t t[num_threads];
  wfs_fetch::start(num_threads,t,NULL);
}

void
wfs_prefetch::fetcher() {

  //tag of the current request we are working on
  req_tag tag;
  //information for the current request
  prefetch_req *req = NULL;

  //keep track of what I am currently working on
  thread_info *cur = new thread_info();

  while(1) {
    pthread_mutex_lock(&_q_lock);
    //decrease the semaphore when we are done with a request
    if (req)
      _req_window--;
    while(_queue.empty()||(_req_window >= _req_window_size)) {
      pthread_cond_wait(&_q_wait,&_q_lock);
    }
    //grab the first thing off of the queue
    tag = _queue.front();
    _queue.pop();
    //increase the semaphore while we are running a request
    _req_window++;

    pthread_mutex_unlock(&_q_lock);

    //check if the request is still pending
    //by trying to look up the request info
    //the only time that the request might be gone is
    //if an application thread took the prefetching request
    pthread_mutex_lock(&_p_lock);
    if( _pending.count(tag) > 0 ) {
      //found the request and it's still pending
      req = (prefetch_req*)_pending[tag];
      pthread_mutex_lock(&(cur->lock));
      cur->working_req = tag;
      req->thread = cur;
      pthread_mutex_unlock(&(cur->lock));
      pthread_mutex_unlock(&_p_lock);

      fetch(req);

      pthread_mutex_lock(&_p_lock);
      delete req;
      _pending.erase(tag);
      pthread_mutex_unlock(&_p_lock);

      pthread_mutex_lock(&(cur->lock));
      cur->working_req = req_tag();
      pthread_cond_broadcast(&(cur->wait));
      pthread_mutex_unlock(&(cur->lock));
    } else {
      pthread_mutex_unlock(&_p_lock);
    }
  }
  delete cur;
}

bool
wfs_prefetch::lookup(req_tag tag) {
  bool r = false;
  pthread_mutex_lock(&_p_lock);
  if (_pending.count(tag) > 0) {
    if (((prefetch_req*)_pending[tag])->thread == NULL) {
      //if the thread info is null, no one has claimed it
      //so, remove it
      delete _pending[tag];
      _pending.erase(tag);
    } else {
      r = true;
    }
  }
  pthread_mutex_unlock(&_p_lock);
  return r;
}

void
wfs_prefetch::wait(req_tag tag) {
  pthread_mutex_lock(&_p_lock);
  if ((_pending.count(tag) > 0) && (((prefetch_req*)_pending[tag])->thread != NULL)) {
    thread_info *t = ((prefetch_req*)_pending[tag])->thread;
    pthread_mutex_unlock(&_p_lock);
    pthread_mutex_lock(&(t->lock));
    while (t->working_req == tag) {
      pthread_cond_wait(&(t->wait), &(t->lock));
    }
    wfs_log(WFS_DBG_2, "Done waiting for prefetching request to finish\n");
    pthread_mutex_unlock(&(t->lock));
  } else {
    pthread_mutex_unlock(&_p_lock);
  }
}

wfs_fileprefetch::wfs_fileprefetch(wfs_cache *cache, wfs_cache_client *cc,
				   wfs_node me, slice_manager *slices,
				   vivaldi *viv, wfs_client_manager *clients)
  : wfs_prefetch(cache, cc, me, slices, viv, clients) {

}

void
wfs_fileprefetch::fetch(prefetch_req *req) {
  if (!_cache->read_block(req->fh, req->version, req->bno)) {
    std::string buf;
    wfs_protocol::status status;

#if RATE_CONTROLLED
    timeval start;
    timeval end;
    timeval diff;
    unsigned int prev_window = _req_window_size;
    unsigned int prev_total = _total_reqs;
#endif

    wfs_log(WFS_DBG_1, "Prefetch: About to read uncached block %d for file %016"PRIx64",%d, size=%d\n",
	     req->bno, req->fh.fid, req->version.version_num,
	     WFS_BLOCK_SIZE);

#if RATE_CONTROLLED
    gettimeofday(&start,NULL);
#endif

    status = _cc->read_uncached_data(req->fh, req->version,
				     (req->bno)*WFS_BLOCK_SIZE,
				     WFS_BLOCK_SIZE,
				     req->user, buf);
#if RATE_CONTROLLED
    gettimeofday(&end,NULL);
    timersub(&end,&start,&diff);
    double latency = (diff.tv_sec*1000000+diff.tv_usec)/1000000.0f;
#endif

    if (status == wfs_protocol::OK) {
      wfs_log(WFS_DBG_1, "Prefetch: Storing block %d to cache for file %016"PRIx64",%d, size=%d\n",
	       req->bno, req->fh.fid, req->version.version_num,
	       WFS_BLOCK_SIZE);

      _cache->store_block(req->fh, req->version, req->bno, buf);
      _cache->apply_buf_writes(req->fh, req->version, req->bno, buf);

#if RATE_CONTROLLED
      //adjust the window size
      double throughput = (prev_window*.1)/latency;
      pthread_mutex_lock(&_win_lock);
      if (_avg_latencies[prev_window-1] == 0) {
        _avg_latencies[prev_window-1] = throughput;
      } else {
        _avg_latencies[prev_window-1] = (_avg_latencies[prev_window-1]*.75+throughput*.25);
      }

      double avg = 0;
      for (int i = 0; i < prev_window-1; i++) {
        avg+=_avg_latencies[i];
      }
      avg/=(prev_window-1);
      if ((_req_window_size < (STEP_UP+1)) ||
          (throughput > avg)) {
        if (_req_window_size+(STEP_UP) <= threadpool_size) {
          pthread_cond_broadcast(&_q_wait);
          _req_window_size+=STEP_UP;
        }
      } else {
        if (_req_window_size/SCALE_DOWN >=1)
          _req_window_size/=SCALE_DOWN;
      }
      _total_reqs++;
      pthread_mutex_unlock(&_win_lock);
#endif
    }
  }
}

bool
wfs_fileprefetch::readahead(wfs_file_handle fh,
			 wfs_version version,
			 std::string user,
			 unsigned int first_block,
			 unsigned int num_blocks) {
  unsigned int num_fetched_blocks = num_blocks;
  first_block += num_blocks;

  wfs_protocol::fileinfo fin;
  bool r = false;

  //grab the file's info out of the cache
  if (!_cache->read_md(fh, fin)) {
    //should be there, otherwise we just won't prefetch
    return r;
  }

  //check the size of the file
  if (fin.size <= (first_block+num_fetched_blocks)*WFS_BLOCK_SIZE) {
    //reduce the number of blocks we prefetch if the file isn't that big
    calc_block_no(first_block*WFS_BLOCK_SIZE, fin.size, num_fetched_blocks);
  }

  //blocks to be prefetched
  std::vector<req_tag> to_enqueue;
  req_tag tag;
  prefetch_req *info;
  for (unsigned int i = first_block; i < first_block+num_fetched_blocks; i++) {
    //check the cache to see if we need to prefetch
    if (!_cache->read_block(fh, version, i)) {
      //make a tag for the request
      tag = req_tag(fh.fid, i);
      //add the request
      pthread_mutex_lock(&_p_lock);
      if (_pending.count(tag) == 0) {
        wfs_log(WFS_DBG_2, "Prefetch: Adding prefetching request for block %d\n",i);
        _pending[tag] = new prefetch_req(fh, i, version, user);;
        to_enqueue.push_back(tag);
	r = true;
      }
      pthread_mutex_unlock(&_p_lock);
    }
  }

  //if file is popular, prefetch in a random order as well
  if (fh.get_cue( WFS_CUE_POPULARITY ).val == WFS_CUE_HOTSPOT) {
    for (int n = to_enqueue.size(); n > 0; n--) {
      int k = (random() >> 4) % n;
      tag = to_enqueue[n-1];
      to_enqueue[n-1] = to_enqueue[k];
      to_enqueue[k] = tag;
    }
  }

  //actually queue up the requests in the proper order
  pthread_mutex_lock(&_q_lock);
  for(std::vector<req_tag>::iterator i = to_enqueue.begin();
      i != to_enqueue.end(); i++) {
    _queue.push(*i);
  }
  pthread_cond_broadcast(&_q_wait);
  pthread_mutex_unlock(&_q_lock);
  return r;
}

bool
wfs_fileprefetch::wholefile(wfs_file_handle fh,
			     wfs_version version,
			     std::string user,
			     unsigned int first_block,
			     unsigned int num_blocks) {
  wfs_protocol::fileinfo fin;
  bool r = false;

  //grab the file's info out of the cache
  if (!_cache->read_md(fh, fin)) {
    //should be there, otherwise we just won't prefetch
    return r;
  }

  //figure out number of blocks in the file
  unsigned int total_blocks;
  calc_block_no(0, fin.size, total_blocks);


  std::vector<req_tag> to_enqueue;
  req_tag tag;
  prefetch_req *info;
  for (unsigned int i = 0; i < total_blocks; i++) {
    if ((i < first_block || (i >= first_block+num_blocks)) &&
	(!_cache->read_block(fh, version, i))) {
      tag = req_tag(fh.fid, i);
      pthread_mutex_lock(&_p_lock);
      if (_pending.count(tag) == 0) {
        wfs_log(WFS_DBG_2, "Prefetch: Adding prefetching request for block %d\n",i);
        _pending[tag] = new prefetch_req(fh,i,version,user);
        to_enqueue.push_back(tag);
        r = true;
      }
      pthread_mutex_unlock(&_p_lock);
    }
  }

  //if file is popular, prefetch in a random order as well
  if (fh.get_cue( WFS_CUE_POPULARITY ).val == WFS_CUE_HOTSPOT) {
    for (int n = to_enqueue.size(); n > 0; n--) {
      int k = (random() >> 4) % n;
      tag = to_enqueue[n-1];
      to_enqueue[n-1] = to_enqueue[k];
      to_enqueue[k] = tag;
    }
  }
  //actually queue up the requests in the proper order
  pthread_mutex_lock(&_q_lock);
  for(std::vector<req_tag>::iterator i = to_enqueue.begin();
      i != to_enqueue.end(); i++) {
    _queue.push(*i);
  }
  pthread_cond_broadcast(&_q_wait);
  pthread_mutex_unlock(&_q_lock);
  return r;
}

bool
wfs_fileprefetch::lookup(wfs_file_handle fh, unsigned int bno) {
  return wfs_prefetch::lookup(req_tag(fh.fid,bno));
}

void
wfs_fileprefetch::wait(wfs_file_handle fh, unsigned int bno) {
  return wfs_prefetch::wait(req_tag(fh.fid,bno));
}

wfs_dirprefetch::wfs_dirprefetch(wfs_cache *cache, wfs_cache_client *cc,
				 wfs_lease_manager *leases, wfs_node me,
				 slice_manager *slices, vivaldi *viv,
				 wfs_client_manager *clients)
  : wfs_prefetch(cache, cc, me, slices, viv, clients) {

  _leases = leases;
}

wfs_dirprefetch::~wfs_dirprefetch() {
}

void
wfs_dirprefetch::fetch(prefetch_req *req) {
  wfs_protocol::fileinfo fin;

#if RATE_CONTROLLED
  timeval start;
  timeval end;
  timeval diff;
  unsigned int prev_window = _req_window_size;
  unsigned int prev_total = _total_reqs;
#endif

  if (!_cache->read_md(req->fh, fin)) {
    wfs_protocol::status status;
    wfs_log(WFS_DBG_1, "Prefetching: About to read uncached meta-data for file %016"PRIx64" version %d\n",
	     req->fh.fid, req->fh.version.version_num);

    _leases->reg_pending_lease(req->fh);


    if (wfs_isdir(req->fh.fid)) {
      wfs_protocol::dirinfo din;
      status = _c.getdir(req->fh, req->user, din, NULL);
      fin.perm_cues = din.perm_cues;
      fin.permissions = din.permissions;
      fin.size = din.size;
      fin.atime = din.atime;
      fin.ctime = din.ctime;
      fin.mtime = din.mtime;
      fin.version = wfs_version();
    } else {

#if RATE_CONTROLLED
      gettimeofday(&start,NULL);
#endif

      status = _c.getfile(req->fh, req->user, fin, NULL);

#if RATE_CONTROLLED
      gettimeofday(&end,NULL);
      timersub(&end,&start,&diff);
      double latency = (diff.tv_sec*1000000+diff.tv_usec)/1000000.0f;

      //if we made a successful request to the server, adjust the
      //window size
      if (status == wfs_protocol::OK || wfs_protocol::OK_FLUSH) {
	//adjust the window size
	double throughput = (prev_window*.1)/latency;
	pthread_mutex_lock(&_win_lock);
	if (_avg_latencies[prev_window-1] == 0) {
	  _avg_latencies[prev_window-1] = throughput;
	} else {
	  _avg_latencies[prev_window-1] = (_avg_latencies[prev_window-1]*.75+throughput*.25);
	}

	double avg = 0;
	for (int i = 0; i < prev_window-1; i++) {
	  avg+=_avg_latencies[i];
	}
	avg/=(prev_window-1);
	if ((_req_window_size < (STEP_UP+1)) ||
          (throughput > avg)) {
	  if (_req_window_size+(STEP_UP) <= threadpool_size) {
          pthread_cond_broadcast(&_q_wait);
          _req_window_size+=STEP_UP;
	  }
	} else {
	  if (_req_window_size/SCALE_DOWN >=1)
	    _req_window_size/=SCALE_DOWN;
	}
	_total_reqs++;
	pthread_mutex_unlock(&_win_lock);
      }
#endif

    }
    bool lease_ok = _leases->unreg_pending_lease(req->fh);

    if (status == wfs_protocol::OK_FLUSH) {
      _cache->clear_md(req->fh, fin.version);
    }

    if (lease_ok && !req->fh.is_version &&
	(status == wfs_protocol::OK || wfs_protocol::OK_FLUSH)) {
      wfs_node file_node = _c.get_primary( file_handle_to_file_id(req->fh) );
      _leases->add_lease( req->fh, file_node, fhlease(req->fh) );
      wfs_protocol::fileinfo xfin;
      if(!(_cache->read_md( req->fh, xfin ) && xfin.version == fin.version )) {

        wfs_log(WFS_DBG_1, "Storing meta-data to cache for file %016"PRIx64"\n",
            req->fh.fid);
        _cache->store_md(req->fh, fin);
      }
    }
  }
}

bool
wfs_dirprefetch::directory(wfs_file_handle fh,
	 std::vector<wfs_storage::dirent> dirs,
	 std::string user) {

  std::vector<req_tag> to_enqueue;
  req_tag tag;
  prefetch_req *info;
  //not used, just for checking if the dir entry is in the cache
  wfs_protocol::fileinfo fin;
  bool r = false;

  for(std::vector<wfs_storage::dirent>::iterator i = dirs.begin();
      i != dirs.end(); i++ ) {
    //check if the meta-data is already in the cache
    if (!_cache->read_md(i->inum, fin)) {
      tag = req_tag(i->inum.fid);
      pthread_mutex_lock(&_p_lock);
      if (_pending.count(tag) == 0) {
        wfs_log(WFS_DBG_2, "Adding prefetching request for meta-data for directory %s\n",
                 i->name.c_str());
        _pending[tag] = new prefetch_req(i->inum, user);
        to_enqueue.push_back(tag);
        r = true;
      }
      pthread_mutex_unlock(&_p_lock);
    } else {
      wfs_log(WFS_DBG_2, "Meta-data for %s already in cache\n", i->name.c_str());
    }
  }

  if (fh.get_cue( WFS_CUE_POPULARITY ).val == WFS_CUE_HOTSPOT) {
    for (unsigned int n = to_enqueue.size(); n > 0; n--) {
      unsigned int k = (random() >> 4) % n;
      tag = to_enqueue[n-1];
      to_enqueue[n-1] = to_enqueue[k];
      to_enqueue[k] = tag;
    }
  }

  pthread_mutex_lock(&_q_lock);
  for(std::vector<req_tag>::iterator i = to_enqueue.begin();
      i != to_enqueue.end(); i++) {
    _queue.push(*i);
  }
  pthread_cond_broadcast(&_q_wait);
  pthread_mutex_unlock(&_q_lock);
  return r;
}

bool
wfs_dirprefetch::lookup(wfs_file_handle fh) {
  return wfs_prefetch::lookup(req_tag(fh.fid));
}

void
wfs_dirprefetch::wait(wfs_file_handle fh) {
  wfs_prefetch::wait(req_tag(fh.fid));
}
