/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


//prefetching for files and directories

#ifndef __WFS_PREFETCH_H__
#define __WFS_PREFETCH_H__ 1

#include <ext/hash_map>
#include <queue>
#include <vector>
#include "wfs_fetch.h"
#include "wfs_client.h"

//using __gnu_cxx::hash;
using __gnu_cxx::hash_map;

#define RATE_CONTROLLED 0

class wfs_cache_client;

//base abstract class for prefetching
class wfs_prefetch : public wfs_fetch {
 protected: 
  int threadpool_size;
  //semaphore that adjusts how many requests in-flight at a time
  //counts up to _req_window_size rather than down to zero
  //i.e. the thread increments this when it starts a request rather
  //than decrementing
  unsigned int _req_window;
  //limits how many requests can be in-flight at a time
  unsigned int _req_window_size;
  
  #if RATE_CONTROLLED
  //lock that protects the request window size 
  //this is so that each thread can adjust the window size after
  //finishing a request
  pthread_mutex_t _win_lock;
  //avg latency for a request
  double _avg_latencies[100];
  unsigned int _total_reqs;
  static const unsigned int STEP_UP = 1;
  static const float SCALE_DOWN = 2;
  #endif
  
  //info on what one thread is currently working on
  //wait for clients that need what the thread is working on
  struct thread_info {
    //must track what the current request for the thread is,
    //so that threads that are waiting have something to check
    //when they wake up
    req_tag working_req;
    pthread_mutex_t lock;
    pthread_cond_t wait;
    
    thread_info() {
      assert(pthread_mutex_init(&lock, NULL) == 0);
      assert(pthread_cond_init(&wait, NULL) == 0);
    };
    
    ~thread_info() {
      pthread_mutex_destroy(&lock);
      pthread_cond_destroy(&wait);
    };
  };
  
  struct prefetch_req : req_info {
    thread_info *thread;

    prefetch_req(wfs_file_handle _fh,
		 unsigned int _bno,
		 wfs_version _version,
		 std::string _user) 
      : req_info(_fh, _bno, _version, _user) {
      thread = NULL;
    };
    
  prefetch_req(wfs_file_handle _fh,
	       std::string _user)
    : req_info(_fh, _user) {
      thread = NULL;
    };
  };
  
  //function that actually calls the server 
  //filled in based on what we are prefetching
  //returns true if something is successfully fetched from the server
  virtual void fetch(prefetch_req *) = 0;
  
 public:
  
  wfs_prefetch(wfs_cache*, wfs_cache_client*, wfs_node, slice_manager*, vivaldi*, 
	    wfs_client_manager*); 
  
  virtual ~wfs_prefetch();
  virtual void fetcher();
  void start(int);
  bool lookup(req_tag);
  void wait(req_tag);
};

//class that performs file prefetching. Prefetches blocks from the servers 
//asynchronously
class wfs_fileprefetch : public wfs_prefetch {
 protected:
  virtual void fetch(prefetch_req *);
  
 public:
  wfs_fileprefetch(wfs_cache*, wfs_cache_client*, wfs_node, slice_manager*, vivaldi*, 
		   wfs_client_manager*); 
  
  virtual ~wfs_fileprefetch() {};

  //add one block to the queue for every block requested
  bool readahead(wfs_file_handle, wfs_version, std::string, unsigned int, unsigned int);
  //add all of the blocks of this file to be prefetched
  //returns true if any blocks where added to the prefetching queue
  bool wholefile(wfs_file_handle, wfs_version, std::string, unsigned int, unsigned int);
    
  bool lookup(wfs_file_handle, unsigned int);

  void wait(wfs_file_handle, unsigned int);
};

//class that performs directory prefetching
class wfs_dirprefetch : public wfs_prefetch {
 protected:
  wfs_lease_manager *_leases;
    
  //function that actually calls the server and fetches blocks
  virtual void fetch(prefetch_req *);
  
 public:
  wfs_dirprefetch(wfs_cache*, wfs_cache_client*, wfs_lease_manager*, wfs_node, 
		  slice_manager*, vivaldi*, wfs_client_manager*);

  virtual ~wfs_dirprefetch();

  //check if the block is currently being processsed
  //if the request is currently being processed, then wait for it to finish
  //returns true if the block was fetched by the prefetcher
  //side effect: if the block is in the queue, but not currently being
  //processed, remove from queue
  bool directory(wfs_file_handle, std::vector<wfs_storage::dirent>, std::string);
  //check if a file meta-data request is currently being processed
  //returns true if the meta-data was fetched by the prefetcher
  //if the request is currently being processed, then wait for it to finish
  //side effect: if the meta-data request is in the queue but not currently 
  //being processed, remove from queue
  
  //lookup whether the meta-data for this file is currently being prefetched
  bool lookup(wfs_file_handle);
  //wait for the prefetcher to finish fetching the meta-data for this file
  void wait(wfs_file_handle);
};

#endif
