/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <chash.h>
#include <dirent.h>
#include <fcntl.h>
#include <map>
#include <method_thread.h>
#include <set>
#include <sys/stat.h>
#include <wfs_file.h>
#include <wfs_misc.h>
#include <wfs_protocol.h>
#include <wfs_metadata.h>

#include "wfs_maint_slices.h"

wfs_maint_slices::wfs_maint_slices( wfs_node me, rpcs *server,
				    std::string dir,
				    wfs_perfile_locker *locker,
				    wfs_storage *storage,
				    wfs_client_manager *clients,
				    int maint_timer ) :
  wfs_maint_dirdisk(me, server, dir, locker, storage, clients, maint_timer) {
  _fh_needs_cues = false;

  server->reg( wfs_maint_protocol::acquire_lock, this,
	       &wfs_maint_slices::acquire_lock );
  server->reg( wfs_maint_protocol::md_exchange_slice, this,
	       &wfs_maint_slices::exchange_metadata_slice );

}

wfs_maint_slices::~wfs_maint_slices() {
}

void wfs_maint_slices::get_replicas( wfs_file_handle fh,
				     std::vector<wfs_node> &reps ) {

  std::vector<host> hreps;
  int rep_num;
  wfs_slice_id sid = wfs_slice::get_slice_id(fh.fid);

  slice_manager::ret_code r =
    _slices->get_slice_replicaset( sid, hreps, rep_num );

  if( r != slice_manager::SM_SUCCESS ) {
    return;
  }

  for( int i = 0; i < rep_num && i < (int)hreps.size(); i++ ) {
    reps.push_back( make_node( hreps[i] ) );
  }

  if( _tmp_replicas.find(sid) != _tmp_replicas.end() ) {
    reps.push_back( _tmp_replicas[sid] );
  }

}

wfs_node wfs_maint_slices::node_from_id( wfs_node_id node_id ) {
  // brute force search through everyone
  std::vector<host> hosts;
  _slices->get_all_active_hosts(hosts);
  for( unsigned i = 0; i < hosts.size(); i++ ) {
    wfs_node node = make_node(hosts[i]);
    if( node_id == node.id ) {
      return node;
    }
  }
  wfs_node x;
  x.id = 0;
  return x;
}

int wfs_maint_slices::exchange_metadata_slice( wfs_node_id node_id, wfs_slice_id slice,
					       std::vector<wfs_maint_protocol::metadata> in_md,
					       std::vector<wfs_maint_protocol::metadata> &out_md ) {


  bool erase = false;
  if( _tmp_replicas.find(slice) == _tmp_replicas.end() ) {
    _tmp_replicas[slice] = _me;
    erase = true;
  }
  build_metadata_list( node_id, out_md );
  // fork off a thread for this, so we don't block the response?
  compare_metadata( node_from_id(node_id), out_md, in_md );

  if(erase) {
    _tmp_replicas.erase(slice);
  }

  return wfs_maint_protocol::OK;

}

// TODO: this duplicates some functionality from the parent class, refactor
// to be less dumb
bool wfs_maint_slices::do_metadata_exchange_slice( wfs_node node,
						   wfs_slice_id slice ) {

  // ask pred for all the meta-data it knows about between his id and
  // my own, if I'm within in the replication level for that data.
  // If the versions don't match, put the data in a queue somewhere to
  // reconcile at some later date

  std::vector<wfs_maint_protocol::metadata> my_md, node_md;
  build_metadata_list( node.id, my_md );

  rpcc *client = _clients->get_client( node, rpcc::to(5000) );
  if( client == NULL ) {
    // TODO: better timeout???
    return false;
  } else {
    wfs_maint_protocol::status s = client->call(wfs_maint_protocol::md_exchange_slice,
						_me.id, slice, my_md, node_md,
						rpcc::to(5000));
    _clients->done_with_client(node);
    //+xavid
    if( s != wfs_maint_protocol::OK ) {
      return false;
    }
  }

  my_md.clear();
  // rebuild list, the old one could already be out of date
  build_metadata_list( node.id, my_md );

  compare_metadata( node, my_md, node_md );

  return true;

}

int wfs_maint_slices::acquire_lock( wfs_slice_id sid, std::vector<host> hreps,
				    int &dummy ) {

  wfs_log(WFS_DBG_1, "%016llx: maintenance -- acquiring lock %016"PRIx64", %d reps\n",
	    _me.id, sid, hreps.size() );

  rep_policy pol;
  slice_manager::ret_code r =_slices->get_policy_for_slice(sid, pol);
  if( r == slice_manager::SM_SUCCESS ) {
    wfs_file_id fid;
    host tryhost;
    _slices->get_new_fid( pol, fid, tryhost, true, sid, hreps );
  }
  return wfs_maint_protocol::OK;

}

void wfs_maint_slices::slices_maint_loop() {

  // give other nodes a chance to register their procs
  sleep(2);

  // first see what slices we have on disk -- we might need to
  // become their primary
  // TODO: remember the last epoch number for the slices, and only
  // become primary if nothing has changed since the last time we were a
  // replica
  DIR *dir = opendir( _server_dir.c_str() );
  if( dir != NULL ) {
    struct dirent *entry;
    while( (entry = readdir(dir)) ) {
      wfs_file_id fid;
      uint32_t fid_1;
      uint32_t fid_2;
      char dummy[20];
      // ignore any dot extensions of a file
      if( sscanf( entry->d_name, "%08"SCNx32"%08"SCNx32".%s",
          &fid_1, &fid_2, dummy ) != 3 &&
          sscanf( entry->d_name, "%08"SCNx32"%08"SCNx32,
              &fid_1, &fid_2 ) == 2 ) {
        fid = two_ints_to_file_id( fid_1, fid_2 );
        _possible_locked_slices.insert( wfs_slice::get_slice_id(fid) );
      }
    }
  }
  closedir(dir);

  // periodically contact predecessors to exchange meta-data and
  // place necessary repairs onto a queue

  while(1) {

    // make sure all the slices have the right set of primaries
    _slices->sync_state();
    std::list<std::pair<wfs_slice_id,host> > repairs;
    slice_manager::ret_code r = _slices->check_all_slices(repairs);
    assert( r == slice_manager::SM_SUCCESS );
    std::list<std::pair<wfs_slice_id,host> >::const_iterator i;
    for( i = repairs.begin(); i != repairs.end(); i++ ) {

      wfs_slice_id slice_id = i->first;
      wfs_node node = make_node( i->second );

      wfs_log(WFS_DBG_1, "Maintenance: told to recruit %016"PRIx64" for slice %016"PRIx64"\n",
          node.id, slice_id );

      if( node.id == _me.id ) {
        continue;
      }

      _tmp_replicas[slice_id] = node;
      wfs_log(WFS_DBG_1, "%016"PRIx64": exchanging with %016"PRIx64"\n", _me.id, node.id );
      // TODO: should probably stop the world for this slice during the
      // exchange
      if( !do_metadata_exchange_slice( node, slice_id ) ) {
        r = _slices->sync_aborted( slice_id, i->second );
        wfs_log(WFS_DBG_1, "Node %016"PRIx64" is down -- reporting\n", node.id );
        r = _slices->report_host_down(i->second);
        wfs_log(WFS_DBG_1, "Report %016"PRIx64" down status = %d\n", node.id, r );
      } else {
        std::vector<host> dropped; // TODO: WTF?
        bool do_lock;

        // save the old replica set
        std::vector<host> hreps;
        int rep_num;
        r = _slices->get_slice_replicaset( slice_id, hreps, rep_num );

        r = _slices->report_sync_complete( slice_id, i->second, dropped,
            do_lock );
        // It's possible that the renew operation within the
        // report_sync_complete will run up against stale data,
        // and if so we'll get an error back here, but I think it's harmless.
        //assert( r == slice_manager::SM_SUCCESS );

        // inform the other guy
        if( do_lock ) {
          rpcc *client = _clients->get_client(node);
          if( client != NULL ) {
            int dummy;

            // make sure the other guy uses the same backups
            wfs_log(WFS_DBG_OFF, "Telling %016"PRIx64" to acquire lock on slice "
                "%016"PRIx64" with %d existing replicas\n",
                node.id, slice_id, hreps.size() );
            client->call( wfs_maint_protocol::acquire_lock, slice_id, hreps,
                dummy, rpcc::to(5000) );
            _clients->done_with_client(node);

          }
        }

      }
      _tmp_replicas.erase(slice_id);

    }

    std::vector<wfs_slice_id> maint_slices;
    _slices->get_maintenence_slices(maint_slices);

    // If we're not in any slices, we must have been disconnected for a while
    // Make sure other people know about us by making a new slice
    if( maint_slices.size() == 0 ) {
      wfs_log(WFS_DBG_1, "Not part of any slices, so making one.\n" );
      rep_policy pol( WFS_REP_LEVEL_DEFAULT, rep_policy::STRICT, _me.site_id,
          WFS_REP_LEVEL_DEFAULT );
      wfs_file_id fid;
      host tryhost;
      _slices->get_new_fid( pol, fid, tryhost, true );
    }

    std::set<wfs_node_id> maintained;
    for( unsigned i = 0; i < maint_slices.size(); i++ ) {

      _possible_locked_slices.insert( maint_slices[i] );
      std::vector<host> slice_repset;
      int repnum;
      r = _slices->get_slice_replicaset( maint_slices[i], slice_repset,
          repnum );
      assert( r == slice_manager::SM_SUCCESS );

      for( int j = 0; j < repnum && j < (int)slice_repset.size(); j++ ) {

        wfs_node node = make_node( slice_repset[j] );
        if( node.id == _me.id ) {
          continue;
        }
        // only need to exchange with the other node once per loop,
        // not once per slice per loop
        if( maintained.find(node.id) != maintained.end() ) {
          continue;
        }
        maintained.insert(node.id);
        wfs_log(WFS_DBG_1, "%016"PRIx64": exchanging with %016"PRIx64" for slice %016"PRIx64"\n",
            _me.id, node.id , maint_slices[i] );
        if( !do_metadata_exchange( node ) ) {
          wfs_log(WFS_DBG_1, "Node %016"PRIx64" is down -- reporting\n", node.id );
          r = _slices->report_host_down( slice_repset[j] );
          wfs_log(WFS_DBG_1, "Report %016"PRIx64" down status = %d\n", node.id, r );
        }
      }
    }
    // now make sure that all possible locked slices with replevel = 1
    // are locked
    // TODO: fix to work for all possible replevels, but this requires making
    // sure there was no alternate configuration for the slice since we
    // were last a replica
    std::set<wfs_slice_id>::iterator j = _possible_locked_slices.begin();
    for( ; j != _possible_locked_slices.end(); j++ ) {
      wfs_slice_id sid = *j;
      host primary;
      r = _slices->get_slice_primary( sid, primary, true );
      if( r == slice_manager::SM_NOLOCKED ) {

        // lock it with the same policy as before
        rep_policy pol;
        r = _slices->get_policy_for_slice( sid, pol );
        if( r == slice_manager::SM_SUCCESS && pol.rep_num == 1 ) {

          wfs_log(WFS_DBG_1, "Found that slice %016"PRIx64
              " is unlocked, going to lock.\n", sid );

          wfs_file_id fid;
          host tryhost;
          _slices->get_new_fid( pol, fid, tryhost, true, sid );
        }
      }
    }

    sleep( MAINT_LOOP_TIMER );

  }

}

void wfs_maint_slices::start_maint( slice_manager *slices, bool do_maint ) {

  _slices = slices;
  _maint_started = true;

  if( do_maint ) {

    // TODO: somehow I need to get up-to-date with every node I should be
    // a primary for

    method_thread( this, true, &wfs_maint_slices::slices_maint_loop );

  }

  method_thread( (wfs_maint_dirdisk *) this, true,
		 &wfs_maint_dirdisk::repair_loop );

}
