/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "base_chan.h"
#include <signal.h>
#include "method_thread.h"
#include "logging/wfs_log.h"

abstract_channel::abstract_channel(pthread_mutex_t *delete_m,
				   bool freebuf, canceller *cancel)
   : th(0), isdead(false), _delete_m(delete_m),
     _freebuf(freebuf), _unsent(0), _cancel(cancel)
{
  nam[0] = '\0';
  signal(SIGPIPE, SIG_IGN);
  assert(pthread_mutex_init(&_unsent_m, NULL) == 0);
  th = method_thread(this, false, &abstract_channel::output_loop);
  //th = method_thread(this, true, &abstract_channel::output_loop);
}

abstract_channel::~abstract_channel() {
  assert(pthread_mutex_destroy(&_unsent_m) == 0);
}

bool
abstract_channel::outstanding() {
  bool x;
  assert(pthread_mutex_lock(&_unsent_m) == 0);
  x = (_unsent != 0);
  assert(pthread_mutex_unlock(&_unsent_m) == 0);
  return x;
}

// output thread
void
abstract_channel::output_loop()
{
  // abstract_channel::abstract_channel() also sets th but we might execute
  // first, and th must be set for die().
  th = pthread_self();
  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);
  //assert(pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldstate) == 0);

  while(1){
    const charbuf pdu = outq.deq();
    if(pdu.sz <= 0) {
      // spurious packet, probably caused by the mutex bug in fifo
      break;
    }
    send1(pdu);
    assert(pthread_mutex_lock(&_unsent_m) == 0);
    _unsent--;
    assert(pthread_mutex_unlock(&_unsent_m) == 0);
  }
}

void cb_buf_cleanup(void *arg) {
  charbuf *cb = (charbuf *) arg;
  if( cb != NULL ) {
    if (cb->free) {
      free(cb->buf);
    }
  }
}

void cb_ref_cleanup(void *arg) {
  charbuf *cb = (charbuf *) arg;
  if( cb != NULL ) {
    cb->decref(); // we must still hold the lock
  }
}

// output thread
void
abstract_channel::send1(const charbuf &pdu)
{
  if (pdu.requser != NULL) {
    if (!verify_user(pdu.requser)) {
      wfs_log(WFS_DBG_2, "send1 dropping packet meant for %s due to user verification failure\n",pdu.requser);
      goto end;
    }
  }

  {

    unsigned long len = htonl(pdu.sz);
    assert(len > 0);
    char *buf = pdu.buf;
    int size = pdu.sz;

    if( dead() ) { 
      goto end;
    }

    int iovcnt = 0;
    struct iovec iov[3];
    iov[iovcnt].iov_base = &len;
    iov[iovcnt++].iov_len = sizeof(len);
#ifdef WFS_RPC_CHECKSUMMING
    // TODO: network vs. host order
    unsigned long long checksum = pdu.checksum();
    iov[iovcnt].iov_base = &checksum;
    iov[iovcnt++].iov_len = sizeof(checksum);
#endif
    iov[iovcnt].iov_base = buf;
    iov[iovcnt++].iov_len = size;
    struct iovec *iovptr = iov;
    int iovlen;

    charbuf *cb_ref = NULL;
    if( _delete_m != NULL ) {
      pthread_mutex_lock( _delete_m );
      // This means the RPC timed out before we had a chance to even send
      // this message (unless free is set, in which case the message was sent
      // once and forgotten about)
      if( pdu.getref() <= 1 && !pdu.free ) {
        pthread_mutex_unlock( _delete_m );
        goto end;
      }
      cb_ref = (charbuf *) &pdu;
    }

    pthread_cleanup_push( &abstract_channel::cleanup_lock, (void *) _delete_m );
    pthread_cleanup_push( &cb_ref_cleanup, (void *) cb_ref );
    pthread_cleanup_push( &cb_buf_cleanup, (void *) &pdu );

    if (wfs_dbg_block(WFS_DBG_2)) {
      int clt_nonce = 0;
      clt_nonce = clt_nonce | (buf[0] & 0xff);
      clt_nonce = clt_nonce | ((buf[1] & 0xff) << 8);
      clt_nonce = clt_nonce | ((buf[2] & 0xff) << 16);
      clt_nonce = clt_nonce | ((buf[3] & 0xff) << 24);
      wfs_log(WFS_DBG_2, "Sending packet for clt_nonce %u and checksum %llu on socket %d\n",
          clt_nonce,
#ifdef WFS_RPC_CHECKSUMMING
          pdu.checksum(),
#else
          (unsigned long long) 0,
#endif
          channo());
    }

    iovlen = iovcnt;

    do {
      ssize_t written = xwritev( iovptr, iovcnt );

      if( written <= 0 ) {
        wfs_log(WFS_DBG_OFF, "bad send1 on socket %d, written=%d\n", channo(), written );
        perror("abstract_channel write");
        die();
        wfs_log(WFS_DBG_OFF, "abstract_channel::send1 die in loop write fd (%d) dead? %d\n", channo(), isdead?1:0);
        break;
      } else {

        // did we at least send the length successfully?
        if (iovcnt == iovlen && written >= (ssize_t)sizeof(len)) {
          iovcnt--;
          iovptr++;
          written -= sizeof(len);
        }

#ifdef WFS_RPC_CHECKSUMMING
        if (iovcnt == (iovlen-1) && written >= (ssize_t)sizeof(checksum)) {
          iovcnt--;
          iovptr++;
          written -= sizeof(checksum);
        }
#endif

        iov[iovlen-1].iov_base = ((char *) iov[iovlen-1].iov_base) + written;
        iov[iovlen-1].iov_len -= written;
      }

    } while( iov[iovlen-1].iov_len > 0 );

    pthread_cleanup_pop(0);
    pthread_cleanup_pop(0);
    pthread_cleanup_pop(0);
    if( _delete_m != NULL ) {
      pthread_mutex_unlock( _delete_m );
    }

  }

  end:

  if (pdu.free) {
    free(pdu.buf);
  }
  if( _delete_m != NULL ) {
    pthread_mutex_lock( _delete_m );
    pdu.decref();
    pthread_mutex_unlock( _delete_m );
  }

}

// non-blocking (except for connect)
void
abstract_channel::send(const charbuf &pdu)
{
  assert(pthread_mutex_lock(&_unsent_m) == 0);
  _unsent++;
  assert(pthread_mutex_unlock(&_unsent_m) == 0);
  outq.enq(pdu);
}


// blocks and reads from the tcp/ssh socket.
// called inside cchan or schan mutex
int
#ifdef WFS_WITH_LIBSSH
abstract_channel::recv(charbuf &cb, ssh_cleanup_frame **frames)
#else
abstract_channel::recv(charbuf &cb)
#endif
{
  unsigned long len;

#ifdef WFS_WITH_LIBSSH
  if(xread(&len, sizeof(len), frames) != 0)
#else
  if(xread(&len, sizeof(len)) != 0)
#endif
    return 0;
  len = ntohl(len);
  if(len > 10*1024*1024){
    wfs_log(WFS_DBG_OFF, "abstract_channel:recv len %lu too big\n", len);
    die();
    return 0;
  }
#ifdef WFS_RPC_CHECKSUMMING
  unsigned long long checksum;
#ifdef WFS_WITH_LIBSSH
  if(xread(&checksum, sizeof(checksum), frames) != 0) {
#else
  if(xread(&checksum, sizeof(checksum)) != 0) {
#endif
    wfs_log(WFS_DBG_OFF, "CHECKSUM XREAD FAILED\n" );
    return 0;
  } else {
    wfs_log(WFS_DBG_2, "Read checksum %llu\n", checksum);
  }
#endif
  char *p = (char *) malloc(sizeof(char)*len);
  int r;
  _cancel->cleanup_push( &free, p );
  assert(p);
#ifdef WFS_WITH_LIBSSH
  r = xread(p, len, frames);
#else
  r = xread(p, len);
#endif
  _cancel->cleanup_pop( &free, p );
  if(r != 0){
    free(p);
    return 0;
  }
  cb.buf = p;
  cb.sz = len;
#ifdef WFS_RPC_CHECKSUMMING
  if (checksum != cb.checksum()) {
    wfs_log(WFS_DBG_OFF,
	    "CHECKSUM DOESN'T MATCH -- %llu (stated) vs. %llu (actual) socket %d\n",
	    checksum, cb.checksum(), s);
  }
#endif

  return cb.sz;
}

// set isdead, so that this abstract_channel and its associated output thread
// will go away.  the thread that reads from this abstract_channel (owned by
// schan/cchan) will see a failure because s has been closed, and then
// it will clean up this abstract_channel.  when the thread calls the
// destructor, the destructor will terminate the output thread
// associated with this abstract_channel.
void
abstract_channel::die()
{
  wfs_log(WFS_DBG_2, "%d: die\n", channo());
  isdead = true;
  // only close the socket in the destructor, so it can't be reused again
  // until the channel is thoroughly deleted and state is cleaned up
}

void
cleanup_frames(void *a)
{
#ifdef WFS_WITH_LIBSSH
  ssh_cleanup_frame **frames = (ssh_cleanup_frame **) a;
  for( unsigned int i = 0; i < WFS_NUM_FRAMES; i++ ) {
    delete frames[i];
    frames[i] = NULL;
  }
  //delete [] frames;
#endif
}

