/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef base_chan_h
#define base_chan_h

#include <iostream>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#ifdef __linux__
#include <sys/signal.h>
#include <openssl/md5.h>
#elif defined(__APPLE__)
#include <openssl/md5.h>
#else
#include <md5.h>  // requires sys/types.h
#endif /*__linux__*/

#include "fifo.h"
#include "canceller.h"

#ifdef WFS_WITH_LIBSSH
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <libssh/libssh.h>
#endif

#if defined(__linux__) || defined(__APPLE__)
unsigned long long md5_to_long( unsigned char *md5_result );
#else
unsigned long long md5_to_long( char *md5_result );
#endif

// base classes for both normal tcpchans and sshchans, 
// including data structures
// used by both

struct charbuf {
charbuf(char *b, int s, const char *_requser=NULL) : buf(b), sz(s), refcount(NULL), free(false), requser(_requser) {}
charbuf():buf(NULL),sz(0),refcount(NULL),free(false),requser(NULL) {}
charbuf(const charbuf &x) : buf(x.buf), sz(x.sz), refcount(x.refcount), 
    free(x.free),requser(x.requser) {}
  charbuf dup() const {
    char *b = NULL;
    if (sz > 0) 
      b = (char *) malloc( sizeof(char)*sz );
    if (b)
      memcpy(b, buf, sz);
    return charbuf(b,sz);
  }
  void incref() const { 
    if (refcount != NULL ) {
      if( *refcount >= 0 ) {
	(*refcount)++;
      } else {
	(*refcount)--;
      }
    }
  }
  void decref( bool flip = false ) const { 
    if (refcount != NULL ) {
      if (abs(*refcount) == 1) {
	::free(refcount);
      } else if (*refcount >= 0) {
	(*refcount)--;
	if (flip) {
	  (*refcount) = -(*refcount);
	}
      } else {
	(*refcount)++;
      }
    }
  }
  int getref() const { 
    if (refcount != NULL ) {
      return (*refcount);
    } else {
      return 0;
    }
  }
  unsigned long long checksum() const {

    MD5_CTX md5_context;
#if defined(__linux__) || defined(__APPLE__)
    MD5_Init( &md5_context );
#else
    MD5Init( &md5_context );
#endif

#if defined(__linux__) || defined(__APPLE__)
    MD5_Update( &md5_context, (const unsigned char *) buf, sz );
#else
    MD5Update( &md5_context, (const unsigned char *) buf, sz );
#endif

#if defined(__linux__) || defined(__APPLE__)
    unsigned char md5_result_int[17];
    MD5_Final( md5_result_int, &md5_context);
    
    unsigned char md5_result[33];
    // have to convert the bytes manually to hex, because there's no MD5End
    char *curr_char = (char *) md5_result;
    for( uint i = 0; i < 16; i++ ) {
      sprintf( curr_char, "%02x", md5_result_int[i] );
      curr_char += 2;
    }
    
#else
    
    char md5_result[33];
    MD5End( &md5_context, md5_result );
    
#endif /*__linux__*/

    return md5_to_long( md5_result );

  }
  char *buf;
  int sz;
  // This is allocated and incremented directly only by people who 
  // know what they're doing, since otherwise we'd need a lock to protect it.
  int *refcount;
  bool free;
  const char *requser;
};

void cb_buf_cleanup(void *arg);
void cb_ref_cleanup(void *arg);

////////////////////

class abstract_channel {
 protected:
  
  pthread_t th;
  bool isdead; // tell owning schan or cchan to stop using this tcpchan
  pthread_mutex_t *_delete_m;
  bool _freebuf; //free buf after finished sending?
  int _unsent;
  pthread_mutex_t _unsent_m;
  canceller *_cancel;

  fifo<charbuf> outq;

  char nam[20];

  void output_loop();
  void send1(const charbuf &);

  virtual ssize_t xwritev(const struct iovec *iovptr, int iovcnt) = 0;
#ifdef WFS_WITH_LIBSSH
  virtual int xread(void *p, unsigned int n, ssh_cleanup_frame **frames) = 0;
#else
  virtual int xread(void *p, unsigned int n) = 0;
#endif

  static void cleanup_lock(void *arg) {
    pthread_mutex_t *l = (pthread_mutex_t *) arg;
    if (l != NULL) {
      pthread_mutex_unlock(l);
    }
  };

 public:
  abstract_channel(pthread_mutex_t *delete_m, bool freebuf,
		   canceller *cancel);
  virtual ~abstract_channel();

  // Called on the client side to connect to a remote server
#ifdef WFS_WITH_LIBSSH
  virtual int connect(const sockaddr_in &dst, ssh_cleanup_frame **frames) = 0;
#else
  virtual int connect(const sockaddr_in &dst) = 0;
#endif
  // Called on a newly-accepted connection on the server side
  virtual int start() {return 0;};

  void send(const charbuf &);
  bool outstanding();
#ifdef WFS_WITH_LIBSSH
  int recv(charbuf &cb, ssh_cleanup_frame **frames);
#else
  int recv(charbuf &cb);
#endif
  virtual bool dead() = 0;
  virtual void lose() = 0;
  void die();

  virtual void close() = 0;

  virtual const char *getpeername() = 0;
  virtual int channo() = 0;

  virtual bool verify_user(const char *username) = 0;
};
class abstract_bind {
 public:
  virtual ~abstract_bind() {};
  virtual abstract_channel *accept(canceller *cancel,
				   pthread_mutex_t *tcpdelete_m = NULL, 
				   bool freebuf = false) = 0;
  virtual void close() = 0;
  virtual unsigned int bound_port() = 0;
};

class chan_hub {
 public:
  virtual ~chan_hub() {};
  virtual abstract_channel *channel(canceller *cancel,
				    pthread_mutex_t *tcpdelete_m = NULL, 
				    bool freebuf = false) = 0;
  virtual abstract_bind *bind(const sockaddr_in &dst) = 0;
  virtual int initial_timeout() = 0;
};

#ifdef WFS_WITH_LIBSSH
#define setup_cleanup_handler(h,a)				\
  pthread_cleanup_push( h, (void *) a );			\
  {								\
    unsigned int i = 0;						\
    while( frames[i] != NULL && ++i < WFS_NUM_FRAMES ) {}	\
    assert( i < WFS_NUM_FRAMES );				\
    frames[i] = new ssh_cleanup_frame();			\
    frames[i]->handler = h;					\
    frames[i]->arg = a;						\
  }
#else
#define setup_cleanup_handler(h,a)		\
  pthread_cleanup_push( h, (void *) a );
#endif

#ifdef WFS_WITH_LIBSSH
#define pop_cleanup_handler(x)					\
  {								\
    unsigned int i = 0;						\
    while( frames[i] != NULL && ++i < WFS_NUM_FRAMES ) {}	\
    assert( i <= WFS_NUM_FRAMES && i > 0 );			\
    delete frames[i-1];						\
    frames[i-1] = NULL;						\
  }								\
  pthread_cleanup_pop(x);
#else
#define pop_cleanup_handler(x)		\
  pthread_cleanup_pop(x);
#endif

const unsigned int WFS_NUM_FRAMES = 6;

void cleanup_frames(void *a);

#ifdef WFS_WITH_LIBSSH
#define setup_cleanup_frames()						\
  ssh_cleanup_frame *frames[WFS_NUM_FRAMES];				\
  bzero( frames, WFS_NUM_FRAMES*sizeof(ssh_cleanup_frame *) );
#else
#define setup_cleanup_frames() ;	    
#endif

#endif
