/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef fifo_h
#define fifo_h 

// fifo template
// synchronized with mutex and cond

#include <assert.h>
#include <errno.h>
#include <list>
#include <sys/time.h>
#include <time.h>
#include <logging/wfs_log.h>

template<class T>
class fifo {
 public:
  fifo();
  ~fifo();
  void enq(T);
  T deq();
  void cleanup() {
    pthread_mutex_unlock(&m);
    waiting = false;
  }
 private:
  std::list<T> q;
  pthread_mutex_t m;
  pthread_cond_t c; // q went non-empty
  volatile bool waiting;
  volatile bool loop;
  bool wait_for_lock();
  static void cleanup_lock(void *arg) {
    fifo<T> *f = (fifo<T> *) arg;
    f->cleanup();
  };
};

template<class T>
fifo<T>::fifo() : waiting(false), loop(true)
{
  assert(pthread_mutex_init(&m, 0) == 0);
  assert(pthread_cond_init(&c, 0) == 0);
}

template<class T>
fifo<T>::~fifo()
{
  // make sure the thread calling recv has been fully cancelled
  if( !wait_for_lock() ) {
    wfs_log( WFS_DBG_1, "fifo memory leak!\n" );
    return;
  }
  int s;
  while((s = pthread_mutex_destroy(&m)) != 0 && errno == EAGAIN) {
#ifndef __APPLE__
    pthread_yield();
#else
    // pthread_yield isn't available on Mac
    pthread_yield_np();
#endif
  }
  if( s != 0 ) {
    perror( "FIFO errno" );
  }
  assert(s == 0);
  assert(pthread_cond_destroy(&c) == 0);
}

template<class T> bool
fifo<T>::wait_for_lock() {
  loop = false;
  int times = 0;
  while(waiting && times++ < 100000) {
#ifndef __APPLE__
    pthread_yield();
#else
    // pthread_yield isn't available on Mac
    pthread_yield_np();
#endif
  }
  return (times <= 100000);
}

template<class T> void
fifo<T>::enq(T e)
{
  assert(pthread_mutex_lock(&m) == 0);
  q.push_back(e);
  assert(pthread_cond_broadcast(&c) == 0);
  assert(pthread_mutex_unlock(&m) == 0);
}

template<class T> T
fifo<T>::deq()
{
  void *t = (void *) this;
  waiting = true;
  T e;
  // avoid crazy race condition I can't figure out
  if(pthread_mutex_lock(&m) != 0) {
    waiting = false;
    return e;
  }
  pthread_cleanup_push(&fifo::cleanup_lock, t);

  while(loop){
    if(q.empty()){
      assert (pthread_cond_wait(&c, &m) == 0);
    } else {
      e = q.front();
      q.pop_front();
      break;
    }
  }

  pthread_cleanup_pop(0);
  assert(pthread_mutex_unlock(&m) == 0);
  waiting = false;
  return e;
}

#endif
