/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef marshall_h
#define marshall_h

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

enum {
  DEFAULT_RPC_SZ = 65000,
  // yes these are equal, but for different reasons
  // (the reply header has coords, but the request header has nonces and such)
  RPC_HEADER_SZ = 20,
  RPC_REPLY_HEADER_SZ = 20
  // when there's no vivaldi, you can use:
  // RPC_REPLY_HEADER_SZ = 8
};

class marshall {
 private:
  char *_buf;
  int _capa;
  int _ind;
  int _sz;
  bool _del_buf;
 public:
 marshall(bool reply_header = false) : _del_buf(true) {
    _buf = (char *) malloc(sizeof(char)*DEFAULT_RPC_SZ);
    if (!_buf) {
      _capa = _ind = _sz = 0;
    } else { 
      _capa = DEFAULT_RPC_SZ;
      _ind = _sz = (reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ);
    }
  }
 marshall(char *b, int sz, bool reply_header = false) : 
  _buf(b), _capa(sz), _ind(reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ),
    _sz(reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ), _del_buf(true) {}
 marshall( const marshall& m ) : _buf(m._buf), _capa(m._capa), 
    _ind(m._ind), _sz(m._sz), _del_buf(m._del_buf) {}
  
  ~marshall() { if (_buf) free(_buf); }
  void zero_ind() { 
    _sz = _ind; //save the size of RPC body marshalled so far
    _ind = 0; 
  }
  void reset_ind() { _ind = _sz;}
  int ind() { return _ind; }
  int size() { return (_sz >_ind)?_sz:_ind; }
  char *cstr() { return _buf;}
  char *take_buf() {
    char *b = (char *) realloc(_buf, _sz);
    assert(b);
    _buf = NULL;
    _sz = _ind = 0;
    return b;
  }

  void rawbyte(unsigned char);
  void rawbytes(const char *, int);

  void dont_del_buf() { _del_buf = false; }
  bool del_buf() { return _del_buf; }
  
  // Return the current contents (including header) as a string
  const std::string str() const {
  	std::string tmps = std::string(_buf,_ind);
  	return tmps;
  }

};
marshall& operator<<(marshall &, unsigned int);
marshall& operator<<(marshall &, unsigned long);
marshall& operator<<(marshall &, int);
marshall& operator<<(marshall &, unsigned char);
marshall& operator<<(marshall &, char);
marshall& operator<<(marshall &, unsigned short);
marshall& operator<<(marshall &, short);
marshall& operator<<(marshall &, unsigned long long);
marshall& operator<<(marshall &, const std::string &);

// i32() &c do not directly signal an error.
// call ok() to check that all calls succeeded.
class unmarshall {
 private:
  char *_buf;
  int _sz;
  int _ind;
  bool _ok;
  bool _reply;
 public:
 unmarshall(bool reply_header = false): _buf(0), _sz(0), _ind(0), 
    _ok(false), _reply(reply_header) {}
 unmarshall(char *buf, int sz, bool reply_header = false):
  _buf(buf),_sz(sz),_ind(reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ),
    _reply(reply_header)
  {
    _ok = _sz >= (reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ)? true:false;
  }
  unmarshall(const std::string str, bool reply_header = false)
  {
  	_buf = (char *) malloc(sizeof(char)*str.size());
  	memcpy(_buf,str.data(), str.size());
  	_sz = str.size();
  	_ind = reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ;
    _reply = reply_header;
  	_ok = _sz >= (reply_header?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ)? true:false;
  }
  
  ~unmarshall() {
    if (_buf) free(_buf);
  }
  //take contents from another unmarshall object
  void take_in(unmarshall &another);
  bool ok() { return _ok; }
  bool okdone();
  unsigned int rawbyte();
  void rawbytes(std::string &s, unsigned int n);
  void zero_ind() { _ind = 0;}
  void reset_ind() { 
    _ind = (_reply?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ); 
    _ok = _sz >= (_reply?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ)? true:false;
  }
  int ind() { return _ind;}
  int size() { return _sz;}
  char *take_buf() {
    char *b = _buf;
    _buf = NULL;
    _sz = _ind = 0;
    _ok = false;
    return b;
  }
};
unmarshall& operator>>(unmarshall &, unsigned char &);
unmarshall& operator>>(unmarshall &, char &);
unmarshall& operator>>(unmarshall &, unsigned short &);
unmarshall& operator>>(unmarshall &, short &);
unmarshall& operator>>(unmarshall &, unsigned int &);
unmarshall& operator>>(unmarshall &, unsigned long &);
unmarshall& operator>>(unmarshall &, int &);
unmarshall& operator>>(unmarshall &, unsigned long long &);
unmarshall& operator>>(unmarshall &, std::string &);

template <class C> marshall &
operator<<(marshall &m, std::vector<C> v)
{
  m << (unsigned int) v.size();
  for(unsigned i = 0; i < v.size(); i++)
    m << v[i];
  return m;
}

template <class C> unmarshall &
operator>>(unmarshall &u, std::vector<C> &v)
{
  unsigned n;
  u >> n;
  for(unsigned i = 0; i < n; i++){
    C z;
    u >> z;
    v.push_back(z);
  }
  return u;
}

template <class A, class B> marshall &
operator<<(marshall &m, const std::map<A,B> &d) {
  typename std::map<A,B>::const_iterator i;

  m << (unsigned int) d.size();

  for (i = d.begin(); i != d.end(); i++) {
    m << i->first << i->second;
  }
  return m;
}

template <class A, class B> unmarshall &
operator>>(unmarshall &u, std::map<A,B> &d) {
  unsigned int n;
  u >> n;
  
  d.clear();

  for (unsigned int lcv = 0; lcv < n; lcv++) {
    A a;
    B b;
    u >> a >> b;
    d[a] = b;
  }
  return u;
}

#endif
