/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
    FUSE: Filesystem in Userspace
    Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.

    To run:
    mkdir /tmp/null_fuse_dir
    mkdir /tmp/null_fuse_mnt
    cd /tmp/null_fuse_dir
    <path>/null_fuse_ll /tmp/null_fuse_mnt

    Then you can access all the files in /tmp/null_fuse_dir through the
    /tmp/null_fuse_mnt mountpoint.

*/

// xavid: This uses a different version than the rest of the project
#undef FUSE_USE_VERSION
#define FUSE_USE_VERSION 26

#include <fuse_lowlevel.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <map>
#include <string>
#ifdef __linux__
#include <openssl/md5.h>
#elif defined(__APPLE__)
#include <openssl/md5.h>
#else
#include <md5.h>  // requires sys/types.h
#endif /*__linux__*/
#include <dirent.h>

static std::map< int, std::string > ino_to_path;
static std::map< std::string, int > linked_paths;

int path_to_ino( std::string path ) {

  if( linked_paths.find(path) != linked_paths.end() ) {
    return linked_paths[path];
  }

  //printf( "path_to_ino(%s)\n", path.c_str() );

  const unsigned char *addr_const = (unsigned char *) path.c_str();
  MD5_CTX md5_context;

#if defined(__linux__) || defined(__APPLE__)

  MD5_Init( &md5_context );
  MD5_Update( &md5_context, addr_const, path.size() );
  unsigned char md5_result_int[17];
  MD5_Final( md5_result_int, &md5_context);
  
  unsigned char md5_result[33];
  // have to convert the bytes manually to hex, because there's no MD5End
  char *curr_char = (char *) md5_result;
  for( uint i = 0; i < 16; i++ ) {
    sprintf( curr_char, "%02x", md5_result_int[i] );
    curr_char += 2;
  }

#else
    
    MD5Init( &md5_context );
    MD5Update( &md5_context, addr_const, 24 );
    char md5_result[33];
    MD5End( &md5_context, md5_result );
    
#endif /*__linux__*/

  // copy first 32 bits to a buffer
  char int1_buf[9];
  for( int i = 0; i < 8; i++ ) {
    int1_buf[i] = md5_result[i];
  }
  int1_buf[8] = '\0';

  const char *res_const = int1_buf;
  unsigned int a = 0;
  sscanf( res_const, "%x", &a );

  return (int) a;

}

static int hello_stat(fuse_ino_t ino, struct stat *stbuf)
{

  //printf( "Stat-ing %d\n", (int) ino );

  if( ino_to_path.find(ino) == ino_to_path.end() ) {
    return -1;
  }

  std::string path = ino_to_path[ino];

  int r = stat( path.c_str(), stbuf );
  stbuf->st_ino = ino;

  return r;
}

static void null_fuse_ll_getattr(fuse_req_t req, fuse_ino_t ino,
                             struct fuse_file_info *fi)
{
    struct stat stbuf;

    (void) fi;

    //printf( "In getattr\n" );

    memset(&stbuf, 0, sizeof(stbuf));
    if (hello_stat(ino, &stbuf) == -1)
        fuse_reply_err(req, ENOENT);
    else
        fuse_reply_attr(req, &stbuf, 1.0);
}

static void null_fuse_ll_lookup(fuse_req_t req, fuse_ino_t parent, const char *name)
{
    struct fuse_entry_param e;


    std::string p = ino_to_path[parent];
    std::string full = p + "/" + std::string(name);
    fuse_ino_t ino = (fuse_ino_t) path_to_ino(full);
    //printf( "In lookup -- full %s -> %d\n", full.c_str(), (int) ino );

    memset(&e, 0, sizeof(e));
    e.ino = ino;
    e.attr_timeout = 0.0;
    e.entry_timeout = 0.0;

    // make sure it's the actual current path, not some old, now 
    // unlinked path
    int x = stat( full.c_str(), &e.attr );
    e.attr.st_ino = ino;

    if( x == 0 ) {
      fuse_reply_entry(req, &e);
    } else {
      fuse_reply_err( req, ENOENT );
    }

}

void null_fuse_ll_mknod( fuse_req_t req, fuse_ino_t parent, 
		     const char *name, mode_t mode, dev_t rdev ) {

  struct fuse_entry_param e;

  std::string p = ino_to_path[parent];
  std::string path = p + "/" + std::string(name);
  int ino = path_to_ino(path);
  ino_to_path[ino] = path;

  //fprintf( stderr, "%s <-> %u\n", path.c_str(), (fuse_ino_t) ino );

  int fd = creat( path.c_str(), 0777 );
  if( fd < 0 ) {
    fuse_reply_err( req, ENOENT ); 
  } else {
    close(fd);
    e.ino = (fuse_ino_t) ino;
    hello_stat(e.ino, &e.attr);
    fuse_reply_entry( req, &e );
  }

}

void null_fuse_ll_create( fuse_req_t req, fuse_ino_t parent, 
			  const char *name, mode_t mode, 
			  struct fuse_file_info *fi ) {

  struct fuse_entry_param e;
  
  std::string p = ino_to_path[parent];
  std::string path = p + "/" + std::string(name);
  int ino = path_to_ino(path);
  ino_to_path[ino] = path;
  
  //fprintf( stderr, "%s <-> %u\n", path.c_str(), (fuse_ino_t) ino );

  int fd = creat( path.c_str(), 0777 );
  if( fd < 0 ) {
    fuse_reply_err( req, ENOENT ); 
  } else {
    close(fd);
    e.ino = (fuse_ino_t) ino;
    hello_stat(e.ino, &e.attr);
    fuse_reply_create( req, &e, fi );
  }

}

void null_fuse_ll_mkdir( fuse_req_t req, fuse_ino_t parent, 
		     const char *name, mode_t mode ) {

  struct fuse_entry_param e;

  std::string p = ino_to_path[parent];
  std::string path = p + "/" + std::string(name);
  int ino = path_to_ino(path);
  ino_to_path[ino] = path;

  //fprintf( stderr, "%s <-> %u\n", path.c_str(), (fuse_ino_t) ino );

  int r = mkdir( path.c_str(), 0777 );
  if( r < 0 ) {
    fuse_reply_err( req, ENOENT ); 
  } else {
    e.ino = (fuse_ino_t) ino;
    hello_stat(e.ino, &e.attr);
    fuse_reply_entry( req, &e );
  }

}

void null_fuse_ll_unlink( fuse_req_t req, fuse_ino_t parent, 
		      const char *name ) {

  std::string p = ino_to_path[parent];
  std::string path = p + "/" + std::string(name);
  //fprintf( stderr, "Unlinking path %s\n", path.c_str() );
  linked_paths.erase(path);
  if( unlink( path.c_str() ) < 0 ) {
    fuse_reply_err( req, ENOENT );
  } else {
    fuse_reply_err( req, 0 );
  }

}

void null_fuse_ll_rmdir( fuse_req_t req, fuse_ino_t parent, 
		     const char *name ) {

  std::string p = ino_to_path[parent];
  std::string path = p + "/" + std::string(name);
  if( rmdir( path.c_str() ) < 0 ) {
    fuse_reply_err( req, ENOENT );
  } else {
    fuse_reply_err( req, 0 );
  }

}


struct dirbuf {
    char *p;
    size_t size;
};

static void dirbuf_add(fuse_req_t req, struct dirbuf *b, const char *name,
                       fuse_ino_t ino)
{
    struct stat stbuf;
    size_t oldsize = b->size;
    b->size += fuse_add_direntry(req, NULL, 0, name, NULL, 0);
    b->p = (char *) realloc(b->p, b->size);
    memset(&stbuf, 0, sizeof(stbuf));
    stbuf.st_ino = ino;
    fuse_add_direntry(req, b->p + oldsize, b->size - oldsize, name, &stbuf,
                      b->size);
}

#define min(x, y) ((x) < (y) ? (x) : (y))

static int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize,
                             off_t off, size_t maxsize)
{
    if ((size_t)off < bufsize)
        return fuse_reply_buf(req, buf + off, min(bufsize - off, maxsize));
    else
        return fuse_reply_buf(req, NULL, 0);
}

static void null_fuse_ll_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
                             off_t off, struct fuse_file_info *fi)
{

    std::string p = ino_to_path[ino];
    DIR *dp = opendir(p.c_str());
    struct dirent *de;

    if( dp == NULL ) {
      fuse_reply_err(req, ENOTDIR);      
      return;
    }

    struct dirbuf b;
    memset(&b, 0, sizeof(b));
    while ((de = readdir(dp)) != NULL) {
      std::string path = p + "/" + std::string(de->d_name);
      int child_ino = path_to_ino(path);
      ino_to_path[child_ino] = path;
      //printf( "Mapping %s to %d\n", path.c_str(), child_ino );
      dirbuf_add(req, &b, de->d_name, child_ino);
    }

    reply_buf_limited(req, b.p, b.size, off, size);
    free(b.p);

    closedir(dp);

}

static void null_fuse_ll_open(fuse_req_t req, fuse_ino_t ino,
                         struct fuse_file_info *fi)
{
  /**
    if (ino != 2)
        fuse_reply_err(req, EISDIR);
    else if ((fi->flags & 3) != O_RDONLY)
        fuse_reply_err(req, EACCES);
    else
  */
  fuse_reply_open(req, fi);
}

static void null_fuse_ll_read(fuse_req_t req, fuse_ino_t ino, size_t size,
                         off_t off, struct fuse_file_info *fi)
{
    (void) fi;

    std::string p = ino_to_path[ino];

    int fd = open(p.c_str(), O_RDONLY);
    if( fd < 0 ) {
      fuse_reply_err( req, EIO );
      return;
    }

    char buf[size];
    if( lseek( fd, off, 0 ) == - 1 ) {
      close(fd);
      fuse_reply_err( req, EIO );
      return;
    }
    int s = read( fd, buf, size );
    close(fd);

    if( s < 0 ) {
      fuse_reply_err( req, EIO );
      return;
    }

    reply_buf_limited(req, buf, s, off, size);
}

void null_fuse_ll_write( fuse_req_t req, fuse_ino_t ino,
		     const char *buf, size_t size, off_t off,
		     struct fuse_file_info *fi ) {

    (void) fi;

    std::string p = ino_to_path[ino];

    //fprintf( stderr, "About to open %s for writing\n", p.c_str() );

    int fd = open(p.c_str(), O_WRONLY);
    if( fd < 0 ) {
      perror( "Couldn't open wronly" );
      fuse_reply_err( req, EIO );
      return;
    }

    //fprintf( stderr, "About to seek to %s\n", p.c_str() );

    if( lseek( fd, off, 0 ) == - 1 ) {
      perror( "Couldn't seek properly" );
      close(fd);
      fuse_reply_err( req, EIO );
      return;
    }

    //fprintf( stderr, "About to write to %s\n", p.c_str() );

    int s = write( fd, buf, size );
    close(fd);

    if( s < 0 ) {
      fuse_reply_err( req, EIO );
      return;
    }

    fuse_reply_write( req, s );

}

void null_fuse_ll_rename( fuse_req_t req, fuse_ino_t old_parent, 
			  const char *old_name,
			  fuse_ino_t new_parent, 
			  const char *new_name) {

  std::string p = ino_to_path[old_parent];
  std::string oldpath = p + "/" + std::string(old_name);
  std::string n = ino_to_path[new_parent];
  std::string newpath = n + "/" + std::string(new_name);

  fuse_ino_t ino = path_to_ino(oldpath);

  rename( oldpath.c_str(), newpath.c_str() );

  //fprintf( stderr, "renamed %s to %s, ino %u\n", oldpath.c_str(), newpath.c_str(), ino );

  linked_paths.erase(oldpath);
  linked_paths[newpath] = ino;
  ino_to_path[ino] = newpath;

  fuse_reply_err( req, 0 );

}

void null_fuse_ll_setattr( fuse_req_t req, fuse_ino_t ino, 
			   struct stat *attr, int to_set, 
			   struct fuse_file_info *fi ) {

  (void) fi;

  std::string p = ino_to_path[ino];

  if ( to_set & FUSE_SET_ATTR_UID ) {
    chown( p.c_str(), attr->st_uid, (gid_t)-1 );
  }
  if ( to_set & FUSE_SET_ATTR_GID ) {
    chown( p.c_str(), (uid_t)-1, attr->st_gid );
  }
  if( to_set & FUSE_SET_ATTR_MODE ) {
    chmod( p.c_str(), attr->st_mode );
  }

  
  struct stat stbuf;
  
  memset(&stbuf, 0, sizeof(stbuf));
  if (hello_stat(ino, &stbuf) == -1)
    fuse_reply_err(req, ENOENT);
  else
    fuse_reply_attr(req, &stbuf, 1.0);

}

void null_fuse_ll_link( fuse_req_t req, fuse_ino_t ino, 
			fuse_ino_t newparent, const char *newname ) {

  
  struct fuse_entry_param e;

  std::string p = ino_to_path[ino];
  std::string newp = ino_to_path[newparent];
  std::string path = newp + "/" + std::string(newname);

  if( link( p.c_str(), path.c_str() ) ) {
    perror( "Link failed!" );
    fuse_reply_err( req, EIO );
    return;
  }
  linked_paths[path] = ino;
  // This is broken if this path gets unlinked before the original path
  ino_to_path[ino] = path;

  //fprintf( stderr, "%s <-> %u\n", path.c_str(), (fuse_ino_t) ino );

  //fprintf( stderr, "Link file %s to new path %s, new ino %u\n", p.c_str(), path.c_str(), (fuse_ino_t) ino );

  memset(&e, 0, sizeof(e));
  e.ino = ino;
  e.attr_timeout = 0.0;
  e.entry_timeout = 0.0;
  int x = hello_stat(e.ino, &e.attr);
  if( x == 0 ) {
    fuse_reply_entry(req, &e);
  } else {
    fuse_reply_err( req, ENOENT );
  }

}

static struct fuse_lowlevel_ops null_fuse_ll_oper;

int main(int argc, char *argv[])
{
    struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
    struct fuse_chan *ch;
    char *mountpoint;
    int err = -1;

    umask(0);

    null_fuse_ll_oper.lookup  = null_fuse_ll_lookup;
    null_fuse_ll_oper.getattr = null_fuse_ll_getattr;
    null_fuse_ll_oper.readdir = null_fuse_ll_readdir;
    null_fuse_ll_oper.open    = null_fuse_ll_open;
    null_fuse_ll_oper.read    = null_fuse_ll_read;
    null_fuse_ll_oper.write   = null_fuse_ll_write;
    null_fuse_ll_oper.mknod   = null_fuse_ll_mknod;
    null_fuse_ll_oper.mkdir   = null_fuse_ll_mkdir;
    null_fuse_ll_oper.unlink  = null_fuse_ll_unlink;
    null_fuse_ll_oper.rmdir   = null_fuse_ll_rmdir;
    null_fuse_ll_oper.setattr = null_fuse_ll_setattr;
    null_fuse_ll_oper.link    = null_fuse_ll_link;
    null_fuse_ll_oper.create  = null_fuse_ll_create;
    null_fuse_ll_oper.rename  = null_fuse_ll_rename;

    ino_to_path[1] = ".";

    if (fuse_parse_cmdline(&args, &mountpoint, NULL, NULL) != -1 &&
        (ch = fuse_mount(mountpoint, &args)) != NULL) {
        struct fuse_session *se;

        se = fuse_lowlevel_new(&args, &null_fuse_ll_oper, 
			       sizeof(null_fuse_ll_oper),
                               NULL);
        if (se != NULL) {
            if (fuse_set_signal_handlers(se) != -1) {
                fuse_session_add_chan(se, ch);
                err = fuse_session_loop_mt(se);
                fuse_remove_signal_handlers(se);
                fuse_session_remove_chan(ch);
            }
            fuse_session_destroy(se);
        }
        fuse_unmount(mountpoint, ch);
    }
    fuse_opt_free_args(&args);

    return err ? 1 : 0;
}
