/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <sys/errno.h>

#include "wfs_lease_manager.h"
#include "wfs_misc.h"

const int32_t wfs_lease_manager::LEASE_BATCH_MARGIN = 5; // seconds

wfs_lease_manager::wfs_lease_manager() :
  _next_expiration(NULL), _last_batch(0), _in_wait_thread(false) {
  assert( pthread_mutex_init( &_lease_mutex, NULL ) == 0 );
  assert( pthread_cond_init( &_expire_cond, NULL ) == 0 );
  assert( pthread_mutex_init( &_pending_leases_lock, NULL ) == 0 );
}

wfs_lease_manager::~wfs_lease_manager() {
  // TODO: delete any existing lease structs in _leases*?
  pthread_mutex_destroy( &_lease_mutex );
  pthread_cond_destroy( &_expire_cond );
  pthread_mutex_destroy( &_pending_leases_lock );
}

void wfs_lease_manager::add_to_node_list( lease *xlease ) {

    lease *nodelease = _leases_by_node[xlease->node.id];
    if( nodelease == NULL ) {
      _leases_by_node[xlease->node.id] = xlease;
      _last_leases_by_node[xlease->node.id] = xlease;
    } else if( xlease->expiration <= nodelease->expiration ) {
      nodelease->prev_by_node = xlease;
      xlease->next_by_node = nodelease;
      _leases_by_node[xlease->node.id] = xlease;
    } else {

      lease *lastlease = _last_leases_by_node[xlease->node.id];
      assert( lastlease->next_by_node == NULL );

      // optimization -- if it's older than all the other
      // leases, dump it at the end without enumeration
      if( xlease->expiration >= lastlease->expiration ) {
	lastlease->next_by_node = xlease;
	xlease->prev_by_node = lastlease;
	_last_leases_by_node[xlease->node.id] = xlease;
      } else {

	while( nodelease != NULL &&
	       xlease->expiration > nodelease->expiration ) {
	  xlease->prev_by_node = nodelease;
	  nodelease = nodelease->next_by_node;
	}

	xlease->next_by_node = xlease->prev_by_node->next_by_node;
	if( xlease->next_by_node != NULL ) {
	  xlease->next_by_node->prev_by_node = xlease;
	}
	xlease->prev_by_node->next_by_node = xlease;

      }

    }

}

void wfs_lease_manager::remove_from_node_list( lease *xlease ) {

  // check if it was last lease
  lease *lastlease = _last_leases_by_node[xlease->node.id];
  if( lastlease == xlease ) {
    lease *newlast = xlease->prev_by_node;
    if( newlast != NULL ) {
      newlast->next_by_node = NULL;
      _last_leases_by_node[xlease->node.id] = newlast;
    } else {
      _last_leases_by_node.erase(xlease->node.id);
    }
  }

  // remove from by_node list
  if( xlease->prev_by_node == NULL ) {
    _leases_by_node[xlease->node.id] = xlease->next_by_node;
    if( xlease->next_by_node == NULL ) {
      _leases_by_node.erase(xlease->node.id);
    }
  } else {
    xlease->prev_by_node->next_by_node = xlease->next_by_node;
  }
  if( xlease->next_by_node != NULL ) {
    xlease->next_by_node->prev_by_node = xlease->prev_by_node;
  }

  xlease->prev_by_node = NULL;
  xlease->next_by_node = NULL;

}


// assumes lock is held
wfs_lease_manager::lease *
wfs_lease_manager::internal_add_lease( wfs_file_handle fh,
				       wfs_node node,
				       time_t expiration /*absolute*/,
				       int32_t stale_time ) {

  lease *xlease = _leases_by_fid[fh.fid];

  while( xlease != NULL && xlease->node.id != node.id ) {
    xlease = xlease->next_by_fid;
  }

  // if we already have one for this node, extend the deadline
  if( xlease != NULL ) {

    xlease->fh == fh; // update lease time
    xlease->expiration = expiration;
    xlease->stale_time = stale_time;

    // update place in node list
    remove_from_node_list(xlease);
    add_to_node_list(xlease);

    wfs_log(WFS_DBG_2, "Updating lease for %016"PRIx64
	      ", node %016"PRIx64", expiration %d\n",
	      fh.fid, node.id, (int)expiration );

  } else {

    wfs_log(WFS_DBG_2, "Adding new lease for %016"PRIx64
	      ", node %016"PRIx64", expiration %d\n",
	      fh.fid, node.id, (int)expiration );

    if( _new_leases.size() > 0 ) {
      xlease = _new_leases.back();
      _new_leases.pop_back();
    } else {
      xlease = new lease();
    }

    xlease->fh = fh;
    xlease->node = node;
    xlease->expiration = expiration;
    xlease->stale_time = stale_time;
    xlease->recently_used = false; // only set to true on a has_lease

    // goes at front of by_fid list
    xlease->next_by_fid = _leases_by_fid[fh.fid];
    _leases_by_fid[fh.fid] = xlease;

    // goes in the right place in by_node list
    add_to_node_list(xlease);
  }

  return xlease;

}

// assume lock is held
void wfs_lease_manager::internal_update_expiration_state( lease *xlease,
							  time_t old_expiration ) {

  wfs_node node = xlease->node;

  if( old_expiration == 0 || xlease->expiration < old_expiration ) {
    // interrupt timer and restart
    wfs_log(WFS_DBG_2, "Updating next_expiration to be %d\n", (int)xlease->expiration );
    _next_expiration = xlease;
    pthread_cond_signal( &_expire_cond );
  } else {
    wfs_log(WFS_DBG_2, "Next expiration of %d no longer than %d, no need to update\n",
	      (int)old_expiration, (int)xlease->expiration );
  }
}

void wfs_lease_manager::add_lease( wfs_file_handle fh, wfs_node node,
				   time_t deadline /*relative*/,
				   int32_t stale_time ) {

  lease *xlease;

  //wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);

  time_t now = time(NULL);
  time_t expiration = now + deadline;

  time_t old_expiration = 0;
  if( _next_expiration != NULL ) {
    old_expiration = _next_expiration->expiration;
  }

  // TODO: this mutex scope might be a bit broad
  pthread_mutex_lock( &_lease_mutex );
  xlease = internal_add_lease( fh, node, expiration, stale_time );

  internal_update_expiration_state( xlease, old_expiration );

  pthread_mutex_unlock( &_lease_mutex );

}

// assume lock is taken!
void wfs_lease_manager::internal_remove_lease( lease *xlease ) {

  wfs_log(WFS_DBG_2,  "Removing lease for (%016"PRIx64",%016"PRIx64")\n",
	    xlease->fh.fid,
	    xlease->node.id );

  // remove from by_fid list
  lease *fidlease = _leases_by_fid[xlease->fh.fid];
  if( fidlease->node.id == xlease->node.id ) {
    _leases_by_fid[xlease->fh.fid] = xlease->next_by_fid;
  } else {
    lease *prevlease = fidlease;
    while( fidlease != NULL && fidlease->node.id != xlease->node.id ) {
      prevlease = fidlease;
      fidlease = fidlease->next_by_fid;
    }
    prevlease->next_by_fid = xlease->next_by_fid;
  }
  if( _leases_by_fid[xlease->fh.fid] == NULL ) {
    _leases_by_fid.erase(xlease->fh.fid);
  }
  xlease->next_by_fid = NULL;

  remove_from_node_list(xlease);

  // and make sure it wasn't the next to expire
  if( _next_expiration != NULL && _next_expiration == xlease ) {
    internal_next_expiration();
    pthread_cond_signal( &_expire_cond );
  }

  _new_leases.push_back(xlease);

}

// assume lock is taken
void wfs_lease_manager::internal_next_expiration() {

  // now find the next expiration over all nodes
  hash_map< wfs_node_id, lease *, node_id_hash >::iterator i =
    _leases_by_node.begin();
  _next_expiration = NULL;

  for( ; i != _leases_by_node.end(); i++ ) {
    
    if( _next_expiration == NULL ||
	(i->second != NULL &&
	 i->second->expiration < _next_expiration->expiration) ) {
      // interrupt timer and restart
      _next_expiration = i->second;
    }
    
  }
  
}

bool wfs_lease_manager::has_lease( wfs_file_handle fh, wfs_node node ) {

  lease *xlease = NULL;

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  time_t now = time(NULL);

  pthread_mutex_lock( &_lease_mutex );

  // check these first so that the access below doesn't create them
  // unnecessarily
  if( (_leases_by_fid.find(fid) == _leases_by_fid.end()) ) {
    pthread_mutex_unlock( &_lease_mutex );
    return false;
  }

  lease *fidlease = _leases_by_fid[fid];
  while( fidlease != NULL ) {
    if( fidlease->node.id == node.id ) {
      xlease = fidlease;
      break;
    } else {
      fidlease = fidlease->next_by_fid;
    }
  }

  bool found = false;

  if( xlease != NULL ) {
    if( xlease->expiration < now ) {
      internal_remove_lease( xlease );
    } else {
      xlease->recently_used = true;
      found = true;
    }
  }

  pthread_mutex_unlock( &_lease_mutex );

  return found;

}

void wfs_lease_manager::leasing_nodes( wfs_file_handle fh,
				       std::vector<wfs_lease_manager::leaser> &nodes ) {

  //wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  time_t now = time(NULL);

  pthread_mutex_lock( &_lease_mutex );

  lease *xlease = _leases_by_fid[fh.fid];
  std::vector<lease *> to_remove;

  while( xlease != NULL ) {

    if( xlease->expiration < now ) {
      to_remove.push_back(xlease);
    } else {
      leaser node;
      node.node = xlease->node;
      node.time_left = xlease->expiration - now;
      node.stale_time = xlease->stale_time;
      nodes.push_back( node );
    }

    xlease = xlease->next_by_fid;

  }

  for( unsigned j = 0; j < to_remove.size(); j++ ) {
    internal_remove_lease( to_remove[j] );
  }

  pthread_mutex_unlock( &_lease_mutex );

}

void wfs_lease_manager::invalidate_lease( wfs_file_handle fh,
					  wfs_node *modifier ) {

  //wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);

  pthread_mutex_lock( &_lease_mutex );

  lease *xlease = _leases_by_fid[fh.fid];
  std::vector<lease *> to_remove;

  while( xlease != NULL ) {
    if( modifier == NULL || modifier->id != xlease->node.id ) {
      wfs_log(WFS_DBG_2, "Removing lease for file %016"PRIx64" for node %016"PRIx64"\n",
          fh.fid, xlease->node.id );
      assert( xlease != NULL );
      to_remove.push_back(xlease);
    }
    xlease = xlease->next_by_fid;
  }

  for( unsigned j = 0; j < to_remove.size(); j++ ) {
    internal_remove_lease( to_remove[j] );
  }

  pthread_mutex_unlock( &_lease_mutex );

  pthread_mutex_lock( &_pending_leases_lock );
  if( _pending_leases.find(fh.fid) != _pending_leases.end() ) {
    if( _pending_leases[fh.fid] > 0 ) {
      _pending_leases[fh.fid] = -_pending_leases[fh.fid];
    }
  }
  pthread_mutex_unlock( &_pending_leases_lock );

}

void wfs_lease_manager::wait_for_next_batch( wfs_node &node,
					     std::vector<wfs_file_handle> &fhs ) {

  pthread_mutex_lock( &_lease_mutex );

  assert(!_in_wait_thread);
  _in_wait_thread = true;

  // now that one is set, wait for it, or wait to be interrupted by
  // something sooner
  bool expired = false;
 outer:
  while( !expired ) {

    // if none are set yet, wait for something to come in first
    while( _next_expiration == NULL ) {
      wfs_log(WFS_DBG_2, "Waiting for some lease expiration to be set\n" );
      pthread_cond_wait( &_expire_cond, &_lease_mutex );
    }

    time_t now = time(NULL);
    while( _next_expiration->expiration < now ) {
      internal_remove_lease( _next_expiration );
      if( _next_expiration == NULL ) {
	expired = false;
	goto outer;
      }
    }

    struct timespec timeout;
    timeout.tv_sec = _next_expiration->expiration - LEASE_BATCH_MARGIN;
    // don't send out batch too often, it's expensive . . .
    if( timeout.tv_sec < (_last_batch + LEASE_BATCH_MARGIN) ) {
      //timeout.tv_sec = _last_batch + LEASE_BATCH_MARGIN;
      timeout.tv_sec = now+1;
    }
    timeout.tv_nsec = 0;

    if( timeout.tv_sec <= now ) {
      expired = true;
    } else {
      wfs_log(WFS_DBG_2, "About to wait until %d for a lease expiration on node %016"PRIx64"\n",
          (int)timeout.tv_sec, _next_expiration->node.id );

      expired = (pthread_cond_timedwait( &_expire_cond, &_lease_mutex,
					 &timeout ) == ETIMEDOUT);

      // it's possible the lease that we're expecting to expire was
      // already removed by another thread, in which case next_expiration
      // could be NULL.  In that case, loop around again
      if( _next_expiration == NULL ) {
	expired = false;
	wfs_log(WFS_DBG_2, "Lease expiration wait over, but next_expiration null\n",
		  expired );
      } else {
        wfs_log(WFS_DBG_2,  "Lease expiration wait over (expired=%d)\n", expired );
      }
    }
  }

  node = _next_expiration->node;

  // find all the leases that have been accessed since the last renewal
  time_t now = time(NULL);
  std::vector<lease *> to_remove;
  if( _leases_by_node.find(node.id) != _leases_by_node.end() ) {
    lease *xlease = _leases_by_node[node.id];
    while( xlease != NULL ) {
      if( xlease->recently_used ) {
	fhs.push_back(xlease->fh);
	xlease->recently_used = false;
      } else if( xlease->expiration < now ) {
	to_remove.push_back(xlease);
	//break;
      }
      xlease = xlease->next_by_node;
    }
  }

  for( unsigned j = 0; j < to_remove.size(); j++ ) {
    internal_remove_lease( to_remove[j] );
  }

  internal_next_expiration();
  _last_batch = now;

  _in_wait_thread = false;
  pthread_mutex_unlock( &_lease_mutex );

}

void wfs_lease_manager::renew_leases( wfs_node node,
				      std::vector<wfs_file_handle> fhs ) {

  pthread_mutex_lock( &_lease_mutex );

  // renew all the leases for their stated lease time
  time_t now = time(NULL);
  for( unsigned i = 0; i < fhs.size(); i++ ) {
    // update or add the lease, but don't mark as recently used!
    time_t expiration = now + fhlease(fhs[i]);
    lease *xlease = internal_add_lease( fhs[i], node,
					expiration,
					fhstaletime(fhs[i]) );
  }

  // update everything else
  // can't just check the lowest lease added above since it could have
  // updated the last _next_expiration, which will cause wait_for_next_batch
  // to skip over all expirations up to the newly-updated one
  internal_next_expiration();

  pthread_mutex_unlock( &_lease_mutex );

}

// We need to register that we have an outstanding pending lease,
// in case the the server grants the lease and then invalidates it
// before we hear that the lease was granted in the first place.

void wfs_lease_manager::reg_pending_lease( wfs_file_handle fh ) {
  pthread_mutex_lock( &_pending_leases_lock );
  if( _pending_leases[fh.fid] >= 0 ) {
    _pending_leases[fh.fid]++;
  } else {
    _pending_leases[fh.fid]--;
  }
  pthread_mutex_unlock( &_pending_leases_lock );
}

bool wfs_lease_manager::unreg_pending_lease( wfs_file_handle fh ) {

  bool lease_ok = false;

  pthread_mutex_lock( &_pending_leases_lock );

  assert(_pending_leases[fh.fid]);
  if( _pending_leases[fh.fid] > 0 ) {
    _pending_leases[fh.fid]--;
    lease_ok = true;
  } else {
    _pending_leases[fh.fid]++;
  }
  if( _pending_leases[fh.fid] == 0 ) {
    _pending_leases.erase(fh.fid);
  }
  pthread_mutex_unlock( &_pending_leases_lock );

  return lease_ok;

}

