/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_metadata_xattr.h"
#include <config.h>
#include <assert.h>
#ifdef HAVE_ATTR_XATTR_H
#include <attr/xattr.h>
#elif defined HAVE_SYS_XATTR_H
#include <sys/xattr.h>
#endif
#include <fcntl.h>
#include "wfs_misc.h"
#include "util/ident_util.h"

wfs_metadata_xattr::wfs_metadata_xattr( wfs_file_id id, int32_t vernum ) {

  fid = id;
  perm_cues = "";
  num_links = 0;

  if (vernum == 0) {
    char buf[WFS_FID_NUM_HEX_BYTES+5];
    snprintf( buf, WFS_FID_NUM_HEX_BYTES+5, "/%016"PRIx64"", fid );
    _filename = _dir + std::string( buf, strlen(buf) );
  } else {
    char buf[WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES];
    snprintf( buf, WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES,
	      "/%016"PRIx64"@%d", fid, vernum );
    _filename = _dir + std::string( buf, strlen(buf) );
  }

  read_in_file();

}

wfs_metadata_xattr::wfs_metadata_xattr( const wfs_metadata_xattr &md ) {

  fid = md.fid;
  version = md.version;
  perm_cues = md.perm_cues;
  permissions = md.permissions;
  num_links = md.num_links;

  char buf[WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES];
  snprintf( buf, WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES,
	    "/%016"PRIx64"", fid );
  _filename = _dir + std::string( buf, strlen(buf) );

}

wfs_metadata_xattr::~wfs_metadata_xattr() {
}

extern int max_md_size;

std::string wfs_metadata_xattr::xattr_name( std::string name ) {

  return "user." + name;

}

int wfs_metadata_xattr::getxattr( int fd, std::string name, char *buf,
				  int max_size ) {
#if defined HAVE_ATTR_XATTR_H || defined HAVE_SYS_XATTR_H
#ifdef __APPLE__
  return fgetxattr( fd, name.c_str(), buf, max_size, 0, 0 );
#else
  return fgetxattr( fd, name.c_str(), buf, max_size );
#endif
#else
  wfs_log(WFS_DBG_OFF, "Your system doesn't support xattr!\n" );
  assert(0);
#endif
}

int wfs_metadata_xattr::setxattr( int fd, std::string name, const char *buf,
				  int size ) {
#if defined HAVE_ATTR_XATTR_H || defined HAVE_SYS_XATTR_H
#ifdef __APPLE__
  return fsetxattr( fd, name.c_str(), buf, size, 0, 0 );
#else
  return fsetxattr( fd, name.c_str(), buf, size, 0 );
#endif
#else
  wfs_log(WFS_DBG_OFF, "Your system doesn't support xattr!\n" );
  assert(0);
#endif
}

bool wfs_metadata_xattr::read_in_file() {


  int fd = open( _filename.c_str(), O_RDONLY );

  if( fd < 0 ) {
    return false;
  }

  // read the version
  char buf[max_md_size];
  int n;
  char *p;
  char *x;

  std::string vername = xattr_name("ver");
  n = getxattr( fd, vername, buf, max_md_size );
  if( n <= 0 ) {
    perror( "ver failed" );
    close(fd);
    wfs_log(WFS_DBG_OFF, "Getxattr %s on file %s failed!!\n", vername.c_str(),
	    _filename.c_str() );
    return false;
  }

  uint32_t version_num = 0;
  wfs_node_id node_id = 0;

  buf[n] = '\0';
  wfs_log(WFS_DBG_2, "wfs_metadata:read_in_file: %s\n", buf);
  p = strtok_r( buf, ",", &x );
  assert( p != NULL );
  version_num = atoi(p);
  wfs_log(WFS_DBG_2, "version num (%s):%d\n", buf, version_num );
  p = strtok_r( NULL, ",", &x );
  assert( p != NULL );
  wfs_log(WFS_DBG_2, "node id (%s):%s\n", buf, p );
  uint32_t id1;
  uint32_t id2;
  sscanf( p, "%08"SCNx32"%08"SCNx32, &id1, &id2 );
  node_id = two_ints_to_node_id( id1, id2 );
  if( wfs_dbg_block(WFS_DBG_2) ) {
    wfs_log(WFS_DBG_2, "id1: %08x\n", id1);
    wfs_log(WFS_DBG_2, "id2: %08x\n", id2);
    wfs_log(WFS_DBG_2, "node id: %016"PRIx64"\n", node_id );
  }
  version = wfs_version( version_num, node_id );

  // read perm cues
  std::string permname = xattr_name("perm_cues");
  n = getxattr( fd, permname, buf, max_md_size );
  if( n < 0 ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Getxattr %s on file %s failed!!\n", permname.c_str(),
	    _filename.c_str() );
    return false;
  }
  buf[n] = '\0';
  perm_cues = std::string( buf, n );

  // read num hard links
  std::string linksname = xattr_name("num_links");
  n = getxattr( fd, linksname, buf, max_md_size );
  if( n < 0 ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Getxattr %s on file %s failed!!\n", linksname.c_str(),
	    _filename.c_str() );
    return false;
  }
  buf[n] = '\0';
  num_links = atoi(std::string( buf, n ).c_str());

  // Permissions
  // get number of permissions
  std::string permissions_name = xattr_name("permissions");
  n = getxattr( fd, permissions_name, buf, max_md_size );
  if( n <= 0 ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Getxattr %s on file %s failed!!\n", permissions_name.c_str(),
	    _filename.c_str() );
    return false;
  }
  buf[n] = '\0';

  p = strtok_r( buf, ",", &x );
  assert( p != NULL );
  int nr;
  sscanf (p, "%08x", &nr);

  for (int lcv = 0; lcv < nr; lcv++) {
    char name[65];
    wfs_permission perm;
    p = strtok_r( NULL, ",", &x );
    assert( p != NULL );
    sscanf(p, "%64[^=,]s=%08"SCNx32, name, &perm);
    std::string sname(name);
    permissions[name] = perm;
  }

  close(fd);
  return true;

}

bool wfs_metadata_xattr::write_helper(const char *fn) {

  int fd = open( fn, O_CREAT | O_RDONLY, 0666 );

  if( fd < 0 ) {
    return false;
  }

  wfs_log(WFS_DBG_2, "In wfs_metadata_xattr:write_helper for file %s\n", fn );

  // write the version
  char ver[WFS_INT_NUM_DEC_BYTES+1+WFS_FID_NUM_HEX_BYTES+2];
  std::string vername = xattr_name("ver");
  snprintf( ver, WFS_INT_NUM_DEC_BYTES+1+WFS_FID_NUM_HEX_BYTES+2,
	    "%d,%016"PRIx64",", version.version_num, version.node_id );
  if( setxattr( fd, vername, ver, strlen(ver) ) ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Setxattr %s on file %s failed!!\n", vername.c_str(), fn );
    return false;
  }

  // permanent cues
  std::string permname = xattr_name("perm_cues");
  if( setxattr( fd, permname, perm_cues.c_str(), perm_cues.size() ) ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Setxattr %s on file %s failed!!\n", permname.c_str(), fn );
    return false;
  }

  // number of hard links
  std::string linksname = xattr_name("num_links");
  snprintf( ver, WFS_INT_NUM_DEC_BYTES+2, "%d,", num_links );
  if( setxattr( fd, linksname, ver, strlen(ver) ) ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Setxattr %s on file %s failed!!\n", linksname.c_str(), fn );
    return false;
  }

  // permissions
  std::map<std::string,wfs_permission>::const_iterator i;
  std::string permissions_name = xattr_name("permissions");
  char permbuf[max_md_size];
  int n = 0;

  n = snprintf( permbuf, WFS_INT_NUM_HEX_BYTES+2,
		"%08"PRIx32",", (uint32_t)permissions.size() );
  for (i = permissions.begin(); i != permissions.end(); i++) {
    assert (i->first.length() <= 64);
    assert (i->first.find_first_of("=,") == std::string::npos);
    n += snprintf( permbuf+n, i->first.length()+1, "%s", i->first.c_str() );
    n += snprintf( permbuf+n, WFS_INT_NUM_HEX_BYTES+3, "=%08x,", i->second );
  }

  if( setxattr( fd, permissions_name, permbuf, n ) ) {
    close(fd);
    wfs_log(WFS_DBG_OFF, "Setxattr %s on file %s failed!!\n", permissions_name.c_str(), fn );
    return false;
  }

  //  fsync(fd);
  close( fd );

  return true;
}

